package ru.daotech.agroinvest

import android.content.SharedPreferences
import com.beust.klaxon.Klaxon
import com.f2prateek.rx.preferences2.RxSharedPreferences
import io.reactivex.Observable
import io.reactivex.subjects.BehaviorSubject
import io.reactivex.subjects.PublishSubject
import ly.count.android.sdk.Countly
import org.joda.time.DateTime
import ru.daotech.agroinvest.database.entities.TaskPlantImpl
import ru.daotech.agroinvest.models.LastFragmentState
import ru.daotech.agroinvest.models.Worker
import ru.daotech.agroinvest.service.events.Event
import ru.daotech.agroinvest.service.events.ServerData
import ru.daotech.agroinvest.service.jsonmodels.PlantPayload
import ru.daotech.agroinvest.service.jsonmodels.TaskPayload
import ru.daotech.agroinvest.utils.ConnectivityReceiver
import ru.daotech.agroinvest.utils.DateUtil
import ru.daotech.agroinvest.utils.DateUtil.Companion.getDateTimeWithDelta
import timber.log.Timber
import java.io.File
import java.io.StringReader
import java.util.*


class AppStateManager(sharedPreferences: SharedPreferences) {

    companion object {
        const val PREF_CURRENT_WORKER = "PREF_CURRENT_WORKER"
        const val PREF_CURRENT_WORKER_CODE = "PREF_CURRENT_WORKER_CODE"
        const val PREF_SESSION_ID = "PREF_SESSION_ID"
        const val PREF_SERVER_TIMESTAMP = "PREF_SERVER_TIMESTAMP"
        const val PREF_DEVICE_TIME = "PREF_DEVICE_TIME"
        const val PREF_ATTENTION_SHOW = "PREF_ATTENTION_SHOW"
        const val PREF_LAST_FRAGMENT_STATE = "PREF_LAST_FRAGMENT_STATE"
        const val PREF_WORK_SESSION = "PREF_WORK_SESSION"
        const val PREF_CURRENT_TROLLEY_NUMBER = "PREF_CURRENT_TROLLEY_NUMBER"
        const val PREF_LAST_ACCIDENT_MESSAGE = "PREF_LAST_ACCIDENT_MESSAGE`"
        const val PREF_LAST_ACCIDENT_TIME = "PREF_LAST_ACCIDENT_TIME`"
        const val ACTUAL_APP_INFO = "ACTUAL_APP_INFO"
        const val CURRENT_ROW_NUMBER = "CURRENT_ROW_NUMBER"
        const val CURRENT_TASK_TIME_START_MARK = "CURRENT_TASK_TIME_START_MARK"
        const val CURRENT_TROLLEY_TIED_TIME_MARK = "CURRENT_TROLLEY_TIED_TIME_MARK"
        const val CURRENT_BARCODE_GROUP_VALUE = "CURRENT_BARCODE_GROUP_VALUE"
        const val CURRENT_TASK_PLANT_ID = "CURRENT_TASK_PLANT_ID"
        const val DEFAULT_SESSION_ID = "-1"
        const val DEFAULT_SERVER_TIMESTAMP = 0
        const val DEFAULT_APP_INFO = ""
        const val SEPARATOR_SINGLE_DOT = "."
    }

    private val rxSharedPreferences = RxSharedPreferences.create(sharedPreferences)

    private val workSessionInfo = rxSharedPreferences.getString(PREF_WORK_SESSION)

    private val deviceTimeInfo = rxSharedPreferences.getString(PREF_DEVICE_TIME)

    private val currentWorkerCodePreferences =
        rxSharedPreferences.getString(PREF_CURRENT_WORKER_CODE)

    private val currentWorkerJson =
        rxSharedPreferences.getString(PREF_CURRENT_WORKER)

    private val sessionId =
        rxSharedPreferences.getString(PREF_SESSION_ID, DEFAULT_SESSION_ID)

    private val serverTimestamp =
        rxSharedPreferences.getInteger(PREF_SERVER_TIMESTAMP, DEFAULT_SERVER_TIMESTAMP)

    private val attentionShow =
        rxSharedPreferences.getBoolean(PREF_ATTENTION_SHOW)

    private val currentFragmentState =
        rxSharedPreferences.getString(PREF_LAST_FRAGMENT_STATE)

    private val actualAppInfo =
        rxSharedPreferences.getString(ACTUAL_APP_INFO, DEFAULT_APP_INFO)

    private val currentRowNumber =
        rxSharedPreferences.getString(CURRENT_ROW_NUMBER)

    private val currentTrolleyNumber =
        rxSharedPreferences.getString(PREF_CURRENT_TROLLEY_NUMBER)

    private val lastAccidentMessage =
        rxSharedPreferences.getString(PREF_LAST_ACCIDENT_MESSAGE)

    private val lastAccidentTime =
        rxSharedPreferences.getString(PREF_LAST_ACCIDENT_TIME)

    private val currentTimeStartMarkInfo =
        rxSharedPreferences.getString(CURRENT_TASK_TIME_START_MARK)

    private val currentTrolleyTiedTimeMarkInfo =
        rxSharedPreferences.getString(CURRENT_TROLLEY_TIED_TIME_MARK)

    private val currentBarcodeGroupValue =
        rxSharedPreferences.getString(CURRENT_BARCODE_GROUP_VALUE)

    private val currentTaskPlant =
        rxSharedPreferences.getString(CURRENT_TASK_PLANT_ID)

    private var currentWorker: Worker? = null
    private var filterRowNumber: String? = null
    private var isFirstLaunch: Boolean = true

    val onInitiateConnectToCouchDb: PublishSubject<Unit> = PublishSubject.create()
    val onCloseCouchDatabaseConnection: PublishSubject<Unit> = PublishSubject.create()
    val onSocketConnectivityState: BehaviorSubject<Boolean> = BehaviorSubject.create()
    val eventSubject: PublishSubject<Event> = PublishSubject.create()
    val isSessionIdChangedSubject: PublishSubject<String> = PublishSubject.create()
    val isWorkerChangedSubject: PublishSubject<Worker> = PublishSubject.create()
    val isFilterRowChangedSubject: PublishSubject<Unit> = PublishSubject.create()
    val onReconnectQueueReceiveSubject: PublishSubject<Unit> = PublishSubject.create()
    val onSignOutQueueReceiveSubject: PublishSubject<Unit> = PublishSubject.create()
    val uploadBackupMessageEventSubject: PublishSubject<Unit> = PublishSubject.create()
    val onSendMessageSocket: PublishSubject<String> = PublishSubject.create()
    val onEmitWorkerSignStatus: PublishSubject<Boolean> = PublishSubject.create()
    val onEmitSessionSend: PublishSubject<Unit> = PublishSubject.create()
    val onSocketStateRestart: PublishSubject<Unit> = PublishSubject.create()
    val queueCountEventSubject: PublishSubject<String> = PublishSubject.create()
    val fileProgressDownload: PublishSubject<Int> = PublishSubject.create()
    val fileLoadResultSubject: BehaviorSubject<File> = BehaviorSubject.create()
    val responseTaskInfoFromHttpSubject: BehaviorSubject<List<TaskPayload>> = BehaviorSubject.create()

    val connectionReceiver = ConnectivityReceiver()
    val screenshotMessageEventSubject: PublishSubject<Unit> = PublishSubject.create()

    fun saveCurrentWorker(worker: Worker) {
        currentWorkerJson.set(
            Klaxon().toJsonString(worker, null)
        )

        currentWorker = worker

        if (Countly.sharedInstance().isInitialized) {
            Countly.sharedInstance()
                .changeDeviceId(
                    worker.id + SEPARATOR_SINGLE_DOT +
                            worker.name + SEPARATOR_SINGLE_DOT + worker.personalNumber
                )
        }

        isWorkerChangedSubject.onNext(worker)
    }

    fun saveServerDeltaTimestamp(serverData: ServerData) {
        serverTimestamp.set(
            DateUtil.getSecondsByServerDateTime(serverData.serverDataPayload.timestamp)
        )
    }

    fun saveActualAppInfo(serverData: ServerData) {
        actualAppInfo.set(
            Klaxon().toJsonString(serverData, null)
        )
    }

    fun saveDeviceTimeInfo() {
        deviceTimeInfo.set(DateTime.now().toString())
    }

    fun saveCurrentRowNumber(rowNumber: String) {
        currentRowNumber.set(rowNumber)
    }

    fun saveFilterRowNumber(rowNumber: String?) {
        filterRowNumber = rowNumber
        isFilterRowChangedSubject.onNext(Unit)
    }

    fun saveCurrentTrolleyTiedTimeMark(currentTimeMark: String) {
        currentTrolleyTiedTimeMarkInfo.set(currentTimeMark)
    }

    fun saveCurrentBarcodeGroupValue(barcodeGroupValue: String) {
        currentBarcodeGroupValue.set(barcodeGroupValue)
    }

    fun getCurrentTrolleyTiedTimeMark(): String? {
        return currentTrolleyTiedTimeMarkInfo.get()
    }

    fun getCurrentBarcodeGroupValue(): String? {
        return currentBarcodeGroupValue.get()
    }

    fun saveCurrentTimeStartMark(currentTimeStartMark: String) {
        currentTimeStartMarkInfo.set(currentTimeStartMark)
    }

    fun getDeviceTimeInfo(): String? {
        return deviceTimeInfo.get()
    }

    fun saveLastFragmentState(lastFragmentState: LastFragmentState) {
        currentFragmentState.set(
            Klaxon().toJsonString(lastFragmentState, null)
        )
    }

    fun saveCurrentTaskPlant(taskPlant: TaskPlantImpl) {
        currentTaskPlant.set(
            Klaxon().toJsonString(
                PlantPayload(
                    plantId = taskPlant.id,
                    name = taskPlant.name
                ), null
            )
        )
    }

    fun getCurrentTaskPlant(): PlantPayload? {
        val taskPlantJsonObject =
            Klaxon().parseJsonObject(StringReader(currentTaskPlant.get()))
        return Klaxon().parseFromJsonObject<PlantPayload>(taskPlantJsonObject)
    }

    fun saveWorkSessionInfo(currentTimestamp: String) {
        workSessionInfo.set(currentTimestamp)
    }

    fun clearCurrentRowNumber() {
        currentRowNumber.delete()
    }

    fun clearWorkSessionPref() {
        workSessionInfo.delete()
    }

    fun clearCurrentWorker() {
        Timber.i("APPSTATEMANAGER: CLEAR CURRENT WORKER" + currentWorker?.id)
        currentWorkerJson.delete()
        currentWorker = null
    }

    fun isWorkSessionOverByDays(): Boolean {
        return if (!workSessionInfo.get().isBlank()) {
            DateUtil.isSessionToday(
                getDateTimeWithDelta(
                    dateString = workSessionInfo.get(),
                    delta = getServerTimestamp()
                )
            )
        } else {
            false
        }
    }

    fun getActualAppInfo(): ServerData? {
        return when (actualAppInfo.get().isBlank()) {
            true -> null
            false -> {
                val serverInfoJsonObject =
                    Klaxon().parseJsonObject(StringReader(actualAppInfo.get()))

                Klaxon().parseFromJsonObject<ServerData>(serverInfoJsonObject)
            }
        }
    }

    fun removeLastFragmentStatePreferences() {
        currentFragmentState.delete()
    }

    fun getLastFragmentState(): Observable<LastFragmentState>? {
        return currentFragmentState.asObservable()
            .take(1)
            .map { fragmentStateJsonString ->
                when (fragmentStateJsonString.isEmpty()) {
                    true -> {
                        return@map LastFragmentState()
                    }
                    false -> {
                        val fragmentStateJsonObject =
                            Klaxon().parseJsonObject(StringReader(fragmentStateJsonString))
                        Klaxon().parseFromJsonObject<LastFragmentState>(fragmentStateJsonObject)!!
                    }
                }
            }
    }

    fun deleteCurrentWorker() {
        Timber.i("APPSTATEMANAGER: DELETE CURRENT WORKER" + currentWorker?.id)

        isWorkerChangedSubject.onNext(currentWorker!!)

        if (Countly.sharedInstance().isInitialized) {
            Countly.sharedInstance().changeDeviceId(COUNTLY_DEFAULT_DEVICE_ID)
        }

        currentWorkerJson.delete()
    }

    fun getCurrentWorker(): Observable<Worker> {
        return currentWorkerJson.asObservable()
            .filter { workerJsonString ->
                workerJsonString.isNotEmpty()
            }
            .doOnError(Timber::e)
            .map { workerJsonString ->
                val workerJsonObject = Klaxon().parseJsonObject(StringReader(workerJsonString))
                Klaxon().parseFromJsonObject<Worker>(workerJsonObject)
            }
    }

    fun getCurrentWorkerSimple(): Worker? {
        return currentWorker
    }

    fun getServerTimestamp(): Int {
        return serverTimestamp.get()
    }

    fun isWorkerSet(): Observable<Boolean> {
        return Observable.just(currentWorkerJson.isSet)
    }

    fun setSessionId(): String {
        val oldSessionId = sessionId.get()

        sessionId.set(UUID.randomUUID().toString())

        isSessionIdChangedSubject.onNext(oldSessionId)
        onEmitSessionSend.onNext(Unit)

        return sessionId.get()
    }

    fun setIsFirstLaunch(value: Boolean): Boolean {
        isFirstLaunch = value
        return isFirstLaunch
    }

    fun getIsFirstLaunch(): Boolean {
        return isFirstLaunch
    }

    fun getSessionId(): String {
        return sessionId.get()
    }

    fun getCurrentRowNumber(): String {
        return currentRowNumber.get()
    }

    fun getFilterRowNumber(): String? {
        return if (filterRowNumber.isNullOrBlank()) {
            ""
        } else {
            filterRowNumber
        }
    }

    fun notShowAttentionAgain() {
        attentionShow.set(true)
    }

    fun markWorkerInfoNotSent(worker: Worker) {
        worker.isWorkerInfoSent = false
        currentWorkerJson.set(
            Klaxon().toJsonString(worker, null)
        )
    }

    fun saveCurrentTrolleyNumber(trolleyNumber: String) {
        Timber.i("APPSTATEMANAGER: NEW TROLLEY %s", trolleyNumber)
        currentTrolleyNumber.set(trolleyNumber)
    }

    fun getCurrentTrolleyNumber(): String {
        Timber.i("APPSTATEMANAGER: GET TROLLEY %s", currentTrolleyNumber.get())
        return currentTrolleyNumber.get()
    }

    fun clearCurrentTrolleyNumber() {
        Timber.i("APPSTATEMANAGER: DELETE TROLLEY %s", currentTrolleyNumber.get())
        currentTrolleyNumber.delete()
    }

    fun saveLastAccidentMessage(message: String) {
        lastAccidentMessage.set(message)
    }

    fun getLastAccidentMessage(): String? =
        lastAccidentMessage.get()

    fun saveLastAccidentTime(time: String) {
        lastAccidentTime.set(time)
    }

    fun getLastAccidentTime(): String? =
        lastAccidentTime.get()

    fun saveCurrentWorkerCode(workerCode: String) {
        currentWorkerCodePreferences.set(workerCode)
    }

    fun getCurrentWorkerCode(): String? = currentWorkerCodePreferences.get()
}