package ru.daotech.agroinvest.adapters

import android.Manifest
import android.content.Context
import android.content.pm.PackageManager
import android.graphics.Bitmap
import android.net.Uri
import android.os.Build
import android.view.LayoutInflater
import android.view.View
import android.view.View.GONE
import android.view.View.VISIBLE
import android.view.ViewGroup
import android.widget.EditText
import androidx.core.app.ActivityCompat
import androidx.core.content.ContextCompat
import androidx.fragment.app.FragmentActivity
import androidx.recyclerview.widget.RecyclerView
import com.google.android.material.textfield.TextInputEditText
import com.jakewharton.rxbinding2.view.clicks
import com.jakewharton.rxbinding2.widget.RxTextView
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.subjects.BehaviorSubject
import kotlinx.android.synthetic.main.item_header_view.view.*
import kotlinx.android.synthetic.main.item_image_binary_view.view.*
import kotlinx.android.synthetic.main.item_numeric_field_view.view.*
import kotlinx.android.synthetic.main.item_text_area_view.view.*
import ru.daotech.agroinvest.R
import ru.daotech.agroinvest.service.jsonmodels.TaskPropertyPayload
import ru.daotech.agroinvest.utils.CAMERA_PERMISSION_CODE
import ru.daotech.agroinvest.utils.PERMISSION_CODE
import ru.daotech.agroinvest.utils.ServiceUtil.Companion.bitmapToFile
import timber.log.Timber
import java.util.concurrent.TimeUnit

class HeaderFieldAdapter(
    private val context: Context,
    private val openFullscreenImage: ((Context, Uri?) -> Unit)?,
    private val takePhoto: () -> Unit,
    private val pickImageFromGallery: () -> Unit,
    private val fragmentActivity: FragmentActivity,
    private val showSnackBarError: (message: String?) -> Unit,
    private val showTextFieldHintDialog: ((String) -> Unit)?
) : RecyclerView.Adapter<HeaderFieldAdapter.BaseViewHolder<*>>() {

    var propertyList: List<Any> = emptyList()
    var disposable = CompositeDisposable()

    private var lastImageClickPos = 0

    companion object {
        private const val TYPE_TEXT = 0
        private const val TYPE_NUMERIC = 1
        private const val TYPE_IMAGE = 2
        private const val TYPE_HEADER = 3
        private const val TYPE_SIMPLE = 4
        private const val TYPE_GROUP = 5
    }

    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int): BaseViewHolder<*> {
        return when (viewType) {
            TYPE_TEXT -> {
                val view = LayoutInflater.from(context)
                    .inflate(R.layout.item_text_area_view, parent, false)
                TextAreaViewHolder(view)
            }
            TYPE_NUMERIC -> {
                val view = LayoutInflater.from(context)
                    .inflate(R.layout.item_numeric_field_view, parent, false)
                NumericFieldViewHolder(view)
            }
            TYPE_IMAGE -> {
                val view = LayoutInflater.from(context)
                    .inflate(R.layout.item_image_binary_view, parent, false)
                ImageBinaryViewHolder(view)
            }
            TYPE_HEADER -> {
                val view = LayoutInflater.from(context)
                    .inflate(R.layout.item_header_view, parent, false)
                HeaderViewHolder(view)
            }
            TYPE_SIMPLE -> {
                val view = LayoutInflater.from(context)
                    .inflate(R.layout.item_numeric_field_view, parent, false)
                SimpleFieldViewHolder(view)
            }
            TYPE_GROUP -> {
                val view = LayoutInflater.from(context)
                    .inflate(R.layout.item_numeric_field_view, parent, false)
                GroupFieldViewHolder(view)
            }
            else -> throw IllegalArgumentException("Invalid view type")
        }
    }

    override fun getItemCount() = propertyList.size

    inner class SimpleFieldViewHolder(itemView: View) :
        BaseViewHolder<SimpleTaskPropertyModel>(itemView) {

        override fun bind(item: SimpleTaskPropertyModel) {
            itemView.header_field_numeric_view.visibility = GONE
        }
    }

    inner class GroupFieldViewHolder(itemView: View) :
        BaseViewHolder<GroupModelView>(itemView) {

        override fun bind(item: GroupModelView) {
            itemView.header_field_numeric_view.visibility = GONE
        }
    }

    inner class TextAreaViewHolder(itemView: View) : BaseViewHolder<TextAreaModelView>(itemView) {

        override fun bind(item: TextAreaModelView) {
            disposable.add(
                item.onValidationSubject
                    .map {
                        with(itemView.header_field_text_area_view) {
                            error = "Обязательное поле для заполнения"
                            requestFocus()
                        }
                    }
                    .subscribe()
            )

            itemView.header_field_text_area_view.hint =
                """${item.property.name} ${item.property.regexp}"""

            (itemView.header_field_text_area_view as TextInputEditText).apply {
                RxTextView.textChanges(this)
                    .debounce(500, TimeUnit.MILLISECONDS)
                    .observeOn(AndroidSchedulers.mainThread())
                    .subscribe(
                        { value ->
                            (propertyList[adapterPosition] as TextAreaModelView).property.value =
                                value.toString()
                        },
                        { error ->
                            Timber.e(error, error.localizedMessage)
                        }
                    )
            }
        }
    }

    inner class NumericFieldViewHolder(itemView: View) :
        BaseViewHolder<NumericFieldModelView>(itemView) {

        override fun bind(item: NumericFieldModelView) {

            disposable.add(
                item.onValidationSubject
                    .map {
                        with(itemView.header_input_numeric_view) {
                            error = "Обязательное поле для заполнения"
                            requestFocus()
                        }
                    }
                    .subscribe()
            )

            itemView.header_field_numeric_view.hint =
                """${item.property.name} ${item.property.regexp}"""

            if ((propertyList[adapterPosition] as NumericFieldModelView)
                    .property.description.isNotBlank()
            ) {
                itemView.hint_drawable_button.visibility = VISIBLE
                itemView.hint_drawable_button.clicks()
                    .doOnNext {
                        showTextFieldHintDialog?.invoke(
                            (propertyList[adapterPosition] as NumericFieldModelView).property.description
                        )
                    }
                    .subscribe()
            } else {
                itemView.hint_drawable_button.visibility = GONE
            }

            (itemView.header_field_numeric_view as EditText).apply {
                RxTextView.textChanges(this)
                    .debounce(500, TimeUnit.MILLISECONDS)
                    .observeOn(AndroidSchedulers.mainThread())
                    .subscribe(
                        { value ->
                            (propertyList[adapterPosition] as NumericFieldModelView).property.value =
                                value.toString()

                            itemView.header_input_numeric_view.error = null
                            itemView.header_input_numeric_view.background?.clearColorFilter()
                        },
                        { error ->
                            Timber.e(error, error.localizedMessage)
                        }
                    )
            }
        }
    }

    inner class ImageBinaryViewHolder(itemView: View) :
        BaseViewHolder<ImageBinaryModelView>(itemView) {

        override fun bind(item: ImageBinaryModelView) {
            disposable.add(
                item.onValidationSubject
                    .observeOn(AndroidSchedulers.mainThread())
                    .subscribeOn(AndroidSchedulers.mainThread())
                    .map {
                        showSnackBarError.invoke("Для отправки необходимо изображение")
                        itemView.requestFocus()
                    }
                    .subscribe()
            )

            itemView.take_exist_view_button.setOnClickListener {
                lastImageClickPos = adapterPosition

                // Check runtime permission
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                    if (ContextCompat.checkSelfPermission(
                            context,
                            Manifest.permission.READ_EXTERNAL_STORAGE
                        ) == PackageManager.PERMISSION_DENIED
                    ) {
                        // Permission denied
                        val permissions = arrayOf(Manifest.permission.READ_EXTERNAL_STORAGE);
                        // Show popup to request runtime permission
                        ActivityCompat.requestPermissions(
                            fragmentActivity,
                            permissions,
                            PERMISSION_CODE
                        )
                    } else {
                        // Permission already granted
                        pickImageFromGallery()
                    }
                } else {
                    // System OS is < Marshmallow
                    pickImageFromGallery()
                }
            }

            itemView.take_photo_view_button.setOnClickListener {
                lastImageClickPos = adapterPosition

                //check runtime permission
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                    if (ContextCompat.checkSelfPermission(context, Manifest.permission.CAMERA) !=
                        PackageManager.PERMISSION_GRANTED
                        || ContextCompat.checkSelfPermission(
                            context,
                            Manifest.permission.READ_EXTERNAL_STORAGE
                        ) != PackageManager.PERMISSION_GRANTED
                    ) {
                        //permission denied
                        val permissions = arrayOf(
                            Manifest.permission.CAMERA,
                            Manifest.permission.READ_EXTERNAL_STORAGE
                        )
                        //show popup to request runtime permission
                        ActivityCompat.requestPermissions(
                            fragmentActivity,
                            permissions,
                            CAMERA_PERMISSION_CODE
                        )
                    } else {
                        //permission already granted
                        takePhoto()
                    }
                } else {
                    //system OS is < Marshmallow
                    takePhoto()
                }
            }

            itemView.phenology_preview_image_view.setOnClickListener {
                // Show image on full screen
                with(item) {
                    when (property.value.isBlank()) {
                        true -> {
                            openFullscreenImage?.invoke(context, null)
                        }
                        false -> {
                            uri?.let {
                                // Show image on full screen
                                openFullscreenImage?.invoke(
                                    context,
                                    it
                                )
                            }
                        }
                    }
                }
            }
        }
    }

    inner class HeaderViewHolder(itemView: View) : BaseViewHolder<HeaderModelView>(itemView) {

        override fun bind(item: HeaderModelView) {
            itemView.header_field_title_view.text = item.property.name
        }
    }

    override fun onBindViewHolder(holder: BaseViewHolder<*>, position: Int) {
        val element = propertyList[position]
        when (holder) {
            is TextAreaViewHolder -> holder.bind(element as TextAreaModelView)
            is NumericFieldViewHolder -> holder.bind(element as NumericFieldModelView)
            is ImageBinaryViewHolder -> holder.bind(element as ImageBinaryModelView)
            is HeaderViewHolder -> holder.bind(element as HeaderModelView)
            is SimpleFieldViewHolder -> holder.bind(element as SimpleTaskPropertyModel)
            is GroupFieldViewHolder -> holder.bind(element as GroupModelView)
            else -> throw IllegalArgumentException()
        }
    }

    fun setBitmapProperty(bitmap: Bitmap) {
        return when (val comparable = propertyList[lastImageClickPos]) {
            is ImageBinaryModelView -> {
                val bitmapMap = bitmapToFile(bitmap, context)
                comparable.property.value = bitmapMap["fileName"] as String
                comparable.uri = bitmapMap["fileUri"] as Uri
            }
            else -> {
            }
        }
    }

    override fun getItemViewType(position: Int): Int {
        val comparable = propertyList[position]
        return when (comparable) {
            is HeaderModelView -> TYPE_HEADER
            is ImageBinaryModelView -> TYPE_IMAGE
            is NumericFieldModelView -> TYPE_NUMERIC
            is TextAreaModelView -> TYPE_TEXT
            is SimpleTaskPropertyModel -> TYPE_SIMPLE
            is GroupModelView -> TYPE_GROUP
            else -> throw IllegalArgumentException("Invalid type of data " + position)
        }
    }

    fun getImageViewPos(): Int {
        return lastImageClickPos
    }

    fun handleSubmitForm(
        onPropertyRetrieveSubject: BehaviorSubject<Pair<List<TaskPropertyPayload?>, List<Uri>>>
    ) {
        // Do validation before send
        propertyList.forEach { property ->
            when (property) {
                is PropertyContainerField -> {
                    if (property.property.required && property.property.value.isBlank()) {
                        property.onValidationSubject.onNext(Unit)
                        onPropertyRetrieveSubject.onNext(
                            Pair(emptyList(), emptyList())
                        )
                        return
                    }
                }
            }
        }

        val uriImageList = propertyList
            .filterIsInstance<ImageBinaryModelView>()
            .map { taskProperty -> taskProperty.uri!! }

        // Send submitted data to arguments to DetailedFragment for task finishing
        onPropertyRetrieveSubject.onNext(
            Pair(
                propertyList.map {
                    when (it) {
                        is TextAreaModelView -> return@map it.property
                        is NumericFieldModelView -> return@map it.property
                        is ImageBinaryModelView -> return@map it.property
                        is HeaderModelView -> return@map it.property
                        is SimpleTaskPropertyModel -> return@map it.property
                        is GroupModelView -> return@map it.property
                        else -> null
                    }
                },
                uriImageList
            )
        )
    }

    abstract class BaseViewHolder<T>(itemView: View) : RecyclerView.ViewHolder(itemView) {
        abstract fun bind(item: T)
    }
}

data class SimpleTaskPropertyModel(
    val property: TaskPropertyPayload
)

data class TextAreaModelView(
    override val property: TaskPropertyPayload
) : PropertyContainerField(property)

data class NumericFieldModelView(
    override val property: TaskPropertyPayload
) : PropertyContainerField(property)

data class ImageBinaryModelView(
    override val property: TaskPropertyPayload,
    var bitmap: Bitmap?,
    var uri: Uri?
) : PropertyContainerField(property)

data class HeaderModelView(
    val property: TaskPropertyPayload
)

data class GroupModelView(
    val property: TaskPropertyPayload
)

abstract class PropertyContainerField(
    open val property: TaskPropertyPayload
) : ValidatorHeaderFieldView()

abstract class ValidatorHeaderFieldView(
    val onValidationSubject: BehaviorSubject<Unit> = BehaviorSubject.create()
)
