package ru.daotech.agroinvest.adapters

import android.graphics.Color
import android.view.LayoutInflater
import android.view.View
import android.view.View.GONE
import android.view.View.VISIBLE
import android.view.ViewGroup
import androidx.appcompat.widget.TooltipCompat
import androidx.recyclerview.widget.RecyclerView
import com.jakewharton.rxbinding2.view.clicks
import io.reactivex.subjects.PublishSubject
import kotlinx.android.synthetic.main.item_task.view.*
import org.joda.time.DateTime
import org.joda.time.Duration
import org.joda.time.Interval
import org.ocpsoft.prettytime.PrettyTime
import ru.daotech.agroinvest.R
import ru.daotech.agroinvest.autoNotify
import ru.daotech.agroinvest.database.entities.Task
import ru.daotech.agroinvest.hide
import ru.daotech.agroinvest.show
import ru.daotech.agroinvest.utils.DateUtil
import timber.log.Timber
import java.util.*
import kotlin.properties.Delegates

class TaskAdapter(
    private val taskClicks: PublishSubject<Task>
) : RecyclerView.Adapter<TaskAdapter.ViewHolderOther>() {

    var serverTimestamp: Int = 0

    var taskList: List<Task> by Delegates.observable(emptyList()) { _, old, new ->
        autoNotify(old, new) { o, n -> o.id == n.id }
    }

    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int): ViewHolderOther {
        val view = LayoutInflater.from(parent.context)
            .inflate(R.layout.item_task, parent, false)
        return ViewHolderOther(view)
    }

    override fun getItemCount() = taskList.size

    override fun onBindViewHolder(holder: ViewHolderOther, position: Int) {
        holder.bind(taskList[position])
    }

    inner class ViewHolderOther(itemView: View) : RecyclerView.ViewHolder(itemView) {
        fun bind(itemTask: Task) {
            itemView.task_title.text = itemTask.name
            itemView.taskRating.rating = itemTask.rating

            // TODO: Add another state (ic_done_all) when task data is synced with CC server
            if (itemTask.isFinished && itemTask.isTaskInfoSent) {
                itemView.task_finish_at.setCompoundDrawablesWithIntrinsicBounds(
                    null,
                    null,
                    itemView.context.getDrawable(
                        R.drawable.ic_cloud_done_25dp
                    ),
                    null
                )
            } else {
                itemView.task_finish_at.setCompoundDrawablesWithIntrinsicBounds(
                    null,
                    null,
                    null,
                    null
                )
            }

            itemView.taskInProgress.visibility = if(itemTask.inProgress) VISIBLE else GONE
            TooltipCompat.setTooltipText(itemView.taskInProgress, "На ряду проводятся работы!")

            if (itemTask.isFinished && itemTask.finishedAt != null) {
                val p = PrettyTime(Locale("ru"))
                val d = DateUtil.getCurrentDateTimestampWithDelta(itemTask.finishedAt!!, serverTimestamp)

                itemView.task_finish_at.text = "Завершено " + p.format(d)
            } else {
                val p = PrettyTime(Locale("ru"))
                val d = DateUtil.getDateWithServerTimeStamp(itemTask.finishAt)

                itemView.task_finish_at.text = "Окончание " + p.format(d)
            }

            if (itemTask.description.isEmpty()) {
                itemView.task_description.hide()
            } else {
                itemView.task_description.text = itemTask.description
                itemView.task_description.show()
            }

            itemView.task_description.hide() // let us not show the description for now

            if (itemTask.position == null) {
                itemView.task_position.hide()

                itemView.task_position_title.hide()
            } else {
                itemView.task_position.text = itemTask.position?.name
                itemView.task_position.show()

                itemView.task_position_title.show()
            }

            if (itemTask.department == null) {
                itemView.task_department.hide()

                itemView.task_department_title.hide()
            } else {
                itemView.task_department.text = itemTask.department?.fullName
                itemView.task_department.show()

                itemView.task_department_title.show()
            }

            if (itemTask.rowNumber == null) {
                itemView.taskRowNumberWrapper.hide()
            } else {
                itemView.taskRowNumber.text = itemTask.rowNumber.toString()
                itemView.taskRowNumberWrapper.show()
            }

            if (itemTask.section == null) {
                itemView.task_section.hide()

                itemView.task_section_title.hide()
            } else {
                itemView.task_section.text = itemTask.section?.fullName
                itemView.task_section.show()

                itemView.task_section_title.show()
            }

            val s: DateTime = DateUtil.getDateTimeWithDelta(itemTask.startAt, serverTimestamp)
            val f: DateTime = DateUtil.getDateTimeWithDelta(itemTask.finishAt, serverTimestamp)

            val quart = Duration(Interval(s, f).toDurationMillis() / 4)
            val sixth = Duration(Interval(s, f).toDurationMillis() / 6)

            val plentyOfTime = Interval(s, quart)
            val goodTimeToStart = Interval(s, quart + sixth)
            val backToWork = Interval(s, quart + sixth + sixth)
            val wellBetterToHurry = Interval(s, quart + sixth + sixth + sixth)

            itemView.task_type_line_container.setBackgroundColor(
                itemView.context.getColor(when {
                    itemTask.isFinished -> R.color.colorPrimary
                    plentyOfTime.containsNow() -> R.color.colorGreyAppText
                    goodTimeToStart.containsNow() -> R.color.colorAppAccentGreen
                    backToWork.containsNow() -> R.color.colorYellowInProgress
                    wellBetterToHurry.containsNow() -> R.color.colorOrangeInProgress
                    else -> R.color.colorAppAttentionRed
                })
            )

            itemView.task_kind_container.setBackgroundResource(
                when {
                    itemTask.isFinished -> {
                        itemView.task_kind_container.text = TaskTypeEnum.TASK_DONE.taskType
                        itemView.task_kind_container.setTextColor(Color.WHITE)
                        R.drawable.item_task_kind_container_done
                    }
                    itemTask.startedAt.isNullOrBlank() && !itemTask.isFinished -> {
                        itemView.task_kind_container.text = TaskTypeEnum.TASK_NEW.taskType
                        itemView.task_kind_container.setTextColor(Color.BLACK)
                        R.drawable.item_task_kind_container_new
                    }
                    !itemTask.startedAt.isNullOrBlank() && itemTask.finishedAt.isNullOrBlank() -> {
                        itemView.task_kind_container.text = TaskTypeEnum.TASK_IN_PROGRESS.taskType
                        itemView.task_kind_container.setTextColor(Color.WHITE)
                        R.drawable.item_task_kind_container_in_progress
                    }
                    else -> {
                        itemView.task_kind_container.text = TaskTypeEnum.TASK_ON_PAUSE.taskType
                        itemView.task_kind_container.setTextColor(Color.WHITE)
                        R.drawable.item_task_kind_container_pause
                    }
                }
            )

            itemView.clicks()
                .map { itemTask }
                .doOnError(Timber::e)
                .subscribe(taskClicks)
        }
    }
}

enum class TaskTypeEnum(val taskType: String) {
    TASK_NEW("Новая"),
    TASK_DONE("Готово"),
    TASK_ON_PAUSE("Пауза"),
    TASK_IN_PROGRESS("В работе");
}
