package ru.daotech.agroinvest.business

import io.reactivex.Observable
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.network.NetworkRepository
import ru.daotech.agroinvest.service.events.*
import timber.log.Timber
import java.util.concurrent.TimeUnit

open class EventObservableObject(
    val appStateManager: AppStateManager,
    private val networkRepository: NetworkRepository
) : EventObservable {

    override val signInEventObservable: Observable<SignIn>
        get() = appStateManager.eventSubject
            .filter { event -> event is SignIn }
            .debounce(500, TimeUnit.MILLISECONDS)
            .doOnError { Timber.e(it) }
            .cast(SignIn::class.java)

    override val signInMessageEventObservable: Observable<*>
        get() = appStateManager.eventSubject
            .filter { event -> event is SignInMessage }
            .timeout(DEFAULT_TIMEOUT_SECONDS, TimeUnit.SECONDS)
            .filter { event ->
                (event as SignInMessage).signInMessagePayload.interrupt?.let { interrupt ->
                    if (!interrupt) {
                        appStateManager.queueCountEventSubject.onNext(
                            event.signInMessagePayload.message
                        )
                        return@filter false
                    }
                    return@filter true
                }
                return@filter true
            }
            .onErrorResumeNext { error: Throwable ->
                Timber.e(error.localizedMessage)
                // Send manual signal on cache URL for last attempt to worker's data uploading
                networkRepository
                    .signInByCacheUrl(appStateManager.getCurrentWorkerCode()!!)
                    .timeout(DEFAULT_TIMEOUT_SECONDS, TimeUnit.SECONDS)
                    .doOnError(Timber::e)
                    .cast(SignIn::class.java)
            }

    override val signOutEventObservable: Observable<SignOut>
        get() = appStateManager.eventSubject
            .filter { event -> event is SignOut }
            .timeout(2, TimeUnit.SECONDS)
            .doOnError(Timber::e)
            .cast(SignOut::class.java)

    override val taskInfoEventObservable: Observable<TaskInfo>
        get() = appStateManager.eventSubject
            .filter { event -> event is TaskInfo }
            .timeout(5, TimeUnit.SECONDS)
            .cast(TaskInfo::class.java)

    override val trolleyInfoEventObservable: Observable<TrolleyInfo>
        get() = appStateManager.eventSubject
            .filter { event -> event is TrolleyInfo }
            .timeout(DEFAULT_TIMEOUT_SECONDS, TimeUnit.SECONDS)
            .take(COUNT_TAKE_RESPONSE_EVENT)
            .cast(TrolleyInfo::class.java)

    override val barcodeBoxEventObservable: Observable<BarcodeBoxMessage>
        get() = appStateManager.eventSubject
            .filter { event -> event is BarcodeBoxMessage }
            .onErrorResumeNext(Observable.empty())
            .doOnError(Timber::e)
            .cast(BarcodeBoxMessage::class.java)

    override val supportMessageEventObservable: Observable<SupportMessageInfo>
        get() = appStateManager.eventSubject
            .filter { event -> event is SupportMessageInfo }
            .timeout(DEFAULT_TIMEOUT_SECONDS, TimeUnit.SECONDS)
            .onErrorResumeNext(Observable.empty())
            .take(1)
            .doOnError(Timber::e)
            .cast(SupportMessageInfo::class.java)

    override val screenshotMessageEventObservable: Observable<ScreenshotMessage>
        get() = appStateManager.eventSubject
            .filter { event -> event is ScreenshotMessage }
            .timeout(DEFAULT_TIMEOUT_SECONDS, TimeUnit.SECONDS)
            .onErrorResumeNext(Observable.empty())
            .take(1)
            .doOnError(Timber::e)
            .cast(ScreenshotMessage::class.java)

    companion object {
        const val COUNT_TAKE_RESPONSE_EVENT = 1L
        const val DEFAULT_TIMEOUT_SECONDS = 20L
    }
}

interface EventObservable {

    val signInMessageEventObservable: Observable<*>

    val trolleyInfoEventObservable: Observable<TrolleyInfo>

    val taskInfoEventObservable: Observable<TaskInfo>

    val signOutEventObservable: Observable<SignOut>

    val signInEventObservable: Observable<SignIn>

    val barcodeBoxEventObservable: Observable<BarcodeBoxMessage>

    val screenshotMessageEventObservable: Observable<ScreenshotMessage>

    val supportMessageEventObservable: Observable<SupportMessageInfo>
}