package ru.daotech.agroinvest.business

import android.content.Context
import android.net.Uri
import android.webkit.URLUtil
import io.reactivex.Completable
import io.reactivex.Flowable
import io.reactivex.Observable
import io.reactivex.Single
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.exceptions.UndeliverableException
import io.reactivex.functions.BiFunction
import io.reactivex.plugins.RxJavaPlugins
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.PublishSubject
import io.reactivex.subjects.ReplaySubject
import ly.count.android.sdk.Countly
import retrofit2.Response
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.BuildConfig.signManageAttemptNumber
import ru.daotech.agroinvest.BuildConfig.signManageTimeoutSeconds
import ru.daotech.agroinvest.business.EventObservableObject.Companion.DEFAULT_TIMEOUT_SECONDS
import ru.daotech.agroinvest.database.DatabaseRepository
import ru.daotech.agroinvest.database.entities.*
import ru.daotech.agroinvest.database.relation.TaskWithProperties
import ru.daotech.agroinvest.exceptions.OutOfTasksInPauseException
import ru.daotech.agroinvest.exceptions.OutOfTasksInProgressException
import ru.daotech.agroinvest.exceptions.SignInWorkerException
import ru.daotech.agroinvest.exceptions.SignOutCamundaException
import ru.daotech.agroinvest.models.LastFragmentState
import ru.daotech.agroinvest.models.Rating
import ru.daotech.agroinvest.models.RequestTypeEnum
import ru.daotech.agroinvest.models.Worker
import ru.daotech.agroinvest.network.NetworkRepository
import ru.daotech.agroinvest.network.models.*
import ru.daotech.agroinvest.service.events.*
import ru.daotech.agroinvest.service.events.TrolleyInfo
import ru.daotech.agroinvest.service.jsonmodels.*
import ru.daotech.agroinvest.utils.*
import ru.daotech.agroinvest.utils.ServiceUtil.Companion.SCREEN_FILE_FORMAT
import ru.daotech.agroinvest.utils.ServiceUtil.Companion.SQLITE_FILE_FORMAT
import ru.daotech.agroinvest.utils.ServiceUtil.Companion.buildUploadedFilename
import ru.daotech.agroinvest.utils.ServiceUtil.Companion.getFileName
import timber.log.Timber
import java.io.*
import java.net.SocketException
import java.net.URL
import java.net.URLConnection
import java.util.*
import java.util.concurrent.TimeUnit
import javax.inject.Inject
import kotlin.collections.ArrayList

class RepositoryImpl @Inject constructor(
    private val networkRepository: NetworkRepository,
    private val databaseRepository: DatabaseRepository,
    private val appStateManager: AppStateManager,
    private val context: Context,
    private val eventObservableObject: EventObservableObject
) : Repository, VariableRequestBuilderImpl(appStateManager) {

    private val compositeDisposable = CompositeDisposable()

    private val onBroadcastMessage: ReplaySubject<BroadcastMessage> = ReplaySubject.create()
    private val onSupportMessage: ReplaySubject<SupportMessageInfo> = ReplaySubject.create()
    private val onSignalStartCamundaSubject: PublishSubject<String> = PublishSubject.create()

    override fun getMyTasksByRowNumber(rowNumber: String): Observable<List<Task>> {
        return appStateManager.getCurrentWorker()
            .flatMap { worker ->
                databaseRepository.getMyTasksByRowNumber(
                    namespace = appStateManager.getSessionId(),
                    workerId = worker.id,
                    rowNumber = rowNumber
                )
            }
    }

    override fun getMyStartedTasks(): Observable<List<Task>> {
        return appStateManager.getCurrentWorker()
            .flatMap { worker ->
                databaseRepository.getMyStartedTasks(
                    namespace = appStateManager.getSessionId(),
                    workerId = worker.id
                )
            }
    }

    override fun getMyPausedTasks(): Observable<List<Task>> {
        return appStateManager.getCurrentWorker()
            .flatMap { worker ->
                databaseRepository.getMyPausedTasks(
                    namespace = appStateManager.getSessionId(),
                    workerId = worker.id
                )
            }
    }

    override fun getMyTasksFlowable(): Flowable<List<Task>> {
        // Check on authorized user data
        if (appStateManager.getCurrentWorkerSimple() == null) {
            return Flowable.empty()
        }
        return databaseRepository.getMyTasksFlowable(
            namespace = appStateManager.getSessionId(),
            workerId = appStateManager.getCurrentWorkerSimple()!!.id,
            rowNumber = appStateManager.getFilterRowNumber()
        ).map { taskImplList ->
            taskImplList.map { it as Task }
        }
    }

    override fun getTasksHistory(): Observable<List<TaskWithProperties>> {
        return appStateManager.getCurrentWorker()
            .flatMap { worker ->
                databaseRepository.getTasksHistory(
                    namespace = appStateManager.getSessionId(),
                    workerId = worker.id
                )
            }
    }

    override fun getRootSections(): Observable<List<Section>> {
        return databaseRepository.getRootSections(namespace = appStateManager.getSessionId())
            .map { sectionImplList ->
                sectionImplList
                    .map { it as Section }
                    .sortedBy { section -> section.name }
            }
    }

    override fun getRatings(): Observable<List<Rating>> {
        return appStateManager.getCurrentWorker()
            .subscribeOn(Schedulers.io())
            .map { worker -> worker.ratings }
    }

    override fun getSectionParent(parentId: String?): Observable<List<Section>> {
        return databaseRepository.getSectionParent(
            parentId = parentId,
            namespace = appStateManager.getSessionId()
        )
            .map { sectionImplList ->
                sectionImplList
                    .map { it as Section }
                    .sortedBy { section -> section.name }
            }
    }

    override fun getSection(sectionId: String): Observable<Section> {
        return databaseRepository.getSection(sectionId = sectionId)
            .map { sectionImpl -> sectionImpl as Section }
            .toObservable()
    }

    override fun getSectionTasks(sectionId: String): Observable<List<TaskWithProperties>> {
        return databaseRepository.getSectionTasks(
            sectionId = sectionId,
            namespace = appStateManager.getSessionId()
        )
            .map { taskImplList ->
                taskImplList.map { it as TaskWithProperties }
            }
    }

    override fun getTaskWithProperties(taskId: String): Observable<TaskWithProperties> {
        return databaseRepository.getTaskWithProperties(
            taskId,
            appStateManager.getSessionId()
        )
            .map { taskWithPropertiesImpl -> taskWithPropertiesImpl as TaskWithProperties }
            .toObservable()
    }

    override fun saveTask(taskPayload: TaskPayload): Observable<Unit> {

        if (appStateManager.getCurrentWorkerSimple() == null) {
            return Observable.just(Unit)
        }

        return databaseRepository.getTaskExistence(
            namespace = appStateManager.getSessionId(),
            taskId = taskPayload.id,
            workerId = appStateManager.getCurrentWorkerSimple()!!.id
        )
            .take(1)
            .flatMapSingle { count -> saveResponseTaskInfo(count, taskPayload) }
    }

    override fun saveSupportMessageInfo(
        supportMessageInfo: SupportMessagePayload,
        messageCode: Int
    ): Observable<Unit> {
        return databaseRepository.insertSupportMessage(
            supportMessage = supportMessageInfo,
            workerId = appStateManager.getCurrentWorkerSimple()!!.id,
            time = DateUtil.getCurrentDateTimestampWithDelta(appStateManager.getServerTimestamp()),
            messageCode = messageCode
        )
            .toSingleDefault(Unit)
            .doOnError(Timber::e)
            .subscribeOn(Schedulers.single())
            .observeOn(AndroidSchedulers.mainThread())
            .toObservable()
    }

    override fun sendSupportUserMessage(message: String): Observable<Unit> {
        // Send message by socket
        appStateManager.onSendMessageSocket.onNext(message)
        return saveSupportMessageInfo(
            SupportMessagePayload(
                SUPPORT_MESSAGE_USER_TAG,
                message
            ),
            CHAT_MESSAGE_WORKER_CODE
        )
    }

    override fun saveTrolleyInfo(trolleyInfoPayload: TrolleyInfoPayload): Observable<Unit> {
        return databaseRepository
            .upsertTrolleyInfo(
                TrolleyInfoImpl(
                    from = trolleyInfoPayload,
                    namespace = appStateManager.getSessionId()
                )
            )
            .toSingleDefault(Unit)
            .doOnError(Timber::e)
            .toObservable()
    }

    private fun saveResponseTaskInfo(
        count: Int,
        taskPayload: TaskPayload
    ): Single<Unit> {
        return when (count) {
            1 -> {
                databaseRepository
                    .setTaskWorker(
                        namespace = appStateManager.getSessionId(),
                        taskId = taskPayload.id,
                        workerId = taskPayload.workerId,
                        workerName = taskPayload.workerName,
                        inProgress = taskPayload.inProgress,
                        orderBy = taskPayload.orderBy
                    )
                    .toSingleDefault(Unit)
                    .doOnError { Timber.e(it) }
                    .subscribeOn(Schedulers.single())
                    .observeOn(AndroidSchedulers.mainThread())
            }
            else -> {
                databaseRepository
                    .upsertTask(
                        TaskImpl(
                            from = taskPayload,
                            sessionId = appStateManager.getSessionId()
                        )
                    )
                    .andThen(
                        databaseRepository.upsertTaskSectionsList(
                            taskPayload.section?.structure!!.map { sectionPayloadLevel ->
                                SectionImpl(
                                    from = sectionPayloadLevel,
                                    sessionId = appStateManager.getSessionId()
                                )
                            }
                        )
                    )
                    .andThen(
                        databaseRepository.upsertTaskPropertiesList(
                            taskPayload.properties.map { taskProperty ->
                                TaskPropertyImpl(taskProperty, taskPayload.id)
                            }
                        )
                    )
                    .andThen(
                        databaseRepository.upsertTaskIntervalsList(
                            taskPayload.intervals.map { taskInterval ->
                                TaskIntervalImpl(taskInterval, taskPayload.id)
                            }
                        )
                    )
                    .andThen(
                        databaseRepository.upsertTaskMeasurementsList(
                            taskPayload.measurements.map { taskMeasurement ->
                                TaskMeasurementImpl(taskMeasurement, taskPayload.id)
                            }
                        )
                    )
                    .andThen(
                        taskPayload.plants?.let {
                            databaseRepository.upsertTaskPlantsList(
                                taskPayload.plants.map { plant ->
                                    TaskPlantImpl(plant, taskPayload.id)
                                }
                            )
                        }
                    )
                    .toSingleDefault(Unit)
                    .doOnError { Timber.e(it) }
                    .subscribeOn(Schedulers.single())
                    .observeOn(AndroidSchedulers.mainThread())
            }
        }
    }

    override fun downloadAppUrl(
        appUrl: String,
        appFolder: String,
        identifier: String
    ): Observable<File> {
        return Observable.fromCallable<File> {
            var count: Int
            try {
                val url = URL(appUrl)
                val connection: URLConnection = url.openConnection()
                connection.connect()
                val lengthOfFile: Int = connection.contentLength
                val input: InputStream = BufferedInputStream(url.openStream(), 8192)
                val file = File(
                    appFolder,
                    URLUtil.guessFileName(appUrl, null, null)
                )
                val output: OutputStream = FileOutputStream(file)
                val data = ByteArray(1024)
                var total: Long = 0
                while (input.read(data).also { count = it } != -1) {
                    total += count.toLong()
                    val progressValue = (total * 100 / lengthOfFile).toInt();
                    appStateManager.fileProgressDownload.onNext(progressValue)
                    output.write(data, 0, count)
                }
                output.flush()
                output.close()
                input.close()
                appStateManager.fileLoadResultSubject.onNext(file)
                return@fromCallable file
            } catch (e: Exception) {
                Timber.e(e)
                throw e
            }
        }
            .subscribeOn(Schedulers.io())
            .doOnError { Timber.e(it) }
    }

    override fun signIn(workerCode: String): Observable<Unit> {
        return databaseRepository
            .queueCamundaRequest(
                type = RequestTypeEnum.SIGN_IN.requestType,
                request = getVariableForSignIn(workerCode),
                namespace = appStateManager.getSessionId(),
                resourceId = null
            )
            .toObservable()
            .concatMap { flushCamundaQueue() }
            .subscribeOn(Schedulers.io())
            .observeOn(Schedulers.io())
            .zipWith(
                filterSignInResponseStatus(),
                BiFunction<Unit, Any, Worker> { _, signInEvent ->
                    // Catch response event for handling
                    when (signInEvent) {
                        is SignInMessage -> {
                            throw SignInWorkerException(signInEvent.signInMessagePayload.message)
                        }
                        is SignIn -> {
                            // Convert to SignIn event for correct handle
                            signInEvent.worker
                        }
                        else -> (signInEvent as SignIn).worker
                    }
                }
            )
            .map { worker ->
                if (appStateManager.getCurrentWorkerSimple() != null
                    && appStateManager.getCurrentWorkerSimple()?.id != worker.id
                ) {
                    // Clear trolley number, for starting day more correctly
                    appStateManager.clearCurrentTrolleyNumber()

                    // Clear raw number, for starting day more correctly
                    appStateManager.clearCurrentRowNumber()
                }
                appStateManager.saveCurrentWorker(worker)
                appStateManager.saveWorkSessionInfo(
                    DateUtil.getCurrentDateTimestampWithDelta(
                        appStateManager.getServerTimestamp()
                    )
                )

                databaseRepository.upsertRecentUser(RecentUserImpl(worker.id, worker.name, worker.personalNumber))

                // Send updated worker info to socket connection
                appStateManager.onEmitWorkerSignStatus.onNext(true)
            }
            .flatMap { LogoutReceiver.start(context) }
            .doOnError(Timber::e)
            .map { Unit }
    }

    override fun updatePersonalChat(chatResponse: ChatDataResponse): Single<Unit> {
        return databaseRepository.deleteAllMessages()
            .andThen(
                databaseRepository.updateUnreadStatusSupportMessages(
                    chatResponse.messages.map { message ->
                        SupportMessageImpl(
                            message = message.text,
                            workerId = appStateManager.getCurrentWorkerSimple()!!.id,
                            time = message.messageTime,
                            messageCode = CHAT_MESSAGE_SUPPORT_CODE
                        )
                    }
                )
            )
            .toSingleDefault(Unit)
            .doOnError { Timber.e(it) }
            .subscribeOn(Schedulers.single())
            .observeOn(AndroidSchedulers.mainThread())
    }

    override fun saveWorkerInfo(worker: Worker): Observable<Unit> {
        return Observable.just(appStateManager.saveCurrentWorker(worker))
    }

    private fun detectSignInResponse(): Observable<*> {
        return Observable.ambArray(
            eventObservableObject.signInEventObservable,
            eventObservableObject.signInMessageEventObservable
        )
    }

    private fun signInRequestByCacheUrl(): Observable<*> {
        return networkRepository.signInByCacheUrlWithAttempts(appStateManager.getCurrentWorkerCode()!!)
            .filter { event -> event is SignInMessage }
            .timeout(DEFAULT_TIMEOUT_SECONDS, TimeUnit.SECONDS)
            .filter { event ->
                (event as SignInMessage).signInMessagePayload.interrupt?.let { interrupt ->
                    if (!interrupt) {
                        appStateManager.queueCountEventSubject.onNext(
                            event.signInMessagePayload.message
                        )
                        signInRequestByCacheUrl()
                        return@filter false
                    }
                    return@filter true
                }
                return@filter false
            }
            .onErrorResumeNext { error: Throwable ->
                Timber.e(error.localizedMessage)
                // Send manual signal on cache URL for last attempt to worker's data uploading
                networkRepository
                    .signInByCacheUrl(appStateManager.getCurrentWorkerCode()!!)
                    .timeout(DEFAULT_TIMEOUT_SECONDS, TimeUnit.SECONDS)
                    .doOnError(Timber::e)
                    .cast(SignIn::class.java)
            }
    }

    private fun filterSignInResponseStatus(): Observable<*> {
        return Observable.range(1, signManageAttemptNumber + 1)
            .concatMap<Triple<Boolean, Int, Event?>> { i ->
                if (i != (signManageAttemptNumber + 1)) {
                    return@concatMap networkRepository.signInByCacheUrlWithAttempts(
                        appStateManager.getCurrentWorkerCode()!!
                    )
                        .timeout(DEFAULT_TIMEOUT_SECONDS, TimeUnit.SECONDS)
                        .onErrorReturnItem(SignInError)
                        .filter { event ->
                            when (event) {
                                is SignIn -> {
                                    return@filter true
                                }
                                is SignInMessage -> {
                                    event.signInMessagePayload.interrupt?.let { interrupt ->
                                        if (!interrupt) {
                                            appStateManager.queueCountEventSubject.onNext(
                                                event.signInMessagePayload.message
                                            )
                                            signInRequestByCacheUrl()
                                            return@filter false
                                        }
                                    }
                                    return@filter true
                                }
                                else -> {
                                    return@filter false
                                }
                            }
                        }
                        .map { event -> Triple(true, i, event) }
                        .timeout(signManageTimeoutSeconds.toLong() + 5, TimeUnit.SECONDS)
                } else {
                    return@concatMap Observable.just(Triple(false, i, null))
                }
            }
            .filter { (responseStatus, attempt, event) ->
                when {
                    !responseStatus && attempt == signManageAttemptNumber + 1 -> {
                        throw(SignOutCamundaException(
                            "Ошибка входа. Повторите попытку"
                        ))
                    }
                    !responseStatus && attempt <= signManageAttemptNumber -> {
                        return@filter false
                    }
                    responseStatus && event != null -> {
                        return@filter true
                    }
                    else -> {
                        return@filter false
                    }
                }
            }
            .take(1)
            .map { it.third }
    }

    override fun signOutFromQueue() {
        appStateManager.onSignOutQueueReceiveSubject.onNext(Unit)
    }

    override fun signOut(workerId: String): Observable<Boolean> {
        return Observable.zip(
            getMyStartedTasks()
                .take(1)
                .timeout(5, TimeUnit.SECONDS)
                .doOnError { Timber.e("SIGNOUT: getMyStartedTasks - TIMEOUT ERROR ") }
                .subscribeOn(Schedulers.io()),
            databaseRepository.getNotUploadedCamundaRequestsWOTaskListCount()
                .take(1)
                .timeout(5, TimeUnit.SECONDS)
                .doOnError { Timber.e("SIGNOUT: getNotUploadedCamundaRequestsWOTaskListCount - TIMEOUT ERROR ") }
                .subscribeOn(Schedulers.io()),
            BiFunction<List<Task>, Int, Pair<List<Task>, Int>> { tasks, count ->
                Pair(tasks, count)
            }
        )
            .flatMap { (tasks, count) ->
                when (tasks.size) {
                    0 -> {
                        when (count) {
                            0 -> filterSignOutResponseStatus(workerId)
                                .flatMap { clearFullSessionData(workerId) }
                                .flatMap {
                                    LogoutReceiver.cancel(context).timeout(10, TimeUnit.SECONDS)
                                }
                                .map { true }
                                .onErrorResumeNext { error: Throwable ->
                                    Timber.e(error.localizedMessage)
                                    Observable.just(false)
                                }
                            else -> throw(SignOutCamundaException(
                                "Есть неотправленные данные. " +
                                        "Найдите сеть"
                            ))
                        }
                    }
                    else -> throw(SignOutCamundaException(
                        "Есть задачи \"В РАБОТЕ\". " +
                                "Завершите задачи перед завершением рабочего дня"
                    ))
                }
            }
    }

    private fun clearFullSessionData(workerId: String): Observable<Unit> {
        return databaseRepository.clearTasksData(
            appStateManager.getSessionId(),
            workerId
        )
            .toSingleDefault(Unit)
            .toObservable()
            .flatMap { databaseRepository.clearAllSections() }
            .flatMap { databaseRepository.clearAllProperties() }
            .flatMap { databaseRepository.clearAllIntervals() }
            .flatMap { databaseRepository.clearAllCamundaRequests() }
            .take(1)
            .doOnError { Timber.e(it.localizedMessage) }
            .timeout(DEFAULT_TIMEOUT_SECONDS, TimeUnit.SECONDS)
    }

    private fun filterSignOutResponseStatus(workerId: String): Observable<Unit> {
        return Observable.range(1, signManageAttemptNumber + 1)
            .concatMap<Pair<Boolean, Int>> { i ->
                if (i != (signManageAttemptNumber + 1)) {
                    return@concatMap networkRepository.signOut(
                        getVariableForSignOut(workerId),
                        // If the current step is first, then need to make a zero delay
                        when (i) {
                            1 -> {
                                0 // Delay value on the first attempt
                            }
                            else -> {
                                signManageTimeoutSeconds.toLong()
                            }
                        }
                    )
                        .map { response ->
                            when (response.code()) {
                                HTTP_RESPONSE_CODE_OK -> {
                                    Timber.d(response.body().toString())
                                    return@map Pair(true, i)
                                }
                                else -> {
                                    Timber.d(response.errorBody().toString())
                                    return@map Pair(false, i)
                                }
                            }
                        }
                        .toObservable()
                        .timeout(signManageTimeoutSeconds.toLong() + 5, TimeUnit.SECONDS)
                        .onErrorResumeNext { error: Throwable ->
                            return@onErrorResumeNext Observable.just(Pair(false, i))
                        }
                } else {
                    return@concatMap Observable.just(Pair(true, i))
                }
            }
            .filter { (responseStatus, attempt) ->
                when {
                    !responseStatus && attempt == signManageAttemptNumber + 1 -> {
                        throw(SignOutCamundaException(
                            "Ошибка во время завершения рабочего дня. Повторите попытку"
                        ))
                    }
                    !responseStatus && attempt <= signManageAttemptNumber -> {
                        return@filter false
                    }
                    responseStatus -> {
                        return@filter true
                    }
                    else -> {
                        return@filter true
                    }
                }
            }
            .take(1)
            .map { Unit }
    }

    override fun signalToTaskListByCacheUrl(): Observable<Unit> {
        return networkRepository
            .loadTasksDataByCacheUrl(appStateManager.getCurrentWorkerSimple()!!.id)
            .map { taskInfoListHttpResponse ->
                appStateManager.responseTaskInfoFromHttpSubject.onNext(
                    taskInfoListHttpResponse
                )
            }
            .onErrorReturn { Unit }
            .timeout(DEFAULT_TIMEOUT_SECONDS, TimeUnit.SECONDS)
            .doOnError(Timber::e)
    }

    override fun signalToTaskList(workerId: String): Observable<Unit> {
        return databaseRepository
            .queueCamundaRequest(
                type = RequestTypeEnum.TASK_LIST.requestType,
                request = getVariableForTaskList(workerId),
                namespace = appStateManager.getSessionId(),
                resourceId = workerId
            )
            .toObservable()
            .zipWith(
                eventObservableObject.taskInfoEventObservable,
                BiFunction<Long, TaskInfo, Unit> { _, _ ->
                    Unit
                }
            )
            .onErrorResumeNext { error: Throwable ->
                Timber.e(error.localizedMessage)
                // Send manual signal on cache URL for task's data uploading
                networkRepository
                    .loadTasksDataByCacheUrl(appStateManager.getCurrentWorkerSimple()!!.id)
                    .map { taskInfoListHttpResponse ->
                        appStateManager.responseTaskInfoFromHttpSubject.onNext(
                            taskInfoListHttpResponse
                        )
                    }
                    .onErrorReturn { Unit }
                    .timeout(DEFAULT_TIMEOUT_SECONDS, TimeUnit.SECONDS)
                    .doOnError(Timber::e)
            }
    }

    override fun loadCurrentAppStatusByUrl(): Observable<ServerDataPayload> {
        return networkRepository.loadCurrentAppStatusByUrl();
    }

    override fun getSupportMessageSubject(): Observable<SupportMessageInfo> {
        appStateManager.eventSubject
            .filter { event -> event is SupportMessageInfo }
            .cast(SupportMessageInfo::class.java)
            .observeOn(AndroidSchedulers.mainThread())
            .subscribeOn(Schedulers.io())
            .doOnError { Timber.e(it) }
            .subscribe(onSupportMessage)

        return onSupportMessage
    }

    override fun signalToTrolleyInfoHistoryList(workerId: String): Observable<Unit> {

        return databaseRepository
            .queueCamundaRequest(
                type = RequestTypeEnum.TROLLEY_INFO.requestType,
                request = getVariableForTaskList(workerId),
                namespace = appStateManager.getSessionId(),
                resourceId = workerId
            )
            .toObservable()
            .zipWith(
                eventObservableObject.trolleyInfoEventObservable,
                BiFunction<Long, TrolleyInfo, Unit> { _, _ ->
                    Unit
                }
            )
            .doOnError { Timber.e(it) }
    }

    override fun signalToTaskStart(taskId: String): Observable<Unit> {
        return uploadTaskStartInfo(taskId)
    }

    override fun signalToTaskCancel(taskId: String): Observable<Unit> {
        return uploadTaskCancelInfo(taskId)
    }

    override fun signalToTaskFinish(taskId: String): Observable<Unit> {
        return uploadTaskFinishInfo(taskId)
    }

    override fun signalToGroupPropertyCreate(
        taskId: String,
        properties: List<TaskPropertyImpl>,
        groupId: String,
        barcodeValue: String
    ): Observable<Unit> {
        return uploadGroupProperty(taskId, properties, groupId, barcodeValue)
    }

    private fun uploadGroupProperty(
        taskId: String,
        properties: List<TaskPropertyImpl>,
        groupId: String,
        barcodeValue: String
    ): Observable<Unit> {
        return databaseRepository.upsertTaskPropertiesList(
            properties.map { taskProperty ->
                TaskPropertyImpl(taskProperty, taskId)
            }
        )
            .toSingleDefault(Unit)
            .toObservable()
            .doOnError { Timber.e(it) }
            .flatMap { signalGroupPropertyDataToUpload(taskId, properties, groupId) }
            .flatMapSingle {
                databaseRepository.upsertTaskMeasurement(
                    TaskMeasurementImpl(
                        groupId = groupId,
                        barcodeValue = barcodeValue,
                        taskId = taskId,
                        plantId = appStateManager.getCurrentTaskPlant()!!.plantId
                    )
                )
                    .toSingleDefault(Unit)
            }
            .map { Unit }
    }

    override fun uploadPropertyImages(uriList: List<Uri?>): Observable<Unit> {
        val uploadImageRequestList = ArrayList<Observable<FileUploadResponse>>()

        uriList.forEach { uri ->
            uri?.let {
                uploadImageRequestList.add(
                    networkRepository.screenshotUploadRequest(
                        file = File(uri.path),
                        meta = getVariableForFileUpload(appStateManager.getCurrentWorkerSimple()),
                        filename = getFileName(uri)!!
                    )
                )
            }
        }

        return Observable.merge(uploadImageRequestList)
            .toList()
            .subscribeOn(Schedulers.io())
            .observeOn(Schedulers.computation())
            .doOnError { Timber.e(it) }
            .map { Unit }
            .toObservable()
    }

    private fun uploadTaskStartInfo(taskId: String): Observable<Unit> {

        compositeDisposable.addAll(onSignalStartCamundaSubject
            .flatMap { signalStartedDataToUpload(taskId) }
            .doOnError { Timber.e(it) }
            .subscribe()
        )

        return Observable.zip(
            appStateManager.getCurrentWorker(),
            getMyStartedTasks(),
            BiFunction<Worker, List<Task>, Pair<Worker, List<Task>>> { worker, tasks ->
                Pair(worker, tasks)
            }
        )
            .flatMap { (worker, tasks) ->
                databaseRepository.getTaskWithProperties(
                    taskId,
                    appStateManager.getSessionId()
                )
                    .map { taskWithProperties ->
                        checkTaskInProgressAlready(taskId, tasks)
                        Pair(worker, taskWithProperties)
                    }
                    .toObservable()
            }
            .map {
                if (it.second.properties.isEmpty()) {
                    Countly.sharedInstance().recordEvent("EMPTY_POPRS " + it.second.embeddedTask.id)
                }
                it
            }
            .flatMapSingle { (worker, taskWithProperties) ->
                startTaskSignalDb(worker = worker, taskWithProperties = taskWithProperties)
            }
            .flatMapSingle { taskWithProperties ->
                if (!taskWithProperties.intervals.isNullOrEmpty()) {
                    addFinishMarkForIntervals(taskWithProperties.intervals)
                } else {
                    Single.just(Unit)
                }
            }
            .flatMapSingle { addStartMarkForIntervals(taskId) }
            .map { onSignalStartCamundaSubject.onNext(taskId) }
    }

    override fun uploadDatabaseFileDump(files: List<File>): Observable<Unit> {
        return networkRepository.fileUploadRequest(
            files = files,
            meta = getVariableForFileUpload(appStateManager.getCurrentWorkerSimple()),
            filename = buildUploadedFilename(
                appStateManager.getCurrentWorkerSimple(),
                SQLITE_FILE_FORMAT,
                appStateManager.getServerTimestamp()
            )
        )
            .doOnError { Timber.e(it) }
            .map { Unit }
            .subscribeOn(Schedulers.io())
            .observeOn(Schedulers.computation())
    }

    override fun sendAccidentReport(reportMessage: String): Observable<Unit> {
        return networkRepository.sendTaskAccidentReport(
            getVariableForAccidentReport(null, reportMessage)
        )
            .doOnError { Timber.e(it) }
            .flatMapCompletable {
                databaseRepository.insertSupportMessage(
                    SupportMessagePayload(
                        SUPPORT_MESSAGE_USER_TAG,
                        reportMessage
                    ),
                    appStateManager.getCurrentWorkerSimple()!!.id,
                    DateUtil.getCurrentDateTimestampWithDelta(appStateManager.getServerTimestamp()),
                    CHAT_MESSAGE_WORKER_CODE
                )
            }
            .toSingleDefault(Unit)
            .doOnError { Timber.e(it) }
            .subscribeOn(Schedulers.single())
            .observeOn(AndroidSchedulers.mainThread())
            .toObservable()
    }

    override fun sendAccidentComplexReport(
        taskWithProperties: TaskWithProperties?,
        reportMessage: String
    ): Observable<Unit> {
        return networkRepository.sendTaskAccidentReport(
            getVariableForAccidentReport(
                taskWithProperties,
                reportMessage
            )
        )
            .doOnError { Timber.e(it) }
            .map { Unit }
            .subscribeOn(Schedulers.io())
            .observeOn(Schedulers.io())
    }

    private fun checkTaskInProgressAlready(taskId: String, tasks: List<Task>) {
        repeat(tasks.filter { task ->
            task.id == taskId
                    && !task.startedAt.isNullOrBlank()
                    && task.finishedAt.isNullOrBlank()
        }.size) { return@checkTaskInProgressAlready }

        tasks
            .map { task ->
                if (task.id != taskId
                    && !task.startedAt.isNullOrBlank()
                    && task.finishedAt.isNullOrBlank()
                ) {
                    throw OutOfTasksInProgressException(
                        "Есть задачи в работе. Поставьте их на паузу, либо завершите задачи."
                    )
                }
            }
    }

    private fun checkTaskInPauseAlready(tasks: List<Task>) {
        repeat(tasks.filter { task ->
            (task.isFinished
                    || (task.startedAt.isNullOrBlank() && !task.isFinished)
                    || (!task.startedAt.isNullOrBlank() && task.finishedAt.isNullOrBlank()))
        }.size) { return@checkTaskInPauseAlready }

        tasks
            .map { task ->
                if (!(task.isFinished
                            || (task.startedAt.isNullOrBlank() && !task.isFinished)
                            || (!task.startedAt.isNullOrBlank() && task.finishedAt.isNullOrBlank()))
                ) {
                    throw OutOfTasksInPauseException(
                        "Есть задачи на паузе. Завершите задачи на паузе, либо запустите их в работу."
                    )
                }
            }
    }

    private fun startTaskSignalDb(
        taskWithProperties: TaskWithProperties,
        worker: Worker
    ): Single<TaskWithProperties> {
        return databaseRepository.upsertTask(
            TaskImpl(
                from = taskWithProperties,
                sessionId = appStateManager.getSessionId(),
                workerId = worker.id,
                startedAt = if (taskWithProperties.startedAt == null) DateUtil.getCurrentDateTimestampWithDelta(
                    appStateManager.getServerTimestamp()
                ) else taskWithProperties.startedAt,
                finishedAt = null,
                isTaskInfoSent = true,
                isFinished = false
            )
        ).toSingleDefault(taskWithProperties)
    }

    private fun finishTaskSignalDb(
        taskWithProperties: TaskWithProperties,
        worker: Worker
    ): Single<TaskWithProperties> {
        return databaseRepository.upsertTask(
            TaskImpl(
                from = taskWithProperties,
                sessionId = appStateManager.getSessionId(),
                workerId = worker.id,
                startedAt = taskWithProperties.startedAt,
                finishedAt = DateUtil.getCurrentDateTimestampWithDelta(
                    appStateManager.getServerTimestamp()
                ),
                isTaskInfoSent = true,
                isFinished = true
            )
        ).toSingleDefault(taskWithProperties)
    }

    private fun cancelTaskSignalDb(
        taskWithProperties: TaskWithProperties,
        worker: Worker
    ): Single<TaskWithProperties> {
        return databaseRepository.upsertTask(
            TaskImpl(
                from = taskWithProperties,
                sessionId = appStateManager.getSessionId(),
                workerId = worker.id,
                startedAt = taskWithProperties.startedAt,
                finishedAt = DateUtil.getCurrentDateTimestampWithDelta(
                    appStateManager.getServerTimestamp()
                ),
                isTaskInfoSent = true,
                isFinished = false
            )
        ).toSingleDefault(taskWithProperties)
    }

    private fun addStartMarkForIntervals(taskId: String): Single<Unit> {
        val taskIntervalMark = TaskIntervalImpl(
            id = UUID.randomUUID().toString(),
            taskId = taskId,
            startedAt = DateUtil.getCurrentDateTimestampWithDelta(
                appStateManager.getServerTimestamp()
            ),
            finishedAt = ""
        )
        return databaseRepository.insertTaskInterval(taskIntervalMark).toSingleDefault(Unit)
    }

    private fun uploadTaskCancelInfo(taskId: String): Observable<Unit> {
        return Observable.zip(
            appStateManager.getCurrentWorker(),
            getMyPausedTasks(),
            BiFunction<Worker, List<Task>, Pair<Worker, List<Task>>> { worker, taskWithProperties ->
                Pair(worker, taskWithProperties)
            }
        )
            .flatMap { (worker, tasks) ->
                databaseRepository.getTaskWithProperties(
                    taskId,
                    appStateManager.getSessionId()
                )
                    .map { taskWithProperties ->
                        checkTaskInPauseAlready(tasks)
                        Pair(worker, taskWithProperties)
                    }
                    .toObservable()
            }
            .flatMapSingle { (worker, taskWithProperties) ->
                cancelTaskSignalDb(worker = worker, taskWithProperties = taskWithProperties)
            }
            .flatMapSingle { taskWithProperties ->
                if (!taskWithProperties.intervals.isNullOrEmpty()) {
                    addFinishMarkForIntervals(taskWithProperties.intervals)
                } else {
                    Single.just(Unit)
                }
            }
            .doOnError { Timber.e(it) }
            .flatMap { signalCanceledDataToUpload(taskId) }
    }

    private fun uploadTaskFinishInfo(taskId: String): Observable<Unit> {
        return Observable.zip(
            appStateManager.getCurrentWorker(),
            databaseRepository.getTaskWithProperties(taskId, appStateManager.getSessionId())
                .toObservable(),
            BiFunction<Worker, TaskWithProperties, Pair<Worker, TaskWithProperties>> { worker, taskWithProperties ->
                Pair(worker, taskWithProperties)
            }
        )
            .flatMapSingle { (worker, taskWithProperties) ->
                finishTaskSignalDb(worker = worker, taskWithProperties = taskWithProperties)
            }
            .flatMapSingle { taskWithProperties ->
                if (!taskWithProperties.intervals.isNullOrEmpty()) {
                    addFinishMarkForIntervals(taskWithProperties.intervals)
                } else {
                    Single.just(Unit)
                }
            }
            .doOnError { Timber.e(it) }
            .flatMap { signalFinishedDataToUpload(taskId) }
            .map { Unit }
    }

    override fun signalAccidentReportToUpload(
        reportMessage: String,
        taskId: String?
    ): Observable<Unit> {
        return when (!taskId.isNullOrBlank()) {
            true -> databaseRepository.getTaskWithProperties(
                taskId,
                appStateManager.getSessionId()
            )
                .flatMapObservable { taskWithProperties ->
                    databaseRepository.queueCamundaRequest(
                        type = RequestTypeEnum.LOG_REPORT.requestType,
                        request = getVariableForAccidentReport(
                            taskWithProperties = null,
                            reportMessage = reportMessage
                        ),
                        namespace = appStateManager.getSessionId(),
                        resourceId = taskWithProperties.id
                    )
                        .map { it }
                        .toObservable()
                        .flatMap {
                            saveSupportMessageInfo(
                                SupportMessagePayload(
                                    SUPPORT_MESSAGE_USER_TAG,
                                    reportMessage
                                ),
                                CHAT_MESSAGE_WORKER_CODE
                            )
                        }
                        .map { Unit }
                }
                .subscribeOn(Schedulers.io())
            false -> databaseRepository.queueCamundaRequest(
                type = RequestTypeEnum.LOG_REPORT.requestType,
                request = getVariableForAccidentReport(
                    taskWithProperties = null,
                    reportMessage = reportMessage
                ),
                namespace = appStateManager.getSessionId(),
                resourceId = null
            )
                .map { it }
                .toObservable()
                .flatMap {
                    saveSupportMessageInfo(
                        SupportMessagePayload(
                            SUPPORT_MESSAGE_USER_TAG,
                            reportMessage
                        ),
                        CHAT_MESSAGE_WORKER_CODE
                    )
                }
                .map { Unit }
                .subscribeOn(Schedulers.io())
        }
    }

    override fun signalTrolleyTiedToUpload(trolleyNumber: String): Observable<Unit> {
        return databaseRepository.queueCamundaRequest(
            type = RequestTypeEnum.TROLLEY_START.requestType,
            request = getVariableForTrolleyTiedRequest(
                type = RequestTypeEnum.TROLLEY_START.requestType,
                trolleyNumber = trolleyNumber
            ),
            namespace = appStateManager.getSessionId(),
            resourceId = null
        )
            .toObservable()
            .flatMap {
                databaseRepository.upsertTrolleyInfo(
                    trolleyInfoImpl = TrolleyInfoImpl(
                        trolleyNumber = trolleyNumber.toInt(),
                        registeredAt = DateUtil.getCurrentDateTimestampWithDelta(
                            appStateManager.getServerTimestamp()
                        ),
                        workerId = appStateManager.getCurrentWorkerSimple()!!.id,
                        type = "start",
                        namespace = appStateManager.getSessionId()
                    )
                )
                    .toSingleDefault(Unit)
                    .toObservable()
            }
            .map { appStateManager.saveCurrentTrolleyNumber(trolleyNumber) }
            .doOnError { Timber.e(it) }
            .subscribeOn(Schedulers.io())
    }

    override fun unbindTrolley(trolleyUnbindPayload: TrolleyUnbindPayload): Observable<Unit> {
        Timber.d("TROLLEY UNBIND RECEIVED: " + trolleyUnbindPayload)

        return signalTrolleyUntiedToUpload(appStateManager.getCurrentTrolleyNumber())
    }

    override fun sendBarcodeBoxValueRequest(barcodeBoxValue: String): Observable<Unit> {
        Timber.d("BARCODE BOX SEND REQUEST ATTEMPT")

        return databaseRepository.queueCamundaRequest(
            type = RequestTypeEnum.BARCODE_BOX.requestType,
            request = getVariableForBarcodeBox(
                barcodeValue = barcodeBoxValue,
                workerId = appStateManager.getCurrentWorkerSimple()!!.id
            ),
            namespace = appStateManager.getSessionId(),
            resourceId = null
        )
            .toObservable()
            .doOnError { Timber.e(it) }
            .zipWith(
                eventObservableObject.barcodeBoxEventObservable,
                BiFunction<Long, BarcodeBoxMessage, Unit> { _, _ ->
                    Unit
                }
            )
    }

    override fun getAllSupportMessagesByCurrentWorker(): Flowable<List<SupportMessage>> {
        return databaseRepository.getAllSupportMessagesByWorkerId(
            workerId = appStateManager.getCurrentWorkerSimple()!!.id
        )
            .map { supportImplList ->
                supportImplList.map { it as SupportMessage }
            }
            .onErrorResumeNext { _: Throwable ->
                Flowable.just(emptyList())
            }
    }

    override fun getUnreadSupportMessageCount(): Flowable<Int> {
        return databaseRepository.getSupportMessageCount(
            workerId = appStateManager.getCurrentWorkerSimple()!!.id
        )
    }

    override fun signalTrolleyUntiedToUpload(trolleyNumber: String): Observable<Unit> {
        return databaseRepository.queueCamundaRequest(
            type = RequestTypeEnum.TROLLEY_END.requestType,
            request = getVariableForTrolleyTiedRequest(
                type = RequestTypeEnum.TROLLEY_END.requestType,
                trolleyNumber = appStateManager.getCurrentTrolleyNumber()
            ),
            namespace = appStateManager.getSessionId(),
            resourceId = null
        )
            .toObservable()
            .flatMap {
                databaseRepository.upsertTrolleyInfo(
                    trolleyInfoImpl = TrolleyInfoImpl(
                        trolleyNumber = trolleyNumber.toInt(),
                        registeredAt = DateUtil.getCurrentDateTimestampWithDelta(
                            appStateManager.getServerTimestamp()
                        ),
                        workerId = appStateManager.getCurrentWorkerSimple()!!.id,
                        type = "finish",
                        namespace = appStateManager.getSessionId()
                    )
                )
                    .toSingleDefault(Unit)
                    .toObservable()
            }
            .map { appStateManager.clearCurrentTrolleyNumber() }
            .doOnError { Timber.e(it) }
            .subscribeOn(Schedulers.io())
    }

    override fun getCollectionHistory(): Observable<List<TrolleyInfoImpl>> {
        return databaseRepository.getFullTrolleyInfoHistoryList(
            appStateManager.getCurrentWorkerSimple()!!.id,
            appStateManager.getSessionId()
        )
    }

    private fun addFinishMarkForIntervals(intervals: List<TaskInterval>): Single<Unit> {
        lateinit var taskInterval: TaskIntervalImpl
        intervals.takeLast(1).map { taskIntervalImpl ->
            taskInterval = TaskIntervalImpl(
                id = taskIntervalImpl.id,
                taskId = taskIntervalImpl.taskId,
                startedAt = taskIntervalImpl.startedAt,
                finishedAt = taskIntervalImpl.finishedAt
            )
        }
        return if (taskInterval.finishedAt.isEmpty()) {
            taskInterval.finishedAt = DateUtil.getCurrentDateTimestampWithDelta(
                appStateManager.getServerTimestamp()
            )
            databaseRepository.updateTaskInterval(taskInterval)
                .toSingleDefault(Unit)
        } else {
            Single.just(Unit)
        }
    }

    private fun addStartTimerMarkForIntervals(
        intervals: List<TaskInterval>,
        taskId: String
    ): Single<Unit> {
        lateinit var taskInterval: TaskIntervalImpl
        intervals.takeLast(1).map { taskIntervalImpl ->
            taskInterval = TaskIntervalImpl(
                id = taskIntervalImpl.id,
                taskId = taskIntervalImpl.taskId,
                startedAt = taskIntervalImpl.startedAt,
                finishedAt = taskIntervalImpl.finishedAt
            )
        }
        val taskIntervalMark = TaskIntervalImpl(
            id = UUID.randomUUID().toString(),
            taskId = taskId,
            startedAt = DateUtil.getCurrentDateTimestampWithDelta(
                appStateManager.getServerTimestamp()
            ),
            finishedAt = ""
        )
        return if (taskInterval.startedAt.isNotEmpty() && taskInterval.finishedAt.isEmpty()) {
            taskInterval.finishedAt = DateUtil.getCurrentDateTimestampWithDelta(
                appStateManager.getServerTimestamp()
            )
            databaseRepository.updateTaskInterval(taskInterval)
                .toSingleDefault(Unit)
                .flatMap {
                    databaseRepository.insertTaskInterval(taskIntervalMark).toSingleDefault(Unit)
                }
        } else {
            return databaseRepository.insertTaskInterval(taskIntervalMark).toSingleDefault(Unit)
        }
    }

    private fun addMarkTaskInfoSent(taskWithProperties: TaskWithProperties): Single<Unit> {
        return databaseRepository.upsertTask(
            TaskImpl(
                from = taskWithProperties,
                isTaskInfoSent = true
            )
        ).toSingleDefault(Unit)
    }

    private fun addMarkCamundaSent(
        camundaResponse: Response<CamundaResponse>,
        camundaRequest: Camunda
    ): Single<Unit> {
        return databaseRepository.upsertCamundaRequest(
            CamundaImpl(
                from = camundaRequest,
                responseCode = camundaResponse.code(),
                responseBody = camundaResponse.body()
            )
        ).toSingleDefault(Unit)
            .subscribeOn(Schedulers.io())
            .observeOn(AndroidSchedulers.mainThread())
    }

    override fun getServerDataSubject(): Observable<ServerData> {
        return appStateManager.eventSubject
            .filter { event -> event is ServerData }
            .cast(ServerData::class.java)
            .observeOn(AndroidSchedulers.mainThread())
            .subscribeOn(Schedulers.io())
            .map { serverData ->
                appStateManager.saveActualAppInfo(serverData)
                appStateManager.saveServerDeltaTimestamp(serverData)
                appStateManager.saveDeviceTimeInfo()
                return@map serverData
            }
            .doOnError { Timber.e(it) }
    }

    override fun getBroadcastMessageSubject(): Observable<BroadcastMessage> {
        appStateManager.eventSubject
            .filter { event -> event is BroadcastMessage }
            .cast(BroadcastMessage::class.java)
            .observeOn(AndroidSchedulers.mainThread())
            .subscribeOn(Schedulers.io())
            .doOnError { Timber.e(it) }
            .subscribe(onBroadcastMessage)

        return onBroadcastMessage
    }

    override fun getScreenshotMessageSubject(): Observable<Unit> {
        appStateManager.eventSubject
            .filter { event -> event is ScreenshotMessage }
            .cast(ScreenshotMessage::class.java)
            .observeOn(AndroidSchedulers.mainThread())
            .subscribeOn(Schedulers.io())
            .doOnError { Timber.e(it) }
            .map { appStateManager.screenshotMessageEventSubject.onNext(Unit) }
            .subscribe()

        return appStateManager.screenshotMessageEventSubject
    }

    override fun updateUnreadStatusSupportMessages(supportMessages: List<SupportMessage>): Completable {
        return databaseRepository.updateUnreadStatusSupportMessages(
            supportMessages.map { message -> (message as SupportMessageImpl).copy(isRead = true) }
        )
    }

    override fun fixTimerStartMark(taskId: String): Observable<Unit> {
        return databaseRepository.getTaskWithProperties(
            taskId,
            appStateManager.getSessionId()
        )
            .toObservable()
            .flatMapSingle { taskWithProperties ->
                if (!taskWithProperties.startedAt.isNullOrBlank() && taskWithProperties.finishedAt.isNullOrBlank()
                    || taskWithProperties.startedAt.isNullOrBlank() && !taskWithProperties.isFinished
                ) {
                    return@flatMapSingle addStartTimerMarkForIntervals(
                        taskWithProperties.intervals,
                        taskId
                    )
                        .map { Unit }
                } else {
                    return@flatMapSingle startTaskSignalDb(
                        worker = appStateManager.getCurrentWorkerSimple()!!,
                        taskWithProperties = taskWithProperties
                    )
                        .flatMap {
                            addStartTimerMarkForIntervals(taskWithProperties.intervals, taskId)
                                .map { Unit }
                        }
                        .map { Unit }
                }
            }
            .observeOn(AndroidSchedulers.mainThread())
            .doOnError { Timber.e(it) }
    }

    override fun uploadScreenshotImage(file: File): Observable<Unit> {
        return networkRepository.screenshotUploadRequest(
            file = file,
            meta = getVariableForFileUpload(appStateManager.getCurrentWorkerSimple()),
            filename = buildUploadedFilename(
                appStateManager.getCurrentWorkerSimple(),
                SCREEN_FILE_FORMAT,
                appStateManager.getServerTimestamp()
            )
        )
            .doOnError { Timber.e(it) }
            .map { Unit }
            .subscribeOn(Schedulers.io())
            .observeOn(Schedulers.computation())
    }

    private fun buildTaskUpdateRequestByTask(
        taskWithProperties: TaskWithProperties
    ): Observable<Long> {
        return databaseRepository.getTaskUpdateRequestByTaskId(taskWithProperties.id)
            .defaultIfEmpty(emptyList())
            .map { requests ->

                val requestElementPayload: SuperTaskRequest.PayloadRequest =
                    if (!requests.isNullOrEmpty()) {
                        parseSuperTaskRequestData(requests.first())
                    } else {
                        SuperTaskRequest.PayloadRequest(taskWithProperties)
                    }

                getSuperTaskRequestByTaskType(
                    RequestTypeEnum.TASK_START,
                    taskWithProperties,
                    requestElementPayload
                )
            }
            .flatMap { taskStartPayloadRequest ->
                writeCamundaRequestsToDb(
                    taskWithProperties,
                    appStateManager.getCurrentWorkerSimple()!!,
                    taskStartPayloadRequest,
                    RequestTypeEnum.TASK_UPDATE
                )
            }
    }

    override fun signalStartedDataToUpload(taskId: String): Observable<Unit> {
        return databaseRepository.getNotUploadedTaskWithProperties(taskId)
            .flatMapSingle { tasksWithProperties ->
                listenTaskStartEvent(tasksWithProperties)
            }
            .doOnError { Timber.e(it) }
            .map { Unit }
    }

    override fun signalFinishedDataToUpload(taskId: String): Observable<Unit> {
        return Observable.zip(
            appStateManager.getCurrentWorker(),
            databaseRepository.getNotUploadedTaskWithProperties(taskId),
            BiFunction<Worker, List<TaskWithProperties>, Pair<Worker, List<TaskWithProperties>>> { worker, taskWithProperties ->
                Pair(worker, taskWithProperties)
            }
        )
            .flatMapSingle { (worker, taskWithProperties) ->
                listenTaskFinishEvent(worker, taskWithProperties)
            }
            .doOnError { Timber.e(it) }
            .map { Unit }
    }

    override fun signalGroupPropertyDataToUpload(
        taskId: String,
        properties: List<TaskPropertyImpl>,
        groupId: String
    ): Observable<Unit> {
        return Observable.zip(
            appStateManager.getCurrentWorker(),
            databaseRepository.getNotUploadedTaskWithProperties(taskId),
            BiFunction<Worker, List<TaskWithProperties>, Pair<Worker, List<TaskWithProperties>>> { worker, taskWithProperties ->
                Pair(worker, taskWithProperties)
            }
        )
            .flatMapSingle { (worker, taskWithProperties) ->
                listenGroupPropertyEvent(worker, taskWithProperties, groupId)
            }
            .doOnError { Timber.e(it) }
            .map { Unit }
    }

    override fun signalCanceledDataToUpload(taskId: String): Observable<Unit> {
        return Observable.zip(
            appStateManager.getCurrentWorker(),
            databaseRepository.getNotUploadedTaskWithProperties(taskId),
            BiFunction<Worker, List<TaskWithProperties>, Pair<Worker, List<TaskWithProperties>>> { worker, taskWithProperties ->
                Pair(worker, taskWithProperties)
            }
        )
            .flatMapSingle { (worker, taskWithProperties) ->
                listenTaskCancelEvent(worker, taskWithProperties)
            }
            .doOnError { Timber.e(it) }
            .map { Unit }
    }

    override fun flushCamundaQueue(): Observable<Unit> {
        // RxJavaPlugins - intercept all undelivered errors from streams
        RxJavaPlugins.setErrorHandler { e ->
            var ex = e.cause
            if (e is UndeliverableException) {
                ex = e
            }
            if ((ex is IOException) || (ex is SocketException)) {
                // that's likely a bug in the application
                Thread.currentThread()
                    .uncaughtExceptionHandler.uncaughtException(Thread.currentThread(), e)
                // fine, irrelevant
                //
                // problem or API that throws on cancellation
                return@setErrorHandler
            }
            if (ex is InterruptedException) {
                // fine, some blocking code was interrupted by a dispose call
                return@setErrorHandler
            }
            if ((ex is NullPointerException) || (ex is IllegalArgumentException)) {
                // that's likely a bug in the application
                Thread.currentThread()
                    .uncaughtExceptionHandler.uncaughtException(Thread.currentThread(), e)
                return@setErrorHandler
            }
            if (ex is IllegalStateException) {
                // that's a bug in RxJava or in a custom operator
                Thread.currentThread().uncaughtExceptionHandler
                    .uncaughtException(Thread.currentThread(), ex)
                return@setErrorHandler
            }
            Timber.e("Undeliverable exception received, not sure what to do")
        }

        Timber.d("CAMUNDA SERVICE: QUEUE FLUSH!")
        return when (appStateManager.connectionReceiver.isConnected) {
            true -> databaseRepository.getNotUploadedCamundaRequests()
                .concatMapSingle { camundaRequests ->
                    sendCamundaRequests(camundaRequests)
                }
                .onErrorResumeNext(Observable.just(Unit))
                .map { Unit }
            false -> Observable.just(Unit)
        }
    }

    private fun listenTaskFinishEvent(
        worker: Worker,
        taskWithProperties: List<TaskWithProperties>
    ): Single<Unit> {
        return Observable
            .merge(getRequestListFromTaskFinish(taskWithProperties, worker))
            .toList()
            .map { Observable.merge(getMarkListFromSentTask(taskWithProperties)) }
            .map { Unit }
            .doOnError { ex -> Timber.e(ex) }
    }

    private fun listenGroupPropertyEvent(
        worker: Worker,
        taskWithProperties: List<TaskWithProperties>,
        groupId: String
    ): Single<Unit> {
        return Observable
            .merge(getRequestListFromGroupProperty(taskWithProperties, worker, groupId))
            .toList()
            .map { Observable.merge(getMarkListFromSentTask(taskWithProperties)) }
            .map { Unit }
            .doOnError { ex -> Timber.e(ex) }
    }

    private fun listenTaskCancelEvent(
        worker: Worker,
        taskWithProperties: List<TaskWithProperties>
    ): Single<Unit> {
        return Observable
            .merge(getRequestListFromTaskCancel(taskWithProperties, worker))
            .toList()
            .map { Observable.merge(getMarkListFromSentTask(taskWithProperties)) }
            .map { Unit }
            .doOnError { ex -> Timber.e(ex) }
    }

    private fun listenTaskStartEvent(taskWithProperties: List<TaskWithProperties>): Single<Unit> {
        return Observable
            .merge(getRequestListFromTaskStart(taskWithProperties))
            .toList()
            .map { Observable.merge(getMarkListFromSentTask(taskWithProperties)) }
            .map {
                compositeDisposable.clear()
                Unit
            }
            .doOnError { ex -> Timber.e(ex) }
    }

    private fun sendCamundaRequests(camundaRequests: List<Camunda>): Single<Unit> {
        return Observable.merge(getCamundaRequestList(camundaRequests)) // TODO: Write actual response to DB
            .toList()
            .map { responseList ->
                Observable.merge(getMarkListFromSentCamunda(responseList, camundaRequests))
                    .toList()
                    .subscribe()
            }
            .map { Unit }
    }

    override fun notShowAttentionAgain(): Observable<Unit> {
        return Observable.just(appStateManager.notShowAttentionAgain())
            .map { Unit }
    }

    override fun getLastFragmentInfo(): Observable<LastFragmentState>? {
        return appStateManager.getLastFragmentState()
    }

    override fun saveCurrentFragmentState(currentInterval: Long, taskId: String): Observable<Unit> {
        return Observable.just(
            appStateManager.saveLastFragmentState(
                LastFragmentState(
                    currentTaskInterval = currentInterval,
                    currentTaskId = taskId
                )
            )
        )
            .map { Unit }
    }

    override fun getCurrentWorker(): Observable<Worker> {
        return appStateManager.getCurrentWorker()
            .subscribeOn(Schedulers.io())
    }

    override fun clearLastTaskStatePreferences(): Observable<Unit> {
        return Observable.just(appStateManager.removeLastFragmentStatePreferences())
    }

    override fun getRecentUsers(): Observable<List<RecentUser>> {
        return databaseRepository.getRecentUsers()
                .map { it as List<RecentUser> }
    }

    private fun getMarkListFromSentCamunda(
        responseList: List<Response<CamundaResponse>>,
        camundaRequests: List<Camunda>
    ): ArrayList<Observable<Unit>> {

        val markList = ArrayList<Observable<Unit>>()
        for (response in responseList) {
            for (camundaRequest in camundaRequests) {
                markList.add(addMarkCamundaSent(response, camundaRequest)
                    .doOnError { Timber.e(it) }
                    .toObservable()
                )
            }
        }
        return markList
    }

    private fun getMarkListFromSentTask(taskWithProperties: List<TaskWithProperties>)
            : ArrayList<Observable<Unit>> {
        val markList = ArrayList<Observable<Unit>>()
        for (taskWithProperty in taskWithProperties) {
            markList.add(addMarkTaskInfoSent(taskWithProperty).toObservable())
        }
        return markList
    }

    private fun getCamundaRequestList(
        camundaRequests: List<Camunda>
    ): ArrayList<Observable<Response<CamundaResponse>>> {
        val requestList = ArrayList<Observable<Response<CamundaResponse>>>()
        camundaRequests
            .filter { camunda -> camunda.responseCode == null }
            .map { camunda ->
                requestList.add(
                    networkRepository.signalToCamunda(
                        camunda.payload
                    )
                        .subscribeOn(Schedulers.io())
                        .observeOn(Schedulers.computation())
                        .toObservable()
                )
            }
        return requestList
    }

    private fun getRequestListFromTaskStart(
        tasksWithProperties: List<TaskWithProperties>
    ): ArrayList<Observable<Long>> {
        val requestList = ArrayList<Observable<Long>>()

        tasksWithProperties
            .filter { taskWithProperties ->
                !taskWithProperties.isFinished
                        && taskWithProperties.finishedAt.isNullOrEmpty()
            }
            .map { taskWithProperties ->
                requestList.add(buildTaskUpdateRequestByTask(taskWithProperties))
            }

        return requestList
    }

    private fun getRequestListFromTaskCancel(
        tasksWithProperties: List<TaskWithProperties>,
        worker: Worker
    ): ArrayList<Observable<Long>> {
        val requestList = ArrayList<Observable<Long>>()

        tasksWithProperties
            .filter { taskWithProperties ->
                !taskWithProperties.isFinished
                        && !taskWithProperties.finishedAt.isNullOrEmpty()
            }
            .map { taskWithProperties ->
                requestList.add(
                    databaseRepository.getTaskUpdateRequestByTaskId(taskWithProperties.id)
                        .defaultIfEmpty(emptyList())
                        .map { requests ->
                            val requestElementPayload: SuperTaskRequest.PayloadRequest =
                                if (!requests.isNullOrEmpty()) {
                                    parseSuperTaskRequestData(requests.first())
                                } else {
                                    SuperTaskRequest.PayloadRequest(taskWithProperties)
                                }

                            getSuperTaskRequestByTaskType(
                                RequestTypeEnum.TASK_CANCEL,
                                taskWithProperties,
                                requestElementPayload
                            )
                        }
                        .flatMap { taskStartPayloadRequest ->
                            writeCamundaRequestsToDb(
                                taskWithProperties,
                                worker,
                                taskStartPayloadRequest,
                                RequestTypeEnum.TASK_UPDATE
                            )
                        }
                )
            }

        return requestList
    }

    private fun getRequestListFromTaskFinish(
        tasksWithProperties: List<TaskWithProperties>,
        worker: Worker
    ): ArrayList<Observable<Long>> {
        val requestList = ArrayList<Observable<Long>>()

        tasksWithProperties
            .filter { taskWithProperties -> taskWithProperties.isFinished }
            .forEach { taskWithProperties ->
                requestList.add(
                    databaseRepository.getTaskUpdateRequestByTaskId(taskWithProperties.id)
                        .defaultIfEmpty(emptyList())
                        .map { requests ->
                            val requestElementPayload: SuperTaskRequest.PayloadRequest =
                                if (!requests.isNullOrEmpty()) {
                                    parseSuperTaskRequestData(requests.first())
                                } else {
                                    SuperTaskRequest.PayloadRequest(taskWithProperties)
                                }

                            getSuperTaskRequestByTaskType(
                                RequestTypeEnum.TASK_FINISH,
                                taskWithProperties,
                                requestElementPayload
                            )
                        }
                        .flatMap { taskStartPayloadRequest ->
                            writeCamundaRequestsToDb(
                                taskWithProperties,
                                worker,
                                taskStartPayloadRequest,
                                RequestTypeEnum.TASK_UPDATE
                            )
                        }
                )
            }

        return requestList
    }

    private fun getRequestListFromGroupProperty(
        tasksWithProperties: List<TaskWithProperties>,
        worker: Worker,
        groupId: String
    ): ArrayList<Observable<Long>> {
        val requestList = ArrayList<Observable<Long>>()

        tasksWithProperties
            .filter { taskWithProperties ->
                !taskWithProperties.isFinished
                        && taskWithProperties.finishedAt.isNullOrEmpty()
            }
            .map { taskWithProperties ->
                requestList.add(
                    databaseRepository.getTaskUpdateRequestByTaskId(taskWithProperties.id)
                        .defaultIfEmpty(emptyList())
                        .map { requests ->
                            val requestElementPayload: GroupPropertyRequest.PayloadRequest =
                                if (!requests.isNullOrEmpty()) {
                                    val groupPropertyRequest =
                                        parseGroupPropertyRequestData(requests.first())

                                    with(groupPropertyRequest) {
                                        barcodeValue?.let {
                                            barcodeValue =
                                                appStateManager.getCurrentBarcodeGroupValue()!!
                                        }
                                        plantId?.let {
                                            plantId =
                                                appStateManager.getCurrentTaskPlant()!!.plantId
                                        }
                                    }

                                    groupPropertyRequest
                                } else {
                                    GroupPropertyRequest.PayloadRequest(
                                        taskWithProperties,
                                        appStateManager.getCurrentBarcodeGroupValue()!!,
                                        appStateManager.getCurrentTaskPlant()!!.plantId
                                    )
                                }

                            getGroupPropertyRequestByTaskType(
                                RequestTypeEnum.PROPERTIES_CREATE,
                                taskWithProperties,
                                requestElementPayload,
                                groupId
                            )
                        }
                        .flatMap { groupPayloadRequest ->
                            writeCamundaRequestsToDb(
                                taskWithProperties,
                                worker,
                                groupPayloadRequest,
                                RequestTypeEnum.PROPERTIES_CREATE
                            )
                        }
                        .doOnError {
                            Timber.e(it)
                        }
                )
            }

        return requestList
    }

    private fun writeCamundaRequestsToDb(
        taskWithProperties: TaskWithProperties,
        worker: Worker,
        taskPayloadRequest: BasePayload,
        requestTypeEnum: RequestTypeEnum
    ): Observable<Long> {
        return databaseRepository.queueCamundaRequest(
            type = requestTypeEnum.requestType,
            request = when {
                (taskPayloadRequest is SuperTaskRequest.PayloadRequest) -> {
                    getVariableForTaskUpdateRequest(
                        taskPayloadRequest,
                        worker
                    )
                }
                (taskPayloadRequest is GroupPropertyRequest.PayloadRequest) -> {
                    getVariableForGroupPropertyRequest(
                        taskPayloadRequest,
                        worker
                    )
                }
                else -> {
                    getVariableForPropertyRequest(
                        taskPayloadRequest as PropertyRequest.PayloadRequest,
                        worker
                    )
                }
            },
            namespace = appStateManager.getSessionId(),
            resourceId = taskWithProperties.id
        )
            .map { it }
            .toObservable()
    }
}