package ru.daotech.agroinvest.components

import android.content.Context
import ru.daotech.agroinvest.adapters.*
import ru.daotech.agroinvest.database.entities.TaskProperty
import ru.daotech.agroinvest.database.relation.TaskWithProperties
import ru.daotech.agroinvest.service.jsonmodels.TaskPropertyPayload
import ru.daotech.agroinvest.utils.*
import timber.log.Timber

open class TextViewComponentBuilder : ComponentBuilderView {

    lateinit var context: Context

    override fun buildPropertyViewModelList(taskWithProperties: TaskWithProperties): ArrayList<Any> {
        val propertyList = ArrayList<Any>()

        val startIndex = taskWithProperties.properties.indexOfFirst { taskProperty ->
            taskProperty.type == TASK_PROPERTY_HEADER_TYPE
        }
        for (i in 0 until startIndex) {
            propertyList.add(
                SimpleTaskPropertyModel(
                    buildTaskPropertyPayloadObj(taskWithProperties.properties[i])
                )
            )
        }
        for (i in startIndex until taskWithProperties.properties.size) {
            when (taskWithProperties.properties[i].type) {
                TASK_PROPERTY_TEXT_TYPE -> {
                    propertyList.add(
                        TextAreaModelView(
                            buildTaskPropertyPayloadObj(taskWithProperties.properties[i])
                        )
                    )
                }
                TASK_PROPERTY_NUMBER_TYPE -> {
                    propertyList.add(
                        NumericFieldModelView(
                            buildTaskPropertyPayloadObj(taskWithProperties.properties[i])
                        )
                    )
                }
                TASK_PROPERTY_IMAGE_TYPE -> {
                    propertyList.add(
                        ImageBinaryModelView(
                            buildTaskPropertyPayloadObj(taskWithProperties.properties[i]),
                            null,
                            null
                        )
                    )
                }
                TASK_PROPERTY_HEADER_TYPE -> {
                    propertyList.add(
                        HeaderModelView(
                            buildTaskPropertyPayloadObj(taskWithProperties.properties[i])
                        )
                    )
                }
                else -> {
                    Timber.i("Type of field: NOT DEFINED")
                }
            }
        }
        return propertyList
    }

    override fun buildGroupPropertyViewModelList(taskWithProperties: TaskWithProperties): ArrayList<Any> {
        // Build list with the last one unfilled properties - groupProperty
        val groupProperties = ArrayList<List<TaskProperty>>()
        taskWithProperties.properties
            .filter { it.type == TASK_PROPERTY_GROUP_TYPE }
            .forEach { groupPropertyList ->
                val propertyDistinct = taskWithProperties.properties
                    .filter { it.parentId == groupPropertyList.parentId }
                if (propertyDistinct.any { it.value.isBlank() }) {
                    groupProperties.add(propertyDistinct)
                    return@forEach
                }
            }
        val groupProperty = groupProperties.first()

        return buildRecycleViewListFromTask(
            groupProperty,
            ONLY_GROUP_BUILD_STATE
        )
    }

    override fun buildGroupPropertyViewModelListByBarcode(
        taskWithProperties: TaskWithProperties,
        barcodeValue: String
    ): ArrayList<Any> {
        // Build list with the last one unfilled properties - groupProperty
        val groupProperties = ArrayList<List<TaskProperty>>()
        taskWithProperties.properties
            .filter { it.type == TASK_PROPERTY_GROUP_TYPE }
            .forEach { taskProperty ->
                val propertyDistinct = taskWithProperties.properties
                    .filter { it.parentId == barcodeValue }
                if (propertyDistinct.any { it.value.isBlank() }) {
                    groupProperties.add(propertyDistinct)
                    return@forEach
                }
            }
        val groupProperty = groupProperties.first()

        return buildRecycleViewListFromTask(
            groupProperty,
            ONLY_GROUP_BUILD_STATE
        )
    }

    private fun buildRecycleViewListFromTask(
        groupProperty: List<TaskProperty>,
        buildState: Int
    ): ArrayList<Any> {
        val propertyList = ArrayList<Any>()

        for (i in groupProperty.indices) {
            when (groupProperty[i].type) {
                TASK_PROPERTY_TEXT_TYPE -> {
                    propertyList.add(
                        TextAreaModelView(
                            buildTaskPropertyPayloadObj(groupProperty[i])
                        )
                    )
                }
                TASK_PROPERTY_NUMBER_TYPE -> {
                    propertyList.add(
                        NumericFieldModelView(
                            buildTaskPropertyPayloadObj(groupProperty[i])
                        )
                    )
                }
                TASK_PROPERTY_IMAGE_TYPE -> {
                    propertyList.add(
                        ImageBinaryModelView(
                            buildTaskPropertyPayloadObj(groupProperty[i]),
                            null,
                            null
                        )
                    )
                }
                TASK_PROPERTY_HEADER_TYPE -> {
                    propertyList.add(
                        HeaderModelView(
                            buildTaskPropertyPayloadObj(groupProperty[i])
                        )
                    )
                }
                TASK_PROPERTY_GROUP_TYPE -> {
                    when (buildState) {
                        ONLY_HEADER_BUILD_STATE -> {

                        }
                        ONLY_GROUP_BUILD_STATE -> {
                            propertyList.add(
                                GroupModelView(
                                    buildTaskPropertyPayloadObj(groupProperty[i])
                                )
                            )
                        }
                    }
                }
                else -> {
                    Timber.i("Type of field: NOT DEFINED")
                }
            }
        }

        return propertyList
    }

    // Build without constructor.
    // The reason is wrong behaviour of Klaxon which
    // response handling of Task Property Payload with that constructor.
    private fun buildTaskPropertyPayloadObj(taskProperty: TaskProperty): TaskPropertyPayload {
        return TaskPropertyPayload(
            id = taskProperty.id,
            parentId = taskProperty.parentId,
            name = taskProperty.name,
            description = taskProperty.description,
            regexp = taskProperty.regexp,
            type = taskProperty.type,
            value = taskProperty.value,
            required = taskProperty.required,
            stage = taskProperty.stage,
            timestamp = taskProperty.timestamp,
            workerId = taskProperty.workerId
        )
    }

    companion object {
        const val ONLY_GROUP_BUILD_STATE = 2
        const val ONLY_HEADER_BUILD_STATE = 1
    }
}