package ru.daotech.agroinvest.container

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import com.rabbitmq.client.AlreadyClosedException
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.exceptions.UndeliverableException
import io.reactivex.functions.BiFunction
import io.reactivex.plugins.RxJavaPlugins
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.BehaviorSubject
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.BuildConfig
import ru.daotech.agroinvest.business.CountlyEventControllerImpl
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.exceptions.SignOutCamundaException
import ru.daotech.agroinvest.models.InternetConnectivityState
import ru.daotech.agroinvest.models.Worker
import timber.log.Timber
import java.io.IOException
import java.net.SocketException
import java.net.UnknownHostException
import java.util.concurrent.TimeUnit
import java.util.concurrent.TimeoutException
import javax.inject.Inject

class ContainerPresenter @Inject constructor(
    private var goToMyTasks: (() -> Unit)?,
    private var goToTaskHistory: (() -> Unit)?,
    private var goToCollectionHistory: (() -> Unit)?,
    private var goToSections: (() -> Unit)?,
    private var goToAccidentReport: (() -> Unit)?,
    private var goToAboutPage: (() -> Unit)?,
    private var goToUpdateApp: (() -> Unit)?,
    private var goToRatings: (() -> Unit)?,
    private var goToEarnings: (() -> Unit)?,
    private var exitApplication: (() -> Unit)?,
    private var goToSupportNotify: (() -> Unit)?,
    private val repository: Repository,
    private val appStateManager: AppStateManager,
    private val countlyEventController: CountlyEventControllerImpl
) : MviBasePresenter<ContainerView, ContainerViewState>() {

    private val compositeDisposable = CompositeDisposable()
    private val onUnreadSupportMessageCountSubject: BehaviorSubject<Int> = BehaviorSubject.create()

    override fun bindIntents() {

        val loadTasksDataFlowable = onUnreadSupportMessageCountSubject
            .map { unreadMessageCount ->
                ContainerPartialViewState.UpdateUnreadMessageCountMenu(
                    unreadMessageCount
                )
            }
            .cast(ContainerPartialViewState::class.java)

        val signalToListenUnreadMessageIntent = intent(ContainerView::onListenUnreadMessageIntent)
            .map { ContainerPartialViewState.ListenUnreadMessageCountState }
            .cast(ContainerPartialViewState::class.java)

        val loadingIntent = intent(ContainerView::loadingIntent)
            .map {
                // RxJavaPlugins - intercept all undelivered errors from streams
                RxJavaPlugins.setErrorHandler { e ->
                    var ex = e.cause
                    if (e is UndeliverableException) {
                        ex = e
                    }
                    if ((ex is IOException) || (ex is SocketException)) {
                        // fine, irrelevant
                        //
                        // problem or API that throws on cancellation
                        return@setErrorHandler
                    }
                    if (ex is InterruptedException) {
                        // fine, some blocking code was interrupted by a dispose call
                        return@setErrorHandler
                    }
                    if ((ex is NullPointerException) || (ex is IllegalArgumentException)) {
                        // that's likely a bug in the application
                        Thread.currentThread()
                            .uncaughtExceptionHandler.uncaughtException(Thread.currentThread(), e)
                        return@setErrorHandler
                    }
                    if (ex is IllegalStateException) {
                        // that's a bug in RxJava or in a custom operator
                        Thread.currentThread().uncaughtExceptionHandler
                            .uncaughtException(Thread.currentThread(), ex)
                        return@setErrorHandler
                    }
                    Timber.e("Undeliverable exception received, not sure what to do")
                }
            }
            .flatMap { appStateManager.getCurrentWorker() }
            .map { worker -> ContainerPartialViewState.InitialData(worker) }

        val onUploadBackupIntent = intent(ContainerView::onUploadBackupSignal)
            .flatMap { files -> repository.uploadDatabaseFileDump(files) }
            .map {
                ContainerPartialViewState.InitialData(appStateManager.getCurrentWorkerSimple()!!)
                        as ContainerPartialViewState
            }
            .onErrorReturn {
                Timber.e(it)
                ContainerPartialViewState.ShowSnackbarError(it.localizedMessage)
            }

        val forcedLogoutIntent = intent(ContainerView::onForcedLogoutIntent)
            .flatMap { appStateManager.getCurrentWorker() }
            .map { currentWorker -> appStateManager.markWorkerInfoNotSent(currentWorker) }
            .map { ContainerPartialViewState.FinishDay }

        intent(ContainerView::onSearchSubmitSubjectIntent)
            .map { rowNumber ->
                appStateManager.saveFilterRowNumber(rowNumber)
            }
            .subscribe()

        val forceLogoutByWorkerSessionIsOver = intent(ContainerView::exitApplicationIntent)
            .map {
                appStateManager.clearCurrentWorker()
                appStateManager.clearWorkSessionPref()
            }
            .map {
                ContainerPartialViewState.LogoutSuccess(
                    SignOutResult(
                        isSuccessResult = true,
                        message = "Рабочая сессия завершена по истечении времени."
                    )
                )
            }

        val onLoginInitIntent = intent(ContainerView::onLoginInit)
            .map {
                if (appStateManager.getCurrentWorkerSimple() != null) {
                    ContainerPartialViewState.GoingToMyTasks
                } else {
                    ContainerPartialViewState.LoginInit
                }
            }

        val onDismissSignOutResultIntent = intent(ContainerView::onDismissSignOutResult)
            .map { ContainerPartialViewState.DismissSignOutResult }

        val onSignOutAcceptIntent = intent(ContainerView::onSignOutAcceptIntent)
            .withLatestFrom(
                appStateManager.getCurrentWorker(),
                BiFunction<Unit, Worker, Worker> { _, worker -> worker })
            .throttleFirst(3, TimeUnit.SECONDS)
            .switchMap { worker -> logout(worker) }

        val internetConnectionChanged = intent(ContainerView::internetConnectionChanged)
            .map { connectivityState ->
                ContainerPartialViewState.InternetConnectionChanged(connectivityState.first)
                return@map connectivityState
            }
            .switchMap { connectionState ->
                return@switchMap synchronizeNotUploadedData(
                    connectionState.first,
                    connectionState.second
                )
            }

        // TODO: disable 156
//        val onSignOutFromQueueIntent = appStateManager.onSignOutQueueReceiveSubject
//            .flatMap {
//                logout(appStateManager.getCurrentWorkerSimple()!!)
//                    .startWith(ContainerPartialViewState.LoadInProgress)
//            }

        val screenshotSignalIntent = intent(ContainerView::screenshotSignalReceived)
            .flatMap { file -> repository.uploadScreenshotImage(file) }
            .map { ContainerPartialViewState.EmptyState as ContainerPartialViewState }
            .onExceptionResumeNext(
                Observable.just(
                    ContainerPartialViewState.EmptyState as ContainerPartialViewState
                )
            )

        val goToMyTaskIntent = intent(ContainerView::goToMyTasksIntent)
            .map { ContainerPartialViewState.GoingToMyTasks }

        val returnToMyTasksIntent = intent(ContainerView::returnToMyTasksIntent)
            .map { ContainerPartialViewState.GoingToMyTasks }

        val goToTasksHistoryIntent = intent(ContainerView::goToTasksHistoryIntent)
            .map { ContainerPartialViewState.GoingToTasksHistory }

        val goToCollectionHistoryIntent = intent(ContainerView::goToCollectionHistoryIntent)
            .map { ContainerPartialViewState.GoingToCollectionHistory }

        val goToSectionsIntent = intent(ContainerView::goToSectionsIntent)
            .map { ContainerPartialViewState.GoingToSections }

        val goToAccidentReportIntent = intent(ContainerView::goToAccidentReportIntent)
            .map { ContainerPartialViewState.GoingToAccidentReport }

        val goToAboutPageIntent = intent(ContainerView::goToAboutPageIntent)
            .map { ContainerPartialViewState.GoingToAboutPage }

        val goToUpdateAppIntent = intent(ContainerView::goToUpdateAppIntent)
            .map { ContainerPartialViewState.GoingToUpdateApp }

        val goToRatingsIntent = intent(ContainerView::goToRatingsIntent)
            .map { ContainerPartialViewState.GoingToRatings }

        val goToEarningsIntent = intent(ContainerView::goToEarningsIntent)
            .map { ContainerPartialViewState.GoingToEarnings }

        val goToSupportMessageIntent = intent(ContainerView::goToSupportMessageIntent)
            .map { ContainerPartialViewState.GoingToSupportNotify }

        val finishDayIntent = intent(ContainerView::finishDayIntent)
            .map { ContainerPartialViewState.FinishDay }

        val invalidateToolbar = intent(ContainerView::toolbarVisibilityChanged)
            .map { ContainerPartialViewState.InvalidateToolbar }

        val broadcastMessageIntent = repository.getBroadcastMessageSubject()
            .map { message -> ContainerPartialViewState.ShowBroadcastMessage(message) }

//        val supportMessageIntent = repository.getSupportMessageSubject()
//            .map { message -> ContainerPartialViewState.ShowSupportMessage(message) }

        val screenshotReceivedIntent = repository.getScreenshotMessageSubject()
            .map { ContainerPartialViewState.EmptyState }

        val serverDataIntent = repository.getServerDataSubject()
            .map { serverData ->
                var isVersionIncrease = false
                if (serverData.serverDataPayload.versionCode != BuildConfig.VERSION_CODE
                    && serverData.serverDataPayload.versionCode > BuildConfig.VERSION_CODE
                ) {
                    isVersionIncrease = true
                }

                ContainerPartialViewState.ShowUpdateAppVersionMenuItem(isVersionIncrease)
            }

        val dismissSnackbarIntent = intent(ContainerView::onDismissSnackbarIntent)
            .map { ContainerPartialViewState.DismissSnackbar }

        val allIntentList = listOf(
            onLoginInitIntent,
            invalidateToolbar,
            internetConnectionChanged,
            loadingIntent,
            goToMyTaskIntent,
            goToTasksHistoryIntent,
            goToSectionsIntent,
            goToAccidentReportIntent,
            goToAboutPageIntent,
            goToUpdateAppIntent,
            goToRatingsIntent,
            goToEarningsIntent,
            goToCollectionHistoryIntent,
            goToSupportMessageIntent,
            finishDayIntent,
            onSignOutAcceptIntent,
            broadcastMessageIntent,
            dismissSnackbarIntent,
            returnToMyTasksIntent,
            onDismissSignOutResultIntent,
            serverDataIntent,
            forcedLogoutIntent,
            forceLogoutByWorkerSessionIsOver,
//            onSignOutFromQueueIntent,
//            supportMessageIntent,
            loadTasksDataFlowable,
            signalToListenUnreadMessageIntent,
            screenshotSignalIntent,
            screenshotReceivedIntent,
            onUploadBackupIntent
        )

        val allIntent = Observable.merge(allIntentList)
            .onErrorResumeNext { error: Throwable ->
                errorProcessing(error)
            }
            .share()

        compositeDisposable.add(
            allIntent.filter { state ->
                state is ContainerPartialViewState.ListenUnreadMessageCountState
            }.subscribe(
                {
                    startListenUnreadMessageCount()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        compositeDisposable.add(
            allIntent.filter { state ->
                state is ContainerPartialViewState.GoingToMyTasks
            }.subscribe(
                {
                    goToMyTasks?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        compositeDisposable.add(
            allIntent.filter { state ->
                state is ContainerPartialViewState.GoingToTasksHistory
            }.subscribe(
                {
                    goToTaskHistory?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        compositeDisposable.add(
            allIntent.filter { state ->
                state is ContainerPartialViewState.GoingToCollectionHistory
            }.subscribe(
                {
                    goToCollectionHistory?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        compositeDisposable.add(
            allIntent.filter { state ->
                state is ContainerPartialViewState.GoingToSections
            }.subscribe(
                {
                    goToSections?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        compositeDisposable.add(
            allIntent.filter { state ->
                state is ContainerPartialViewState.GoingToAccidentReport
            }.subscribe(
                {
                    goToAccidentReport?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        compositeDisposable.add(
            allIntent.filter { state ->
                state is ContainerPartialViewState.GoingToAboutPage
            }.subscribe(
                {
                    goToAboutPage?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        compositeDisposable.add(
            allIntent.filter { state ->
                state is ContainerPartialViewState.GoingToUpdateApp
            }.subscribe(
                {
                    goToUpdateApp?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        compositeDisposable.add(
            allIntent.filter { state ->
                state is ContainerPartialViewState.GoingToRatings
            }.subscribe(
                {
                    goToRatings?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        compositeDisposable.add(
            allIntent.filter { state ->
                state is ContainerPartialViewState.GoingToEarnings
            }.subscribe(
                {
                    goToEarnings?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        compositeDisposable.add(
            allIntent.filter { state ->
                state is ContainerPartialViewState.GoingToSupportNotify
            }.subscribe(
                {
                    goToSupportNotify?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        compositeDisposable.add(
            allIntent.filter { state ->
                state is ContainerPartialViewState.LogoutSuccess
            }.subscribe(
                {
                    exitApplication?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        val initialState = ContainerViewState(
            name = "",
            personalNumber = "",
            isFinishedDayDialogOpen = false,
            signOutResult = null,
            broadcastMessage = null,
            isLoadInProgress = false,
            isInternetConnected = true,
            isSnackBarNeedShowing = false,
            taskId = null,
            isVersionIncrease = false,
            isQueueConnected = true,
            isSocketConnected = true
        )

        val stateObservable = allIntent.scan(initialState, { viewState, partialViewState ->
            partialViewState.reduce(viewState)
        })

        subscribeViewState(stateObservable, ContainerView::render)
    }

    private fun errorProcessing(error: Throwable): Observable<ContainerPartialViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .take(1)
            .map { ContainerPartialViewState.EmptyState as ContainerPartialViewState }
            .startWith(ContainerPartialViewState.ShowSnackbarError(errorMessage = error.localizedMessage))
    }

    private fun startListenUnreadMessageCount() {
        compositeDisposable.add(
            repository.getUnreadSupportMessageCount()
                .observeOn(AndroidSchedulers.mainThread())
                .subscribe { unreadMessageCount ->
                    onUnreadSupportMessageCountSubject.onNext(unreadMessageCount)
                }
        )
    }

    private fun synchronizeNotUploadedData(
        connectionState: Boolean,
        internetConnectivityState: InternetConnectivityState
    ): Observable<ContainerPartialViewState> {
        return when (connectionState) {
            true -> {
                Observable.just(Unit)
                    .map {
                        ContainerPartialViewState.UploadDataSuccess as ContainerPartialViewState
                    }
                    .map {
                        countlyEventController.sendCountlyEvent(
                            COUNTLY_EVENT_CONNECTIVITY_STATE_ON,
                            internetConnectivityState
                        )
                    }
                    .map {
                        // Open new socket connection
                        appStateManager.onSocketStateRestart.onNext(Unit)
                    }
                    .switchMap {
                        Observable.just(
                            ContainerPartialViewState.InternetConnectionChanged(
                                connectionState
                            ) as ContainerPartialViewState
                        )
                    }
            }
            false -> {
                countlyEventController.sendCountlyEvent(
                    COUNTLY_EVENT_CONNECTIVITY_STATE_OFF,
                    internetConnectivityState
                )

                Observable.just(
                    ContainerPartialViewState.InternetConnectionChanged(connectionState)
                            as ContainerPartialViewState
                )
            }
        }
    }

    private fun logout(worker: Worker): Observable<ContainerPartialViewState> {
        if (!appStateManager.getCurrentTrolleyNumber().isBlank()) {
            return Observable.just(
                ContainerPartialViewState.LogoutError(
                    SignOutResult(
                        isSuccessResult = false,
                        message = "Есть тележка в работе. " +
                                "Пожалуйста отвяжите тележку перед завершением рабочего дня."
                    )
                )
            )
        }
        if (!appStateManager.connectionReceiver.isConnected) {
            return Observable.just(
                ContainerPartialViewState.LogoutError(
                    SignOutResult(
                        isSuccessResult = false,
                        message = "Отсутствует интернет соединение.\n" +
                                "Пожалуйста, проверьте сеть."
                    )
                )
            )
        }
        return repository.signOut(worker.id)
            .observeOn(AndroidSchedulers.mainThread())
            .map {
                if (it) {
                    // Clear worker info from local storage
                    appStateManager.clearWorkSessionPref()
                    appStateManager.clearCurrentWorker()

                    // Disconnect from Couch database instance
                    appStateManager.onCloseCouchDatabaseConnection.onNext(Unit)

                    exitApplication?.invoke()

                    ContainerPartialViewState.LogoutSuccess(
                        SignOutResult(
                            isSuccessResult = true,
                            message = "Рабочий день успешно завершен"
                        )
                    ) as ContainerPartialViewState
                } else {

                    ContainerPartialViewState.LogoutError(
                        SignOutResult(
                            isSuccessResult = false,
                            message = "Произошла ошибка во время завершения рабочего дня. Повторите попытку."
                        )
                    ) as ContainerPartialViewState
                }
            }
            .startWith(ContainerPartialViewState.LoadInProgress)
            .onErrorResumeNext { error: Throwable ->
                Timber.e(error)

                // If worker info has not sent to server,
                // Need to make mark into preferences
                appStateManager.markWorkerInfoNotSent(worker)

                // First we have to forced reconnect queue for stabling connection check
//                appStateManager.onReconnectQueueReceiveSubject.onNext(Unit)

                return@onErrorResumeNext Observable.just(managePartialErrorState(error))
            }
    }

    private fun managePartialErrorState(error: Throwable): ContainerPartialViewState.LogoutError {
        return when (error) {
            is SignOutCamundaException -> {
                ContainerPartialViewState
                    .LogoutError(
                        SignOutResult(
                            isSuccessResult = false,
                            message = error.localizedMessage
                        )
                    )
            }
            is AlreadyClosedException -> {
                ContainerPartialViewState
                    .LogoutError(
                        SignOutResult(
                            isSuccessResult = true,
                            message = "Произошла ошибка, повторите попытку. Код: [AYCD]"
                        )
                    )
            }
            is TimeoutException -> {
                ContainerPartialViewState
                    .LogoutError(
                        SignOutResult(
                            isSuccessResult = true,
                            message = "Произошла ошибка, повторите попытку. Код: [TIUT]"
                        )
                    )
            }
            is UnknownHostException ->
                ContainerPartialViewState
                    .LogoutError(
                        SignOutResult(
                            isSuccessResult = true,
                            message = "Отсутствует интернет соединение. " +
                                    "Проверьте подключение и повторите попытку."
                        )
                    )
            else ->
                ContainerPartialViewState
                    .LogoutError(
                        SignOutResult(
                            isSuccessResult = false,
                            message = "Ошибка во время отправки данных на сервер." +
                                    "Проверьте подключение и повторите попытку."
                        )
                    )
        }
    }

    override fun unbindIntents() {
        super.unbindIntents()
//        compositeDisposable.dispose()

        goToMyTasks = null
        goToTaskHistory = null
        goToCollectionHistory = null
        goToSections = null
        goToAccidentReport = null
        goToAboutPage = null
        goToUpdateApp = null
        goToRatings = null
        goToEarnings = null
        exitApplication = null
    }

    companion object {
        const val COUNTLY_EVENT_CONNECTIVITY_STATE_ON = "CONNECTIVITY_STATE_ON"
        const val COUNTLY_EVENT_CONNECTIVITY_STATE_OFF = "CONNECTIVITY_STATE_OFF"
    }
}
