package ru.daotech.agroinvest.database

import com.beust.klaxon.Klaxon
import io.reactivex.Completable
import io.reactivex.Flowable
import io.reactivex.Observable
import io.reactivex.Single
import io.reactivex.schedulers.Schedulers
import ru.daotech.agroinvest.database.entities.*
import ru.daotech.agroinvest.database.relation.TaskWithPropertiesImpl
import ru.daotech.agroinvest.models.RequestTypeEnum.*
import ru.daotech.agroinvest.network.models.VariableRequest
import ru.daotech.agroinvest.service.jsonmodels.SupportMessagePayload
import timber.log.Timber
import javax.inject.Inject

class DatabaseRepositoryImpl @Inject constructor(
    private val appDatabase: AppDatabase
) : DatabaseRepository {

    override fun updateTaskInterval(updatedTaskInterval: TaskIntervalImpl): Completable {
        return Completable.fromAction {
            appDatabase.getTaskIntervalDao()
                .update(updatedTaskInterval)
        }
            .subscribeOn(Schedulers.io())
    }

    override fun insertTaskInterval(taskInterval: TaskIntervalImpl): Completable {
        return appDatabase.getTaskIntervalDao()
            .insertTaskInterval(taskInterval)
            .subscribeOn(Schedulers.io())
    }

    override fun upsertTaskPropertiesList(taskPropertyList: List<TaskPropertyImpl>): Completable {
        return Completable.fromAction {
            appDatabase.getTaskPropertyDao().upsert(taskPropertyList)
        }
            .subscribeOn(Schedulers.io())
    }

    override fun getMyTasksByRowNumber(
        namespace: String,
        workerId: String,
        rowNumber: String
    ): Observable<List<TaskImpl>> {
        return appDatabase.getTaskDao().queryTaskListByRowNumber(
            namespace = namespace,
            workerId = workerId,
            rowNumber = rowNumber
        )
            .subscribeOn(Schedulers.io())
    }

    override fun getMyStartedTasks(
        namespace: String,
        workerId: String
    ): Observable<List<TaskImpl>> {
        return appDatabase.getTaskDao()
            .queryTaskStartedList(namespace = namespace, workerId = workerId)
            .subscribeOn(Schedulers.io())
    }

    override fun getMyPausedTasks(
        namespace: String,
        workerId: String
    ): Observable<List<TaskImpl>> {
        return appDatabase.getTaskDao()
            .queryTaskPausedList(namespace = namespace, workerId = workerId)
            .subscribeOn(Schedulers.io())
    }

    override fun getMyTasksFlowable(
        namespace: String,
        workerId: String,
        rowNumber: String?
    ): Flowable<List<TaskImpl>> {
        return when (rowNumber.isNullOrEmpty()) {
            true -> appDatabase.getTaskDao().queryTaskListFlowable(
                namespace = namespace,
                workerId = workerId
            )
                .subscribeOn(Schedulers.io())
            else -> appDatabase.getTaskDao().queryTaskListFilteredFlowable(
                namespace = namespace,
                workerId = workerId,
                rowNumber = rowNumber
            )
                .subscribeOn(Schedulers.io())
        }
    }

    override fun getTasksHistory(
        namespace: String,
        workerId: String
    ): Observable<List<TaskWithPropertiesImpl>> {
        return appDatabase.getTaskDao()
            .queryTaskHistoryList(namespace = namespace, workerId = workerId)
            .subscribeOn(Schedulers.io())
    }

    override fun getRootSections(namespace: String): Observable<List<SectionImpl>> {
        return appDatabase.getSectionDao().queryRootSections(namespace)
            .subscribeOn(Schedulers.io())
    }

    override fun getSectionParent(
        parentId: String?,
        namespace: String
    ): Observable<List<SectionImpl>> {
        return appDatabase.getSectionDao().querySectionParent(parentId, namespace)
            .subscribeOn(Schedulers.io())
    }

    override fun getSection(sectionId: String): Single<SectionImpl> {
        return appDatabase.getSectionDao().querySection(sectionId)
            .subscribeOn(Schedulers.io())
    }

    override fun getSectionTasks(
        sectionId: String,
        namespace: String
    ): Observable<List<TaskWithPropertiesImpl>> {
        return appDatabase.getTaskDao().querySectionTasksList(sectionId, namespace)
            .subscribeOn(Schedulers.io())
    }

    override fun getTaskWithProperties(
        taskId: String,
        sessionId: String
    ): Single<TaskWithPropertiesImpl> {
        return appDatabase.getTaskDao().queryTaskWithProperties(taskId, sessionId)
            .subscribeOn(Schedulers.io())
    }

    override fun upsertTask(task: TaskImpl): Completable {
        return Completable.fromAction {
            appDatabase.getTaskDao().upsert(task)
            Timber.d(Thread.currentThread().toString())
        }
            .subscribeOn(Schedulers.io())
    }

    override fun upsertTaskIntervalsList(taskIntervalsList: List<TaskIntervalImpl>): Completable {
        return Completable.fromAction { appDatabase.getTaskIntervalDao().upsert(taskIntervalsList) }
            .subscribeOn(Schedulers.io())
    }

    override fun upsertTaskMeasurementsList(
        taskMeasurementsList: List<TaskMeasurementImpl>
    ): Completable {
        return Completable.fromAction {
            appDatabase.getTaskMeasurementDao().upsert(taskMeasurementsList)
        }
            .subscribeOn(Schedulers.io())
    }

    override fun upsertTaskMeasurement(taskMeasurement: TaskMeasurementImpl): Completable {
        return Completable.fromAction {
            appDatabase.getTaskMeasurementDao().upsert(taskMeasurement)
        }
            .subscribeOn(Schedulers.io())
    }

    override fun upsertTaskSectionsList(taskSectionsList: List<SectionImpl>): Completable {
        return Completable.fromAction { appDatabase.getSectionDao().upsert(taskSectionsList) }
            .subscribeOn(Schedulers.io())
    }

    override fun getNotUploadedTaskWithProperties(taskId: String): Observable<List<TaskWithPropertiesImpl>> {
        return appDatabase.getTaskDao().queryNotUploadedTaskWithProperties(taskId)
            .subscribeOn(Schedulers.io())
    }

    override fun getFullTrolleyInfoHistoryList(
        workerId: String,
        namespace: String
    ): Observable<List<TrolleyInfoImpl>> {
        return appDatabase.getTrolleyInfo().getFullTrolleyHistory(workerId, namespace)
            .subscribeOn(Schedulers.io())
    }

    override fun getSupportMessageCount(workerId: String): Flowable<Int> {
        return appDatabase.getSupportMessageDao().getSupportMessageCount(workerId)
            .subscribeOn(Schedulers.io())
    }

    override fun getAllSupportMessagesByWorkerId(workerId: String): Flowable<List<SupportMessageImpl>> {
        return appDatabase.getSupportMessageDao().getAllSupportMessagesByWorkerId(workerId)
            .subscribeOn(Schedulers.io())
    }

    override fun updateUnreadStatusSupportMessages(supportMessageList: List<SupportMessageImpl>): Completable {
        return Completable.fromAction {
            appDatabase.getSupportMessageDao().upsert(supportMessageList)
        }
            .subscribeOn(Schedulers.io())
    }

    override fun queueCamundaRequest(
        type: String,
        request: VariableRequest,
        namespace: String,
        resourceId: String?
    ): Single<Long> {
        val camunda = CamundaImpl(
            type = type,
            payload = Klaxon().toJsonString(request),
            namespace = namespace,
            resourceId = resourceId
        )

        return appDatabase.getCamundaDao()
            .insertCamunda(camunda)
            .map { id ->
                when {
                    setOf(
                        SIGN_IN.requestType,
                        TASK_LIST.requestType,
                        SIGN_OUT.requestType,
                        TASK_UPDATE.requestType
                    ).contains(type) -> {
                        appDatabase.getCamundaDao()
                            .removeNotSentDuplicates(id = id, type = type, resourceId = resourceId)
                            .subscribeOn(Schedulers.io())
                            .observeOn(Schedulers.single())
                            .subscribe()
                    }
                }
                id
            }
            .subscribeOn(Schedulers.io())
    }

    override fun getTaskUpdateRequestByTaskId(taskId: String): Observable<List<CamundaImpl>> {
        return appDatabase.getCamundaDao().getTaskUpdateRequestByTaskId(taskId)
            .take(1)
            .subscribeOn(Schedulers.io())
    }

    override fun getNotUploadedCamundaRequests(): Observable<List<CamundaImpl>> {
        return appDatabase.getCamundaDao().queryNotSentRequests()
            .take(1)
            .subscribeOn(Schedulers.io())
    }

    override fun getNotUploadedCamundaRequestsWOTaskListCount(): Observable<Int> {
        return appDatabase.getCamundaDao().queryNotSentRequestsWOTaskListCount()
            .subscribeOn(Schedulers.io())
    }

    override fun upsertCamundaRequest(camundaImpl: CamundaImpl): Completable {
        return Completable.fromAction { appDatabase.getCamundaDao().upsert(camundaImpl) }
            .subscribeOn(Schedulers.io())
    }

    override fun getTaskCount(
        namespace: String,
        workerId: String,
        startPeriod: String,
        finishPeriod: String
    ): Observable<Int> {
        return appDatabase.getTaskDao().queryTaskCount(
            namespace = namespace,
            workerId = workerId,
            startPeriod = startPeriod,
            finishPeriod = finishPeriod
        )
            .subscribeOn(Schedulers.io())
    }

    override fun getTaskExistence(
        namespace: String,
        taskId: String,
        workerId: String
    ): Observable<Int> {
        return appDatabase.getTaskDao()
            .queryTaskExistence(namespace = namespace, taskId = taskId, workerId = workerId)
            .subscribeOn(Schedulers.io())
    }

    override fun upsertTaskPlantsList(taskPlantList: List<TaskPlantImpl>): Completable {
        return Completable.fromAction {
            appDatabase.getTaskPlantDao().upsert(taskPlantList)
        }
            .subscribeOn(Schedulers.io())
    }

    override fun insertSupportMessage(
        supportMessage: SupportMessagePayload,
        workerId: String,
        time: String,
        messageCode: Int
    ): Completable {
        return Completable.fromAction {
            appDatabase.getSupportMessageDao()
                .insert(
                    SupportMessageImpl(
                        supportMessage.message,
                        workerId,
                        time,
                        messageCode
                    )
                )
        }
            .subscribeOn(Schedulers.io())
    }

    override fun deleteAllMessages(): Completable {
        return Completable.fromAction {
            appDatabase.getSupportMessageDao().deleteAllMessages()
            Timber.d(Thread.currentThread().toString())
        }
            .subscribeOn(Schedulers.io())
    }

    override fun upsertTrolleyInfo(trolleyInfoImpl: TrolleyInfoImpl): Completable {
        return Completable.fromAction {
            appDatabase.getTrolleyInfo().upsert(trolleyInfoImpl)
            Timber.d(Thread.currentThread().toString())
        }
            .subscribeOn(Schedulers.io())
    }

    override fun setTaskWorker(
        namespace: String,
        taskId: String,
        workerId: String?,
        workerName: String?,
        inProgress: Boolean,
        orderBy: Int
    ): Completable {
        return Completable.fromAction {
            appDatabase.getTaskDao().setTaskWorker(
                namespace = namespace,
                taskId = taskId,
                workerId = workerId,
                workerName = workerName,
                inProgress = inProgress,
                orderBy = orderBy
            ).subscribe()
        }
            .subscribeOn(Schedulers.io())
    }

    override fun clearTasksData(namespace: String, workerId: String): Completable {
        return Completable.fromAction {
            appDatabase.getTaskDao().clearAllTasksData().subscribe()
        }
            .subscribeOn(Schedulers.io())
    }

    override fun clearAllCamundaRequests(): Observable<Unit> {
        return appDatabase.getCamundaDao().getAllCamundaData()
            .doOnNext { camundaList ->
                appDatabase.getCamundaDao()
                    .clearAllCamundaRequestData(camundaList)
                    .subscribe()
            }
            .map { Unit }
    }

    override fun clearAllSections(): Observable<Unit> {
        return appDatabase.getSectionDao().getAllSectionsData()
            .doOnNext { sectionList ->
                appDatabase.getSectionDao()
                    .clearAllSections(sectionList)
                    .subscribe()
            }
            .map { Unit }
    }

    override fun clearAllIntervals(): Observable<Unit> {
        return appDatabase.getTaskIntervalDao().getAllTaskIntervalData()
            .doOnNext { taskIntervalList ->
                appDatabase.getTaskIntervalDao()
                    .clearAllTaskIntervalData(taskIntervalList)
                    .subscribe()
            }
            .map { Unit }
    }

    override fun clearAllProperties(): Observable<Unit> {
        return appDatabase.getTaskPropertyDao().getAllTaskPropertyData()
            .doOnNext { propertyList ->
                appDatabase.getTaskPropertyDao()
                    .clearAllTaskPropertyData(propertyList)
                    .subscribe()
            }
            .map { Unit }
    }

    override fun getRecentUsers(): Observable<List<RecentUserImpl>> {
        return appDatabase.getRecentUserDao().getMostRecentUsers()
    }

    override fun upsertRecentUser(user: RecentUserImpl) {
        appDatabase.getRecentUserDao().upsert(user)
    }
}
