package ru.daotech.agroinvest.database.entities

import androidx.room.*
import com.beust.klaxon.Klaxon
import ru.daotech.agroinvest.database.relation.TaskWithProperties
import ru.daotech.agroinvest.models.Department
import ru.daotech.agroinvest.service.jsonmodels.Position
import ru.daotech.agroinvest.service.jsonmodels.SectionPayload
import ru.daotech.agroinvest.service.jsonmodels.TaskPayload
import java.io.StringReader

interface Task {
    val id: String
    val namespace: String
    val name: String
    val createdAt: String?
    val updatedAt: String?
    val requireFinish: Boolean?
    val isFinished: Boolean
    val startedAt: String?
    val finishedAt: String?
    val requireOnlineStart: Boolean?
    val orderBy: Int
    val color: String
    val description: String
    val position: Position?
    val department: Department?
    val section: SectionPayload?
    val startAt: String
    val finishAt: String
    val parentId: String?
    val workerId: String?
    val workerName: String?
    val sectionId: String?
    val isTaskInfoSent: Boolean
    val rating: Float
    val rowNumber: Int?
    val inProgress: Boolean
    val isHarvesting: Boolean?
    val taskLimit: Int? // minutes
}

class TaskConverter {
    @TypeConverter
    fun fromPosition(position: Position?): String? {
        return when (position) {
            null -> null
            else -> Klaxon().toJsonString(position)
        }
    }

    @TypeConverter
    fun toPosition(data: String?): Position? {
        return when (data) {
            null -> null
            else -> Klaxon().parseFromJsonObject<Position>(
                Klaxon().parseJsonObject(
                    StringReader(
                        data
                    )
                )
            )
        }
    }

    @TypeConverter
    fun fromDepartment(department: Department?): String? {
        return when (department) {
            null -> null
            else -> Klaxon().toJsonString(department)
        }
    }

    @TypeConverter
    fun toDepartment(data: String?): Department? {
        return when (data) {
            null -> null
            else -> Klaxon().parseFromJsonObject<Department>(
                Klaxon().parseJsonObject(
                    StringReader(
                        data
                    )
                )
            )
        }
    }

    @TypeConverter
    fun fromSection(section: SectionPayload?): String? {
        return when (section) {
            null -> null
            else -> Klaxon().toJsonString(section)
        }
    }

    @TypeConverter
    fun toSection(data: String?): SectionPayload? {
        return when (data) {
            null -> null
            else -> Klaxon().parseFromJsonObject<SectionPayload>(
                Klaxon().parseJsonObject(
                    StringReader(data)
                )
            )
        }
    }
}

@Entity(tableName = "tasks")
@TypeConverters(TaskConverter::class)
data class TaskImpl(
    @PrimaryKey
    @ColumnInfo(name = "id") override val id: String,

    @ColumnInfo(name = "namespace") override val namespace: String,

    @ColumnInfo(name = "name") override val name: String,

    @ColumnInfo(name = "isFinished") override val isFinished: Boolean,

    @ColumnInfo(name = "orderBy") override val orderBy: Int,

    @ColumnInfo(name = "color") override val color: String,

    @ColumnInfo(name = "description") override val description: String,

    @ColumnInfo(name = "position") override val position: Position?,

    @ColumnInfo(name = "department") override val department: Department?,

    @ColumnInfo(name = "section") override val section: SectionPayload?,

    @ColumnInfo(name = "startAt") override val startAt: String,

    @ColumnInfo(name = "finishAt") override val finishAt: String,

    @ColumnInfo(name = "createdAt") override val createdAt: String?,

    @ColumnInfo(name = "updatedAt") override val updatedAt: String?,

    @ColumnInfo(name = "requireFinish") override val requireFinish: Boolean?,

    @ColumnInfo(name = "startedAt") override val startedAt: String?,

    @ColumnInfo(name = "finishedAt") override val finishedAt: String?,

    @ColumnInfo(name = "requireOnlineStart") override val requireOnlineStart: Boolean?,

    @ColumnInfo(name = "parentId") override val parentId: String?,

    @ColumnInfo(name = "workerId") override val workerId: String?,

    @ColumnInfo(name = "workerName") override val workerName: String?,

    @ColumnInfo(name = "sectionId") override val sectionId: String?,

    @ColumnInfo(name = "isTaskInfoSent") override val isTaskInfoSent: Boolean = true,

    @ColumnInfo(name = "rating") override val rating: Float,

    @ColumnInfo(name = "rowNumber") override val rowNumber: Int?,

    @ColumnInfo(name = "inProgress") override val inProgress: Boolean,

    @ColumnInfo(name = "isHarvesting") override val isHarvesting: Boolean?,

    @ColumnInfo(name = "taskLimit") override val taskLimit: Int?
) : Task {
    constructor(from: TaskPayload, sessionId: String) : this(
        id = from.id,
        name = from.name,
        isFinished = from.isFinished,
        orderBy = from.orderBy,
        color = from.color,
        position = from.position,
        department = from.department,
        section = from.section,
        description = from.description,
        startAt = from.startAt,
        finishAt = from.finishAt,
        workerId = from.workerId,
        workerName = from.workerName,
        finishedAt = from.finishedAt,
        startedAt = from.startedAt,
        createdAt = null,
        sectionId = from.sectionId,
        namespace = sessionId,
        parentId = from.parentId,
        requireFinish = null,
        requireOnlineStart = null,
        updatedAt = null,
        rating = from.rating,
        rowNumber = from.rowNumber,
        inProgress = from.inProgress,
        isHarvesting = from.isHarvesting,
        taskLimit = from.taskLimit
    )

    constructor(
        from: TaskWithProperties,
        sessionId: String,
        workerId: String,
        startedAt: String?,
        finishedAt: String?,
        isTaskInfoSent: Boolean,
        isFinished: Boolean
    ) : this(
        id = from.id,
        name = from.name,
        isFinished = isFinished,
        orderBy = from.orderBy,
        color = from.color,
        position = from.position,
        department = from.department,
        section = from.section,
        description = from.description,
        startAt = from.startAt,
        finishAt = from.finishAt,
        workerId = workerId,
        finishedAt = finishedAt,
        startedAt = startedAt,
        workerName = from.workerName,
        createdAt = from.createdAt,
        sectionId = from.sectionId,
        namespace = sessionId,
        parentId = from.parentId,
        requireFinish = from.requireFinish,
        requireOnlineStart = from.requireOnlineStart,
        updatedAt = from.updatedAt,
        isTaskInfoSent = isTaskInfoSent,
        rating = from.rating,
        rowNumber = from.rowNumber,
        inProgress = from.inProgress,
        isHarvesting = from.isHarvesting,
        taskLimit = from.taskLimit
    )

    constructor(from: TaskWithProperties, isTaskInfoSent: Boolean) : this(
        id = from.id,
        name = from.name,
        isFinished = from.isFinished,
        orderBy = from.orderBy,
        color = from.color,
        position = from.position,
        department = from.department,
        section = from.section,
        description = from.description,
        startAt = from.startAt,
        finishAt = from.finishAt,
        workerId = from.workerId,
        workerName = from.workerName,
        finishedAt = from.finishedAt,
        startedAt = from.startedAt,
        createdAt = from.createdAt,
        sectionId = from.sectionId,
        namespace = from.namespace,
        parentId = from.parentId,
        requireFinish = from.requireFinish,
        requireOnlineStart = from.requireOnlineStart,
        updatedAt = from.updatedAt,
        isTaskInfoSent = isTaskInfoSent,
        rating = from.rating,
        rowNumber = from.rowNumber,
        inProgress = from.inProgress,
        isHarvesting = from.isHarvesting,
        taskLimit = from.taskLimit
    )
}
