package ru.daotech.agroinvest.fragments.barcode

import android.content.Context
import android.graphics.Matrix
import android.graphics.SurfaceTexture
import android.os.Bundle
import android.view.LayoutInflater
import android.view.Surface
import android.view.View
import android.view.ViewGroup
import android.widget.Toast
import androidx.camera.core.*
import androidx.lifecycle.LifecycleOwner
import com.hannesdorfmann.mosby3.mvi.MviFragment
import com.sembozdemir.permissionskt.askPermissions
import dagger.android.support.AndroidSupportInjection
import io.reactivex.Observable
import io.reactivex.subjects.PublishSubject
import kotlinx.android.synthetic.main.fragment_barcode_scanner.*
import ru.daotech.agroinvest.BarcodeAnalyzer
import ru.daotech.agroinvest.R
import java.util.concurrent.Executors
import javax.inject.Inject

class BarcodeScannerFragment : MviFragment<BarcodeScannerView, BarcodeScannerPresenter>(),
    BarcodeScannerView, LifecycleOwner {

    @Inject
    lateinit var barcodeScannerPresenter: BarcodeScannerPresenter

    private val scanningResult: PublishSubject<String> = PublishSubject.create()

    private val executor = Executors.newSingleThreadExecutor()

    override fun onBarcodeRetrieved(): Observable<String> = scanningResult

    override fun render(state: BarcodeScannerViewState) {}

    override fun createPresenter(): BarcodeScannerPresenter {
        return barcodeScannerPresenter
    }

    override fun onAttach(context: Context) {
        AndroidSupportInjection.inject(this)
        super.onAttach(context)
    }

    override fun onCreateView(
        inflater: LayoutInflater,
        container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View? {
        return inflater.inflate(R.layout.fragment_barcode_scanner, container, false)
    }

    override fun onResume() {
        super.onResume()
        requestPermissionAndStartCamera()
    }

    private fun requestPermissionAndStartCamera() {
        activity?.askPermissions(android.Manifest.permission.CAMERA) {
            onGranted {
                view_finder.post { startScanning() }
            }

            onDenied {
                Toast.makeText(
                    context,
                    resources.getString(R.string.not_grant_camera_permission),
                    Toast.LENGTH_LONG
                ).show()
            }
        }
    }

    private fun startScanning() {
        val previewConfig = PreviewConfig.Builder().build()

        val preview = Preview(previewConfig)

        preview.setOnPreviewOutputUpdateListener {
            updateViewSurfaceTexture(it.surfaceTexture)
        }

        val analyzerConfig = ImageAnalysisConfig.Builder().apply {
            setImageReaderMode(
                ImageAnalysis.ImageReaderMode.ACQUIRE_LATEST_IMAGE
            )
        }.build()

        // Build the image analysis use case and instantiate our analyzer
        val analyzerUseCase = ImageAnalysis(analyzerConfig).apply {
            setAnalyzer(executor, BarcodeAnalyzer { retrievedBarcode ->
                scanningResult.onNext(retrievedBarcode)
                removeAnalyzer()
            })
        }
        val imageCaptureConfig = ImageCaptureConfig.Builder().build()

        // Build the image capture use case and attach button click listener
        val imageCapture = ImageCapture(imageCaptureConfig)

        CameraX.bindToLifecycle(this, preview, analyzerUseCase, imageCapture)
    }

    private fun updateViewSurfaceTexture(surfaceTexture: SurfaceTexture) {

        if (view_finder != null) {
            (view_finder.parent as ViewGroup).apply {
                removeView(view_finder)
                addView(view_finder, VIEW_INDEX)
            }
            view_finder.surfaceTexture = surfaceTexture
            updateTransform()
        }
    }

    private fun updateTransform() {
        val matrix = Matrix()

        // Compute the center of the view finder
        val centerX = view_finder.width / CENTER_COEFFICIENT
        val centerY = view_finder.height / CENTER_COEFFICIENT

        // Correct preview output to account for display rotation
        val rotationDegrees = when (view_finder.display.rotation) {
            Surface.ROTATION_0 -> ROTATION_DEGREES_0
            Surface.ROTATION_90 -> ROTATION_DEGREES_90
            Surface.ROTATION_180 -> ROTATION_DEGREES_180
            Surface.ROTATION_270 -> ROTATION_DEGREES_270
            else -> return
        }
        matrix.postRotate(-rotationDegrees.toFloat(), centerX, centerY)

        // Finally, apply transformations to our TextureView
        view_finder.setTransform(matrix)
    }

    override fun onDestroyView() {
        CameraX.unbindAll()
        super.onDestroyView()
    }

    companion object {
        const val ROTATION_DEGREES_0 = 0
        const val ROTATION_DEGREES_90 = 90
        const val ROTATION_DEGREES_180 = 180
        const val ROTATION_DEGREES_270 = 270
        const val CENTER_COEFFICIENT = 2f
        const val VIEW_INDEX = 0
    }
}