package ru.daotech.agroinvest.fragments.barcode_box

import android.content.Context
import android.os.Bundle
import android.view.*
import android.view.View.INVISIBLE
import android.view.View.VISIBLE
import androidx.drawerlayout.widget.DrawerLayout
import com.google.zxing.ResultPoint
import com.hannesdorfmann.mosby3.mvi.MviDialogFragment
import com.journeyapps.barcodescanner.BarcodeCallback
import com.journeyapps.barcodescanner.BarcodeResult
import com.journeyapps.barcodescanner.CaptureManager
import dagger.android.support.AndroidSupportInjection
import io.reactivex.Observable
import io.reactivex.subjects.PublishSubject
import kotlinx.android.synthetic.main.activity_container.*
import kotlinx.android.synthetic.main.fragment_barcode_box.*
import ru.daotech.agroinvest.R
import ru.daotech.agroinvest.components.NavigationFragmentComponentImpl
import ru.daotech.agroinvest.container.AppBarController
import ru.daotech.agroinvest.container.ContainerActivity
import ru.daotech.agroinvest.dialog.builder.BarcodeBoxDialogBuilder
import ru.daotech.agroinvest.fragments.detailed_task.DetailedTaskFragment
import ru.daotech.agroinvest.hide
import ru.daotech.agroinvest.show
import ru.daotech.agroinvest.utils.KEY_TASK_ID_BARCODE_BOX
import ru.daotech.agroinvest.utils.ServiceUtil.Companion.makeVibrationEffect
import ru.daotech.agroinvest.utils.ViewUtil.Companion.manageInternetConnectState
import ru.daotech.agroinvest.utils.ViewUtil.Companion.manageSocketConnectState
import timber.log.Timber
import javax.inject.Inject

class BarcodeBoxFragment :
    MviDialogFragment<BarcodeBoxView, BarcodeBoxPresenter>(), BarcodeBoxView {

    companion object {
        fun newInstance(): BarcodeBoxFragment {
            return BarcodeBoxFragment()
        }
    }

    @Inject
    lateinit var barcodeBoxPresenter: BarcodeBoxPresenter

    @Inject
    lateinit var dialogBuilder: BarcodeBoxDialogBuilder

    @Inject
    lateinit var navigationComponent: NavigationFragmentComponentImpl

    lateinit var captureManager: CaptureManager

    private val snackbarDismissSubject: PublishSubject<Unit> = PublishSubject.create()
    private val initiateSubject: PublishSubject<Unit> = PublishSubject.create()
    private val errorShowingSubject: PublishSubject<String> = PublishSubject.create()
    private val scanningResult: PublishSubject<String> = PublishSubject.create()
    private val scanningGroupResult: PublishSubject<Pair<String, String>> = PublishSubject.create()

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        setHasOptionsMenu(true)
        retainInstance = true
        this.activity?.drawer_layout?.setDrawerLockMode(DrawerLayout.LOCK_MODE_LOCKED_CLOSED)

        (this.activity as ContainerActivity).actionBar?.setDisplayShowHomeEnabled(true)
        (this.activity as ContainerActivity).actionBar?.setDisplayHomeAsUpEnabled(true)
    }

    override fun onAttach(context: Context) {
        AndroidSupportInjection.inject(this)
        super.onAttach(context)
    }

    override fun onPrepareOptionsMenu(menu: Menu) {
        super.onPrepareOptionsMenu(menu)
        (menu.findItem(R.id.menu_search) as MenuItem).isVisible = false

        activity?.let { activity ->
            manageInternetConnectState(activity, menu)
            manageSocketConnectState(activity, menu)
        }
    }

    override fun onCreateView(
        inflater: LayoutInflater,
        container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View? {
        (activity as? AppBarController)?.apply {
            setupAppBarWithNavController()
            showAppBar()
        }

        return inflater.inflate(R.layout.fragment_barcode_box, container, false)
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)

        // Initial intent
        initiateSubject.onNext(Unit)

        captureManager = CaptureManager(this.activity, barcodeView)
        captureManager.initializeFromIntent(requireActivity().intent, savedInstanceState)

        buildBarcodeScanCallback()

        barcode_box_button.setOnClickListener {
            buildBarcodeScanCallback()
        }
    }

    private fun buildBarcodeScanCallback() {
        barcodeView.decodeSingle(object : BarcodeCallback {
            override fun barcodeResult(result: BarcodeResult?) {
                result?.let {
                    showProgressbar()

                    dialogBuilder.buildBarcodeBoxAttentionDialog(
                        requireContext(),
                        {
                            when (parentFragment) {
                                is DetailedTaskFragment -> {
                                    scanningGroupResult.onNext(
                                        Pair(
                                            result.text!!,
                                            arguments?.getString(KEY_TASK_ID_BARCODE_BOX)!!
                                        )
                                    )
                                }
                                else -> {
                                    // Send barcode value
                                    scanningResult.onNext(result.text)
                                }
                            }
                        },
                        { hideProgressbar() },
                        { buildBarcodeScanCallback() },
                        result.text
                    )

                    makeVibrationEffect(requireContext())
                }
            }

            override fun possibleResultPoints(resultPoints: MutableList<ResultPoint>?) {
            }
        })
    }

    private fun showProgressbar() {
        barcode_box_progressbar.show()
        barcode_box_button.hide()
    }

    private fun hideProgressbar() {
        barcode_box_progressbar.hide()
        barcode_box_button.show()
    }

    override fun onResume() {
        super.onResume()
        captureManager.onResume()
    }

    override fun onPause() {
        super.onPause()
        captureManager.onPause()
    }

    override fun onDestroy() {
        super.onDestroy()
        captureManager.onDestroy()
    }

    override fun loadingIntent(): Observable<Unit> = Observable.just(Unit)

    override fun errorShowingIntent(): Observable<String> = errorShowingSubject

    override fun snackbarDismiss(): Observable<Unit> = snackbarDismissSubject

    override fun initiateSubjectIntent(): Observable<Unit> = initiateSubject

    override fun onBarcodeRetrieved(): Observable<String> = scanningResult

    override fun onDetailedTaskBarcodeScanRetrieved(): Observable<Pair<String, String>> =
        scanningGroupResult

    override fun render(state: BarcodeBoxViewState) {
        (activity as ContainerActivity).activity_container_custom_toolbar?.title = getString(R.string.barcode_box_scan_title)
        (activity as ContainerActivity).activity_container_custom_toolbar?.subtitle = null

        manageErrors(state)

        with(state) {
            if (state !is BarcodeBoxViewState.Loading) {
                hideProgressbar()
            }

            if (state is BarcodeBoxViewState.SendSuccess) {
                barcode_box_progressbar.visibility = INVISIBLE
                barcode_box_button.visibility = VISIBLE

                dialogBuilder.buildResponseDialog(
                    requireContext(),
                    requireFragmentManager(),
                    requireActivity()
                )
            }

            if (state is BarcodeBoxViewState.ShowGroupPropertyForm) {
                when (parentFragment) {
                    is DetailedTaskFragment -> {
                        (parentFragment as DetailedTaskFragment).loadGroupPropertyForm(
                            state.groupCoincidence
                        )
                    }
                }
                // Close current dialog
                dialog?.dismiss()
            }

            if (state is BarcodeBoxViewState.ShowChooseGroupPropertyDialog) {
                navigationComponent.buildAndShowChooseGroupTypeDialog(
                    requireFragmentManager(),
                    state.currentTaskId
                )

                dialog?.dismiss()
                dismiss()
            }

            if (state is BarcodeBoxViewState.ShowErrorBarcodeScanGroupPropertyDialog) {
                dialogBuilder.buildAndShowErrorBarcodeScanDialog(
                    requireContext(),
                    requireFragmentManager(),
                    requireActivity(),
                    getString(R.string.error_message_barcode_box_scanned)
                )

                dialog?.dismiss()
                dismiss()
            }
        }
    }

    // On view handling errors
    private fun manageErrors(state: BarcodeBoxViewState) {
        with(state) {
            when (state) {
                BarcodeBoxViewState.ClearError -> {
                    Timber.e("Ошибка")
                    barcode_box_progressbar.visibility = INVISIBLE
                    barcode_box_button.visibility = VISIBLE
                }
            }
        }
    }

    override fun createPresenter(): BarcodeBoxPresenter {
        return barcodeBoxPresenter
    }
}