package ru.daotech.agroinvest.fragments.barcode_box

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.Disposable
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.database.entities.TaskPlantImpl
import ru.daotech.agroinvest.utils.TASK_PROPERTY_GROUP_TYPE
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

class BarcodeBoxPresenter @Inject constructor(
    private val appStateManager: AppStateManager,
    private val repository: Repository
) : MviBasePresenter<BarcodeBoxView, BarcodeBoxViewState>() {

    private lateinit var disposable: Disposable

    override fun bindIntents() {

        val loadingIntent = intent(BarcodeBoxView::loadingIntent)
            .delay(3, TimeUnit.SECONDS)
            .flatMap { appStateManager.getCurrentWorker() }
            .map { BarcodeBoxViewState.SendFailed as BarcodeBoxViewState }
            .onErrorReturn {
                Timber.e(it)
                BarcodeBoxViewState.Error(it.localizedMessage)
            }

        val errorMessageIntent = intent(
            BarcodeBoxView::errorShowingIntent
        )
            .map { BarcodeBoxViewState.Error(it) }

        val initiateSubjectIntent = intent(
            BarcodeBoxView::initiateSubjectIntent
        )
            .map { BarcodeBoxViewState.EmptyState }

        val detailedTaskBarcodeScanIntent =
            intent(BarcodeBoxView::onDetailedTaskBarcodeScanRetrieved)
                .flatMap { (barcodeValue, taskId) ->
                    repository.getTaskWithProperties(taskId)
                        .observeOn(AndroidSchedulers.mainThread())
                        .map { taskWithProperties ->
                            appStateManager.saveCurrentBarcodeGroupValue(barcodeValue)

                            taskWithProperties.properties
                                .filter { it.type == TASK_PROPERTY_GROUP_TYPE }
                                .forEach { _ ->
                                    if (taskWithProperties.measurements
                                            .any {
                                                it.barcodeValue == barcodeValue
                                                        && it.taskId == taskId
                                                        && it.isMeasured
                                            }
                                    ) {
                                        return@map BarcodeBoxViewState.ShowErrorBarcodeScanGroupPropertyDialog
                                    }

                                    if (!taskWithProperties.measurements
                                            .any { it.barcodeValue == barcodeValue }
                                    ) {
                                        return@map BarcodeBoxViewState.ShowChooseGroupPropertyDialog(
                                            taskId
                                        )
                                    } else {
                                        val plant = taskWithProperties.plants
                                            .first {
                                                plant -> taskWithProperties.measurements
                                                .first { taskMeasurement ->
                                                    taskMeasurement.barcodeValue == barcodeValue
                                                }.plantId == plant.id
                                            } as TaskPlantImpl

                                        appStateManager.saveCurrentTaskPlant(plant)

                                        val groupList = taskWithProperties.properties
                                            .filter { it.type == TASK_PROPERTY_GROUP_TYPE }
                                        // Take group which coincidence with measurements
                                        val groupCoincidence = groupList
                                            .first { taskProperty ->
                                                taskWithProperties.measurements
                                                    .first { taskMeasurement ->
                                                        taskMeasurement.barcodeValue == barcodeValue
                                                    }.parentId == taskProperty.parentId
                                            }
                                        return@map BarcodeBoxViewState.ShowGroupPropertyForm(
                                            taskId,
                                            barcodeValue,
                                            groupCoincidence
                                        )
                                    }
                                }
                        }
                        .cast(BarcodeBoxViewState::class.java)
                }
                .onErrorReturn {
                    Timber.e(it)
                    BarcodeBoxViewState.Error(it.localizedMessage)
                }

        val snackbarDismissIntent = intent(
            BarcodeBoxView::snackbarDismiss
        )
            .map { BarcodeBoxViewState.ClearError }

        val onBarcodeRetrievedIntent = intent(
            BarcodeBoxView::onBarcodeRetrieved
        )
            .flatMap { repository.sendBarcodeBoxValueRequest(it) }
            .observeOn(AndroidSchedulers.mainThread())
            .map { BarcodeBoxViewState.SendSuccess as BarcodeBoxViewState }
            .startWith(BarcodeBoxViewState.Loading as BarcodeBoxViewState)
            .onErrorResumeNext { error: Throwable ->
                Observable.just(BarcodeBoxViewState.ClearError)
                    .map {
                        BarcodeBoxViewState.Error(
                            "Ошибка во время загрузки. Повоторите попытку."
                        )
                    }
                    .cast(BarcodeBoxViewState::class.java)
            }

        val allIntentList = listOf(
            loadingIntent,
            errorMessageIntent,
            onBarcodeRetrievedIntent,
            initiateSubjectIntent,
            detailedTaskBarcodeScanIntent
        )

        val allIntent = Observable.merge(
            allIntentList
        )
            .onErrorResumeNext { error: Throwable ->
                errorProcessing(error)
            }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        disposable = loadingIntent
            .filter { state ->
                state is BarcodeBoxViewState.SendSuccess
            }
            .map { state ->
                state as BarcodeBoxViewState.SendSuccess
            }
            .observeOn(AndroidSchedulers.mainThread())
            .subscribe(
                {

                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )


        subscribeViewState(allIntent, BarcodeBoxView::render)
    }

    private fun errorProcessing(error: Throwable): Observable<BarcodeBoxViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .map { BarcodeBoxViewState.ClearError as BarcodeBoxViewState }
            .startWith(
                BarcodeBoxViewState.Error(errorMessage = error.localizedMessage)
            )
    }
}