package ru.daotech.agroinvest.fragments.detailed_task

import android.annotation.SuppressLint
import android.net.Uri
import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.functions.BiFunction
import io.reactivex.schedulers.Schedulers
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.BuildConfig
import ru.daotech.agroinvest.business.CountlyEventControllerImpl
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.database.DatabaseRepository
import ru.daotech.agroinvest.database.entities.TaskPropertyImpl
import ru.daotech.agroinvest.database.relation.TaskWithProperties
import ru.daotech.agroinvest.exceptions.OutOfTasksInProgressException
import ru.daotech.agroinvest.models.InternetConnectivityState
import ru.daotech.agroinvest.utils.DateUtil
import ru.daotech.agroinvest.utils.ServiceUtil
import timber.log.Timber
import java.text.ParseException
import java.util.*
import java.util.concurrent.TimeUnit
import javax.inject.Inject

class DetailedTaskPresenter @Inject constructor(
    private val repository: Repository,
    private val appStateManager: AppStateManager,
    private var goToMyTasks: (() -> Unit)?,
    private var goToAccidentReport: ((String) -> Unit)?,
    private var goToTrolleyTied: ((String?) -> Unit)?,
    private var goToGroupProperty: ((String) -> Unit)?,
    private var goToBarcodeScan: (() -> Unit)?,
    private var databaseRepository: DatabaseRepository,
    private var countlyEventController: CountlyEventControllerImpl
) :
    MviBasePresenter<DetailedTaskView, DetailedTaskViewState>() {

    private var disposable: CompositeDisposable = CompositeDisposable()

    private lateinit var taskId: String
    private lateinit var taskWithProperties: TaskWithProperties
    private lateinit var internetConnectivityState: InternetConnectivityState

    @SuppressLint("CheckResult")
    override fun bindIntents() {
        val loadDetailedTaskFromDb = intent(DetailedTaskView::loadingIntent)
            .doOnNext { taskId -> this.taskId = taskId }
            .switchMap { taskId ->
                this.loadTaskWithProperties(
                    taskId,
                    isTaskStarted = false
                )
            }

        intent(DetailedTaskView::internetConnectivityStateIntent)
            .subscribe { internetConnectivityState ->
                this.internetConnectivityState = internetConnectivityState
            }

        val taskFinishAcceptIntent = intent(DetailedTaskView::taskFinishAcceptClick)
            .switchMap { pairValue ->
                taskId = pairValue.second
                handleNfcMarkScan(pairValue.first)
            }

        val taskFinishWithPropsIntent = intent(DetailedTaskView::taskFinishWithPropsAcceptClick)
            .switchMap { (nfcMark, taskPropertiesPayload, uriList) ->
                val properties = taskPropertiesPayload.map {
                    TaskPropertyImpl(it!!, taskId)
                }
                return@switchMap databaseRepository.upsertTaskPropertiesList(properties)
                    .andThen { repository.uploadPropertyImages(uriList) }
                    .subscribeOn(Schedulers.single())
                    .observeOn(AndroidSchedulers.mainThread())
                    .toSingleDefault(Unit)
                    .toObservable()
                    .flatMap { handleNfcMarkScan(nfcMark) }
                    .doOnError { Timber.e(it) }
            }

        val sendGroupPropertyIntent = intent(DetailedTaskView::onSendGroupPropertyClick)
            .switchMap { (propertyList, uriList) ->
                val properties = propertyList.map { propertyPayload ->
                    TaskPropertyImpl(
                        propertyPayload!!,
                        taskId,
                        DateUtil.getCurrentDateTimestampWithDelta(
                            appStateManager.getServerTimestamp()
                        )
                    )
                }
                val groupId = propertyList.first()!!.parentId
                return@switchMap signalToUpdateGroupProperty(
                    groupId,
                    properties,
                    uriList
                )
            }

        val onDismissResponseDialogIntent = intent(DetailedTaskView::onDismissResponseDialogIntent)
            .map { DetailedTaskPartialViewState.DismissResponseDialogState }

        val onStartButtonClick = intent(DetailedTaskView::startButtonClick)
            .switchMap { nfcMark ->
                val rowNumber = taskWithProperties.rowNumber.toString()

                if (nfcMark != rowNumber && nfcMark.isNotBlank()) {
                    return@switchMap Observable.just(
                        DetailedTaskPartialViewState.Error(
                            "Номера рядов не совпадают. Отсканируйте другую NFC метку."
                        )
                    )
                }

                // Saving current row number to local storage
                appStateManager.saveCurrentRowNumber(rowNumber)

                countlyEventController.sendCountlyEvent(
                    COUNTLY_EVENT_NETWORK_INFO,
                    internetConnectivityState
                )
                countlyEventController.sendCountlyEvent("ROW_$rowNumber", internetConnectivityState)

                // Save current time's mark for calculate scanning nfc mark interval
                appStateManager.saveCurrentTimeStartMark(
                    DateUtil.getCurrentDateTimestampWithDelta(appStateManager.getServerTimestamp())
                )

                this.startTaskWithProperties()
            }

        val onTaskCancelClick = intent(DetailedTaskView::taskCancelAcceptClick)
            .switchMap { this.cancelTaskWithProperties() }

        val onErrorDismiss = intent(DetailedTaskView::onErrorDismiss)
            .map { DetailedTaskPartialViewState.ClearError }

        val onShowAttentionCheck = intent(DetailedTaskView::showAttentionCheck)
            .switchMap { this.notShowAttention() }

        val goToMyTaskIntent = intent(DetailedTaskView::goToMyTasksIntent)
            .map { DetailedTaskPartialViewState.GoingToMyTasks }

        val goToBarcodeBoxFragmentIntent = intent(DetailedTaskView::onGoToBarcodeBoxFragmentIntent)
            .map { DetailedTaskPartialViewState.GoingToBarcodeScan }

        val goToCurrentTrolleyTiedIntent = intent(DetailedTaskView::onCurrentTrolleyTiedIntent)
            .map { DetailedTaskPartialViewState.GoingToTrolleyTiedPage }

        val dismissTaskSates = intent(DetailedTaskView::dismissTaskStates)
            .map { DetailedTaskPartialViewState.DismissTaskStates }

        val snackbarDismissIntent = intent(DetailedTaskView::snackbarDismiss)
            .map { DetailedTaskPartialViewState.DismissSnackbarErrorStates }

        val allIntentList = listOf(
            loadDetailedTaskFromDb,
            onStartButtonClick,
            onTaskCancelClick,
            goToMyTaskIntent,
            onShowAttentionCheck,
            dismissTaskSates,
            onErrorDismiss,
            taskFinishAcceptIntent,
            taskFinishWithPropsIntent,
            goToCurrentTrolleyTiedIntent,
            sendGroupPropertyIntent,
            goToBarcodeBoxFragmentIntent,
            onDismissResponseDialogIntent,
            snackbarDismissIntent
        )

        val allIntent = Observable.merge(allIntentList)
            .onErrorResumeNext { error: Throwable -> errorProcessing(error) }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        disposable.add(
            allIntent
                .filter { state -> state is DetailedTaskPartialViewState.GoingToMyTasks }
                .takeUntil { state -> state is DetailedTaskPartialViewState.GoingToMyTasks }
                .subscribe(
                    { goToMyTasks?.invoke() },
                    { error -> Timber.e(error, error.localizedMessage) }
                )
        )

        disposable.add(
            allIntent
                .filter { state -> state is DetailedTaskPartialViewState.GoingToTrolleyTiedPage }
                //.takeUntil { state -> state is DetailedTaskPartialViewState.GoingToTrolleyTiedPage }
                .subscribe(
                    { goToTrolleyTied?.invoke(null) },
                    { error -> Timber.e(error, error.localizedMessage) }
                )
        )

        disposable.add(
            allIntent
                .filter { state -> state is DetailedTaskPartialViewState.GoingToBarcodeScan }
                //.takeUntil { state -> state is DetailedTaskPartialViewState.GoingToTrolleyTiedPage }
                .subscribe(
                    { goToBarcodeScan?.invoke() },
                    { error -> Timber.e(error, error.localizedMessage) }
                )
        )

        val initialState =
            DetailedTaskViewState(
                isLoading = false,
                taskWithProperties = null,
                errorMessage = null,
                isTaskStartedSuccess = false,
                isGroupPropertyUploadSuccess = false,
                nfcMode = appStateManager.getCurrentWorkerSimple()?.nfcMode
            )

        val stateObservable = allIntent
            .scan(initialState) { previousState, partialViewState ->
                partialViewState.reduce(previousState)
            }

        subscribeViewState(stateObservable, DetailedTaskView::render)
    }

    private fun handleNfcMarkScan(nfcMark: String): Observable<DetailedTaskPartialViewState> {
        return if (nfcMark.isNotBlank()) {
            if (nfcMark != taskWithProperties.rowNumber.toString()
                || (nfcMark != appStateManager.getCurrentRowNumber()
                        && appStateManager.getCurrentRowNumber().isNotBlank())
                && nfcMark.isNotBlank()
            ) {
                Observable.just(
                    DetailedTaskPartialViewState.Error(
                        "Метки ряда не совпадают."
                    ) as DetailedTaskPartialViewState
                )
            } else {
                checkTaskForFinish()
            }
        } else {
            checkTaskForFinish()
        }
    }

    private fun errorProcessing(error: Throwable): Observable<DetailedTaskPartialViewState> {
        Timber.e(error)
        return Observable.timer(TIMER_DELAY_SECONDS, TimeUnit.SECONDS)
            .map { Unit }
            .map { DetailedTaskPartialViewState.ClearError as DetailedTaskPartialViewState }
            .startWith(
                DetailedTaskPartialViewState
                    .Error(errorMessage = error.localizedMessage)
            )
    }

    private fun checkTaskForFinish(): Observable<DetailedTaskPartialViewState> {
        return databaseRepository.getTaskWithProperties(
            taskId,
            appStateManager.getSessionId()
        )
            .toObservable()
            .observeOn(AndroidSchedulers.mainThread())
            .flatMap { currentTaskWithProperties ->
                if (!currentTaskWithProperties.startedAt.isNullOrBlank()
                    && currentTaskWithProperties.finishedAt.isNullOrBlank()
                    || !(currentTaskWithProperties.startedAt.isNullOrBlank()
                            && !currentTaskWithProperties.isFinished)
                    || !currentTaskWithProperties.isFinished
                ) {
                    currentTaskWithProperties.startedAt?.let { lastTaskTimeStartMark ->
                        if (!lastTaskTimeStartMark.isBlank()) {

                            // If current task is phenology task then it could be finish
                            if (currentTaskWithProperties.sectionId == BuildConfig.phenologySectionId) {
                                return@flatMap finishTaskWithProperties()
                            }

                            val currentTaskIntervalSumSeconds = TimeUnit.MILLISECONDS.toSeconds(
                                ServiceUtil.calculateTaskIntervalsSum(
                                    currentTaskWithProperties,
                                    appStateManager.getServerTimestamp()
                                )
                            )

                            val startTime = Calendar.getInstance().timeInMillis
                            var diffSeconds: Long = 0

                            try {
                                diffSeconds =
                                    (Calendar.getInstance().timeInMillis - startTime + currentTaskIntervalSumSeconds)
                            } catch (e: ParseException) {
                                e.printStackTrace()
                            }

                            Timber.d("TIME DIFF: $currentTaskIntervalSumSeconds")

                            val secondsTaskLimit: Int = when (BuildConfig.BUILD_TYPE) {
                                "debug" -> {
                                    10
                                }
                                else -> {
                                    taskWithProperties.taskLimit?.let { it * 60 } ?: 60
                                }
                            }

                            return@flatMap if (diffSeconds >= secondsTaskLimit) {
                                finishTaskWithProperties()
                            } else {
                                Observable.just(
                                    DetailedTaskPartialViewState.Error(
                                        "Вы не можете завершить задачу менее, " +
                                                "чем за " +
                                                "$secondsTaskLimit" +
                                                " секунд."
                                    )
                                )
                            }
                        }
                    }
                }
                return@flatMap Observable.just(DetailedTaskPartialViewState.EmptyState)
                    .cast(DetailedTaskPartialViewState::class.java)
            }
    }

    private fun notShowAttention(): Observable<DetailedTaskPartialViewState> {
        return repository.notShowAttentionAgain()
            .observeOn(AndroidSchedulers.mainThread())
            .map {
                DetailedTaskPartialViewState.EmptyState
            }
            .doOnError { error ->
                Timber.e(error, error.localizedMessage)
            }
            .cast(DetailedTaskPartialViewState::class.java)
    }

    private fun startTaskWithProperties(): Observable<DetailedTaskPartialViewState> {
        if ((taskWithProperties.isHarvesting != null && taskWithProperties.isHarvesting == true)
            && appStateManager.getCurrentTrolleyNumber().isBlank()
        ) {
            return Observable.just(
                DetailedTaskPartialViewState.Error(
                    "Перед запуском задачи, возьмите доступную тележку в работу."
                )
            )
        }
        return Observable.combineLatest(
            repository.signalToTaskStart(taskId),
            loadTaskWithProperties(taskId, isTaskStarted = true),
            BiFunction<Unit, DetailedTaskPartialViewState, DetailedTaskPartialViewState> { _, state ->
                state
            }
        )
            .observeOn(AndroidSchedulers.mainThread())
            .cast(DetailedTaskPartialViewState::class.java)
            .startWith(DetailedTaskPartialViewState.Loading)
            .onErrorResumeNext { error: Throwable ->
                Observable.timer(TIMER_DELAY_SECONDS, TimeUnit.SECONDS)
                    .map {
                        DetailedTaskPartialViewState
                            .Error(
                                "Есть задачи в работе. " +
                                        "Поставьте их на паузу, начиная с первой"
                            )
                                as DetailedTaskPartialViewState
                    }
            }
            .doOnError {
                if (it.cause is OutOfTasksInProgressException) {
                    throw RuntimeException()
                }
                Timber.e(it)
            }
            .onExceptionResumeNext(
                Observable.timer(TIMER_DELAY_SECONDS, TimeUnit.SECONDS)
                    .map {
                        DetailedTaskPartialViewState
                            .Error(
                                "Есть задачи в работе. " +
                                        "Поставьте их на паузу, начиная с первой"
                            )
                                as DetailedTaskPartialViewState
                    }
            )
    }

    private fun loadTaskWithProperties(
        taskId: String,
        isTaskStarted: Boolean
    ): Observable<DetailedTaskPartialViewState> {
        return repository.getTaskWithProperties(taskId)
            .map { taskWithProperties ->
                this.taskWithProperties = taskWithProperties

                // Saving current row number to local storage
                taskWithProperties.rowNumber?.let {
                    appStateManager.saveCurrentRowNumber(it.toString())
                }

                when (isTaskStarted) {
                    true -> {
                        DetailedTaskPartialViewState.SuccessStart(taskWithProperties)
                    }
                    false -> {
                        DetailedTaskPartialViewState.SuccessLoadTasks(taskWithProperties)
                    }
                }
            }
            .doOnError { error ->
                Timber.e(error, error.localizedMessage)
            }
            .cast(DetailedTaskPartialViewState::class.java)
    }

    private fun cancelTaskWithProperties(): Observable<DetailedTaskPartialViewState> {
        return repository.signalToTaskCancel(taskId)
            .doOnNext { appStateManager.clearCurrentRowNumber() }
            .observeOn(AndroidSchedulers.mainThread())
            .map { DetailedTaskPartialViewState.GoingToMyTasks }
            .cast(DetailedTaskPartialViewState::class.java)
    }

    private fun finishTaskWithProperties(): Observable<DetailedTaskPartialViewState> {
        return repository.signalToTaskFinish(taskId)
            .observeOn(AndroidSchedulers.mainThread())
            .doOnNext { appStateManager.clearCurrentRowNumber() }
            .map { DetailedTaskPartialViewState.GoingToMyTasks }
            .cast(DetailedTaskPartialViewState::class.java)
    }

    private fun signalToUpdateGroupProperty(
        groupId: String,
        properties: List<TaskPropertyImpl>,
        uriList: List<Uri>
    ): Observable<DetailedTaskPartialViewState> {
        return repository.signalToGroupPropertyCreate(
            taskId,
            properties,
            groupId,
            appStateManager.getCurrentBarcodeGroupValue()!!
        )
            .flatMap { repository.uploadPropertyImages(uriList) }
            .observeOn(AndroidSchedulers.mainThread())
            .map { DetailedTaskPartialViewState.SuccessGroupPropertyUpload }
            .cast(DetailedTaskPartialViewState::class.java)
    }

    override fun unbindIntents() {
        super.unbindIntents()

        disposable.dispose()

        goToMyTasks = null
        goToAccidentReport = null
        goToTrolleyTied = null
        goToGroupProperty = null
        goToBarcodeScan = null
    }

    companion object {
        const val COUNTLY_EVENT_NETWORK_INFO = "NETWORK_INFO"
        const val TIMER_DELAY_SECONDS = 3L
    }
}
