package ru.daotech.agroinvest.fragments.detailed_task.detailed_task_in_progress

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.business.Repository
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

class DetailedTaskProcessingPresenter @Inject constructor(
    private val repository: Repository,
    private val appStateManager: AppStateManager,
    private var goToTrolleyTied: ((String?) -> Unit)?
) :
    MviBasePresenter<DetailedTaskProcessingTabView, DetailedTaskProcessingViewState>() {

    private var disposable: CompositeDisposable = CompositeDisposable()

    override fun bindIntents() {

        val errorIntent = intent(DetailedTaskProcessingTabView::errorShown)

        val fragmentChangeIntent = intent(DetailedTaskProcessingTabView::fragmentChange)
            .switchMap { Observable.just(DetailedTaskProcessingPartialViewState.TaskStarted as DetailedTaskProcessingPartialViewState) }
            .onErrorResumeNext { error: Throwable ->
                Timber.e(error)
                Observable.timer(3, TimeUnit.SECONDS)
                    .map { Unit }
                    .mergeWith(errorIntent)
                    .take(1)
                    .map {
                        DetailedTaskProcessingPartialViewState.TaskError as DetailedTaskProcessingPartialViewState
                    }
                    .startWith(
                        DetailedTaskProcessingPartialViewState.TaskErrorWithMessage(
                            errorMessage = error.localizedMessage
                        )
                    )
            }

        val onCurrentTrolleyIntent = intent(
            DetailedTaskProcessingTabView::onCurrentTrolleyIntent
        )
            .switchMap { trolleyNumber ->
                when (!appStateManager.getCurrentRowNumber().isBlank()) {
                    true -> {
                        return@switchMap Observable.just(
                            DetailedTaskProcessingPartialViewState.GoToTrolleyTied(trolleyNumber)
                                    as DetailedTaskProcessingPartialViewState
                        )
                    }
                    false -> {
                        Observable.just(
                            DetailedTaskProcessingPartialViewState.GoToTrolleyTied(trolleyNumber)
                                    as DetailedTaskProcessingPartialViewState
                        )
                    }
                }
            }

        val errorWrongNfcTypeIntent = intent(
            DetailedTaskProcessingTabView::errorWrongNfcTypeIntent
        )
            .switchMap {
                Observable.just(
                    DetailedTaskProcessingPartialViewState.TaskErrorWithMessage(
                        "Ошибка. Неверный тип NFC метки."
                    )
                )
            }

        val allIntentList = listOf(
            fragmentChangeIntent,
            onCurrentTrolleyIntent,
            errorWrongNfcTypeIntent
        )

        val initialState =
            DetailedTaskProcessingViewState(
                isStarted = false,
                errorMessage = null,
                nfcTrolleyMark = null
            )


        val allIntent = Observable.merge(allIntentList)
            .onErrorResumeNext { error: Throwable ->
                Timber.e(error)
                Observable.timer(3, TimeUnit.SECONDS)
                    .map { Unit }
                    .mergeWith(errorIntent)
                    .take(1)
                    .map {
                        DetailedTaskProcessingPartialViewState.TaskError as DetailedTaskProcessingPartialViewState
                    }
                    .startWith(
                        DetailedTaskProcessingPartialViewState.TaskErrorWithMessage(
                            errorMessage = error.localizedMessage
                        )
                    )
            }
            .observeOn(AndroidSchedulers.mainThread())
            .share()


        disposable.add(
            allIntent
                .filter { state -> state is DetailedTaskProcessingPartialViewState.GoToTrolleyTied }
                .map { state -> state as DetailedTaskProcessingPartialViewState.GoToTrolleyTied }
                .subscribe(
                    { state -> goToTrolleyTied?.invoke(state.nfcTrolleyMark) },
                    { error -> Timber.e(error, error.localizedMessage) }
                )
        )

        val stateObservable =
            allIntent.scan(initialState) { previousState, partialViewState ->
                partialViewState.reduce(previousState)
            }

        subscribeViewState(stateObservable, DetailedTaskProcessingTabView::render)
    }
}