package ru.daotech.agroinvest.fragments.detailed_task.detailed_task_in_progress.header_field_container

import android.Manifest
import android.app.Activity
import android.content.Context
import android.content.Intent
import android.content.pm.PackageManager
import android.graphics.Bitmap
import android.net.Uri
import android.os.Build
import android.os.Bundle
import android.provider.MediaStore
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.LinearLayout
import androidx.core.content.ContextCompat
import androidx.core.graphics.scale
import androidx.recyclerview.widget.LinearLayoutManager
import com.hannesdorfmann.mosby3.mvi.MviFragment
import dagger.android.support.AndroidSupportInjection
import io.reactivex.Observable
import io.reactivex.subjects.BehaviorSubject
import io.reactivex.subjects.PublishSubject
import kotlinx.android.synthetic.main.fragment_header_field.*
import kotlinx.android.synthetic.main.fragment_header_field.view.*
import kotlinx.android.synthetic.main.item_image_binary_view.view.*
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.R
import ru.daotech.agroinvest.adapters.HeaderFieldAdapter
import ru.daotech.agroinvest.components.TextViewComponentBuilder
import ru.daotech.agroinvest.dialog.builder.HeaderDialogBuilder
import ru.daotech.agroinvest.service.jsonmodels.TaskPropertyPayload
import ru.daotech.agroinvest.utils.*
import timber.log.Timber
import javax.inject.Inject

class HeaderFieldFragment : MviFragment<HeaderFieldView, HeaderFieldPresenter>(),
    HeaderFieldView {

    @Inject
    lateinit var headerFieldPresenter: HeaderFieldPresenter

    @Inject
    lateinit var textViewComponentBuilder: TextViewComponentBuilder

    @Inject
    lateinit var headerDialogBuilder: HeaderDialogBuilder

    @Inject
    lateinit var appStateManager: AppStateManager

    private val onErrorShownSubject: PublishSubject<Unit> = PublishSubject.create()
    private val dismissSnackbarSubject: PublishSubject<Unit> = PublishSubject.create()
    private val buildGroupFormSubject: PublishSubject<Unit> = PublishSubject.create()
    private val buildHeaderFormSubject: PublishSubject<Unit> = PublishSubject.create()

    companion object {
        const val GALLERY_IMAGE_PATH = "image/*"
        const val THUMBNAIL_TARGET_HEIGHT = 100
        const val THUMBNAIL_TARGET_WIDTH = 100

        fun newInstance(): HeaderFieldFragment {
            return HeaderFieldFragment()
        }
    }

    override fun onAttach(context: Context) {
        AndroidSupportInjection.inject(this)
        super.onAttach(context)
    }

    override fun onCreateView(
        inflater: LayoutInflater,
        container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View? {
        val view = inflater.inflate(R.layout.fragment_header_field, container, false)

        view.recycler_header_fields.apply {
            layoutManager = LinearLayoutManager(context)
            adapter = HeaderFieldAdapter(
                requireContext(),
                headerDialogBuilder::buildAndShowFullscreenImageDialog,
                { takePhoto() },
                { pickImageFromGallery() },
                requireActivity(),
                { showAlertDialog(context.getString(R.string.header_submit_image_attention_dialog_title)) },
                ::showTextFieldHintDialog
            )
        }
        return view
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)
        with(arguments?.getString(KEY_TASK_ID_HEADER_FIELD)) {
            if (this.isNullOrBlank()) {
                buildGroupFormSubject.onNext(Unit)
            } else {
                buildHeaderFormSubject.onNext(Unit)
            }
        }
    }

    override fun loadingIntent(): Observable<Bundle> = Observable.just(Unit)
        .map { arguments }

    override fun loadingGroupPropertiesIntent(): Observable<String> = buildGroupFormSubject
        .map { arguments?.getString(KEY_TASK_ID_GROUP_PROPERTY)?.let { it } }

    override fun errorShown(): Observable<Unit> = onErrorShownSubject

    override fun dismissSnackbar(): Observable<Unit> = dismissSnackbarSubject

    override fun render(state: HeaderFieldViewState) {
        manageErrors(state)

        if (state is HeaderFieldViewState.BuildForm) {
            with((recycler_header_fields.adapter as HeaderFieldAdapter)) {
                when (propertyList.isNullOrEmpty()) {
                    true -> {
                        propertyList =
                            textViewComponentBuilder.buildPropertyViewModelList(state.taskWithProperties)
                        notifyDataSetChanged()
                    }
                }
            }
        }

        if (state is HeaderFieldViewState.BuildGroupPropertyForm) {
            with((recycler_header_fields.adapter as HeaderFieldAdapter)) {
                when (propertyList.isNullOrEmpty()) {
                    true -> {
                        arguments?.getString(KEY_BARCODE_VALUE_GROUP_PROPERTY)
                            ?.let { barcodeValue ->
                                propertyList =
                                    textViewComponentBuilder.buildGroupPropertyViewModelListByBarcode(
                                        state.taskWithProperties,
                                        barcodeValue
                                    )
                                notifyDataSetChanged()
                                return
                            }
                        propertyList = textViewComponentBuilder.buildGroupPropertyViewModelList(
                            state.taskWithProperties
                        )
                        notifyDataSetChanged()
                    }
                }
            }
        }
    }

    // On view handling errors
    private fun manageErrors(state: HeaderFieldViewState) {
        when (state) {
            is Error -> {
            }
            else -> {
            }
        }
    }

    override fun onRequestPermissionsResult(
        requestCode: Int,
        permissions: Array<out String>,
        grantResults: IntArray
    ) {
        when (requestCode) {
            PERMISSION_CODE -> {
                if (grantResults.isNotEmpty() && grantResults[0] ==
                    PackageManager.PERMISSION_GRANTED
                ) {
                    // Permission from popup granted
                    pickImageFromGallery()
                } else {
                    // Permission from popup denied
                    Timber.i("Permission denied")
                }
            }
            CAMERA_PERMISSION_CODE -> {
                if (grantResults.isNotEmpty() && grantResults[0]
                    == PackageManager.PERMISSION_GRANTED
                ) {
                    // Permission from popup granted
                    takePhoto()
                } else {
                    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                        if (ContextCompat.checkSelfPermission(requireContext(), Manifest.permission.CAMERA)
                            != PackageManager.PERMISSION_GRANTED
                        ) {
                            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                                // Show popup to request runtime permission
                                requestPermissions(
                                    arrayOf(
                                        Manifest.permission.CAMERA,
                                        Manifest.permission.READ_EXTERNAL_STORAGE
                                    ),
                                    CAMERA_PERMISSION_CODE
                                )
                            }
                        }
                    }
                }
            }
        }
    }

    // Handle result of picked image
    override fun onActivityResult(requestCode: Int, resultCode: Int, data: Intent?) {
        super.onActivityResult(requestCode, resultCode, data)

        if (resultCode == Activity.RESULT_OK && data != null) {
            when (requestCode) {
                REQUEST_PHOTO_CAPTURE -> {
                    val bitmap = (data.extras!!.get("data") as Bitmap)

                    with(recycler_header_fields) {
                        val imageViewPos =
                            (adapter as HeaderFieldAdapter).getImageViewPos()

                        (adapter as HeaderFieldAdapter).setBitmapProperty(bitmap)

                        (findViewHolderForAdapterPosition(imageViewPos)?.itemView
                                as LinearLayout)
                            .phenology_preview_image_view
                            .setImageBitmap(
                                bitmap.scale(
                                    THUMBNAIL_TARGET_WIDTH,
                                    THUMBNAIL_TARGET_HEIGHT
                                )
                            )

                    }
                }
                IMAGE_PICK_CODE -> {
                    with(recycler_header_fields) {
                        val imageViewPos =
                            (adapter as HeaderFieldAdapter).getImageViewPos()
                        val bitmap = MediaStore.Images.Media.getBitmap(
                            requireActivity().contentResolver,
                            data.data
                        )

                        (adapter as HeaderFieldAdapter).setBitmapProperty(bitmap)

                        (findViewHolderForAdapterPosition(imageViewPos)?.itemView as LinearLayout)
                            .phenology_preview_image_view
                            .setImageBitmap(
                                bitmap.scale(
                                    THUMBNAIL_TARGET_WIDTH,
                                    THUMBNAIL_TARGET_HEIGHT
                                )
                            )
                    }
                }
            }
        }
    }

    private fun showAlertDialog(message: String) {
        headerDialogBuilder.buildAndShowErrorDialog(requireContext(), message)
    }

    private fun showTextFieldHintDialog(message: String) {
        headerDialogBuilder.buildAndShowHintDialog(requireContext(), message)
    }

    fun signalToFormSubmit(
        onPropertyRetrieveSubject: BehaviorSubject<Pair<List<TaskPropertyPayload?>, List<Uri>>>
    ) {
        val propertyList = (recycler_header_fields.adapter as HeaderFieldAdapter)
            .propertyList

        headerDialogBuilder.buildAndShowAttentionDialog(
            propertyList,
            {
                (recycler_header_fields.adapter as HeaderFieldAdapter)
                    .handleSubmitForm(onPropertyRetrieveSubject)
            },
            appStateManager.getCurrentTaskPlant()!!,
            requireContext()
        )
    }

    private fun pickImageFromGallery() {
        //Intent to pick image
        val intent = Intent(Intent.ACTION_PICK)
        intent.type = GALLERY_IMAGE_PATH
        startActivityForResult(intent, IMAGE_PICK_CODE)
    }

    private fun takePhoto() {
        val cameraIntent = Intent(MediaStore.ACTION_IMAGE_CAPTURE)
        startActivityForResult(cameraIntent, REQUEST_PHOTO_CAPTURE)
    }

    override fun createPresenter(): HeaderFieldPresenter {
        return headerFieldPresenter
    }
}