package ru.daotech.agroinvest.fragments.detailed_task.detailed_task_in_progress.timer_container

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.exceptions.OutOfTasksInPauseException
import ru.daotech.agroinvest.fragments.detailed_task.DetailedTaskPresenter.Companion.TIMER_DELAY_SECONDS
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

class TimerContainerViewPresenter @Inject constructor(
    private val repository: Repository,
    private val appStateManager: AppStateManager,
    private var goToTrolleyTied: ((String?) -> Unit)?
) :
    MviBasePresenter<TimerContainerView, TimerContainerViewState>() {

    private var disposable: CompositeDisposable = CompositeDisposable()

    override fun bindIntents() {

        val onTimerControlClick = intent(TimerContainerView::timerControlClick)
            .throttleFirst(2, TimeUnit.SECONDS)
            .switchMap { (isTaskStarted, taskId) ->
                if (!isTaskStarted) {
                    repository.fixTimerStartMark(taskId)
                        .observeOn(AndroidSchedulers.mainThread())
                        .map { TimerContainerPartialViewState.EmptyState }
                        .cast(TimerContainerPartialViewState::class.java)
                } else {
                    this.cancelTaskWithProperties(taskId)
                }
            }

        val onSaveCurrentTimerState = intent(TimerContainerView::saveCurrentTimerState)
            .map { isTimerStateRun ->
                TimerContainerPartialViewState.FixTaskTimerState(
                    isTimerStateRun
                )
            }

        val dismissTimerState = intent(TimerContainerView::dismissTimerState)
            .map { TimerContainerPartialViewState.DismissTimerChangedState }

        val fragmentSaveCurrentStateIntent =
            intent(TimerContainerView::saveCurrentFragmentState)
                .switchMap { lastFragmentState ->
                    Observable.just(
                        TimerContainerPartialViewState.FragmentStateSaving(
                            taskId = lastFragmentState.currentTaskId,
                            taskInterval = lastFragmentState.currentTaskInterval
                        ) as TimerContainerPartialViewState
                    )
                }

        val onCurrentTrolleyIntent = intent(
            TimerContainerView::onCurrentTrolleyIntent
        )
            .switchMap { trolleyNumber ->
                when (!appStateManager.getCurrentRowNumber().isBlank()) {
                    true -> {
                        return@switchMap Observable.just(
                            TimerContainerPartialViewState.GoToTrolleyTied(trolleyNumber) as TimerContainerPartialViewState
                        )
                    }
                    false -> {
                        Observable.just(
                            TimerContainerPartialViewState.GoToTrolleyTied(trolleyNumber) as TimerContainerPartialViewState
                        )
                    }
                }
            }

        val errorWrongNfcTypeIntent = intent(
            TimerContainerView::errorWrongNfcTypeIntent
        )
            .switchMap {
                Observable.just(
                    TimerContainerPartialViewState.TaskErrorWithMessage(
                        "Ошибка. Неверный тип NFC метки."
                    )
                )
            }

        val errorIntent = intent(TimerContainerView::errorShown)

        val initialState =
            TimerContainerViewState(
                isStarted = false,
                errorMessage = null,
                currentTaskId = null,
                currentTaskInterval = null,
                isTimerRunState = false,
                isStateChanged = false,
                nfcTrolleyMark = null
            )

        val allIntentList = listOf(
            onTimerControlClick,
            fragmentSaveCurrentStateIntent,
            onSaveCurrentTimerState,
            dismissTimerState,
            onCurrentTrolleyIntent,
            errorWrongNfcTypeIntent
        )

        val allIntent = Observable.merge(allIntentList)
            .onErrorResumeNext { error: Throwable ->
                Timber.e(error)
                Observable.timer(3, TimeUnit.SECONDS)
                    .map { Unit }
                    .mergeWith(errorIntent)
                    .take(1)
                    .map {
                        TimerContainerPartialViewState.TaskErrorWithoutMessage as TimerContainerPartialViewState
                    }
                    .startWith(
                        TimerContainerPartialViewState.TaskErrorWithMessage(
                            errorMessage = error.localizedMessage
                        )
                    )
            }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        disposable.add(
            allIntent
                .filter { state -> state is TimerContainerPartialViewState.FragmentStateSaving }
                .map { state -> state as TimerContainerPartialViewState.FragmentStateSaving }
                .map { state ->
                    repository.saveCurrentFragmentState(
                        state.taskInterval,
                        state.taskId
                    )
                }
                .subscribe(
                    { Unit },
                    { error ->
                        Timber.e(error, error.localizedMessage)
                    }
                )
        )

        disposable.add(
            allIntent
                .filter { state -> state is TimerContainerPartialViewState.GoToTrolleyTied }
                .map { state -> state as TimerContainerPartialViewState.GoToTrolleyTied }
                .subscribe(
                    { state -> goToTrolleyTied?.invoke(state.nfcTrolleyMark) },
                    { error -> Timber.e(error, error.localizedMessage) }
                )
        )

        val stateObservable = allIntent.scan(initialState) { previousState, partialViewState ->
            partialViewState.reduce(previousState)
        }

        subscribeViewState(stateObservable, TimerContainerView::render)
    }

    private fun cancelTaskWithProperties(taskId: String): Observable<TimerContainerPartialViewState> {
        return repository.signalToTaskCancel(taskId)
            .doOnNext { appStateManager.clearCurrentRowNumber() }
            .observeOn(AndroidSchedulers.mainThread())
            .map { TimerContainerPartialViewState.EmptyState }
            .cast(TimerContainerPartialViewState::class.java)
            .onErrorResumeNext { error: Throwable ->
                Observable.timer(1L, TimeUnit.SECONDS)
                    .map {
                        TimerContainerPartialViewState
                            .DisablePauseForThisTaskState(
                                "Есть задачи на паузе. " +
                                        "Завершите задачи на паузе, либо запустите их в работу."
                            )
                                as TimerContainerPartialViewState
                    }
            }
            .doOnError {
                if (it.cause is OutOfTasksInPauseException) {
                    throw RuntimeException()
                }
                Timber.e(it)
            }
            .onExceptionResumeNext(
                Observable.timer(1L, TimeUnit.SECONDS)
                    .map {
                        TimerContainerPartialViewState
                            .DisablePauseForThisTaskState(
                                "Есть задачи на паузе. " +
                                        "Завершите задачи на паузе, либо запустите их в работу."
                            )
                                as TimerContainerPartialViewState
                    }
            )
    }
}