package ru.daotech.agroinvest.fragments.detailed_task.detailed_task_in_progress.timer_container

import android.annotation.SuppressLint
import android.os.Bundle
import android.view.LayoutInflater
import android.view.View
import android.view.View.GONE
import android.view.View.VISIBLE
import android.view.ViewGroup
import androidx.core.content.ContextCompat
import androidx.core.view.marginStart
import io.reactivex.Flowable
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.processors.PublishProcessor
import io.reactivex.subjects.PublishSubject
import kotlinx.android.synthetic.main.fragment_timer_container.*
import org.greenrobot.eventbus.EventBus
import org.greenrobot.eventbus.Subscribe
import org.greenrobot.eventbus.ThreadMode
import org.joda.time.DateTime
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.R
import ru.daotech.agroinvest.database.relation.TaskWithProperties
import ru.daotech.agroinvest.fragments.base.BaseMviFragment
import ru.daotech.agroinvest.fragments.detailed_task.DetailedTaskFragment
import ru.daotech.agroinvest.models.LastFragmentState
import ru.daotech.agroinvest.utils.*
import ru.daotech.agroinvest.utils.ServiceUtil.Companion.calculateTaskIntervalsSum
import ru.daotech.agroinvest.utils.components.CircleAngleAnimation
import timber.log.Timber
import java.text.ParseException
import java.util.*
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicBoolean
import javax.inject.Inject

const val PERIOD_OF_TIMER_TIME = 1L

class TimerMviFragment :
    BaseMviFragment<TimerContainerViewState, TimerContainerView, TimerContainerViewPresenter>(),
    TimerContainerView {

    private val onTimerLaunch: PublishProcessor<Boolean> = PublishProcessor.create()
    private val onTimerControlClick: PublishSubject<Pair<Boolean, String>> = PublishSubject.create()
    private val onSaveCurrentFragmentState: PublishSubject<LastFragmentState> =
        PublishSubject.create()
    private val onSaveCurrentTimerState: PublishSubject<Boolean> = PublishSubject.create()
    private val onErrorShown: PublishSubject<Unit> = PublishSubject.create()
    private val onDismissTimerState: PublishSubject<Unit> = PublishSubject.create()
    private val onCurrentTrolleySubject: PublishSubject<String> = PublishSubject.create()
    private val errorWrongNfcTypeSubject: PublishSubject<Unit> = PublishSubject.create()
    private val onCurrentRawTaskFinishSubject: PublishSubject<String> = PublishSubject.create()

    private var disposable = CompositeDisposable()

    private var isTimerRun: AtomicBoolean = AtomicBoolean(false)
    private var timerValueHolder: Long = 0L
    private var globalTimerValueHolder: Long = 0L
    private var startTime: Long = 0L
    private var isViewStartedFormat = false
    private lateinit var currentTaskId: String
    private lateinit var taskWithProperties: TaskWithProperties

    @Inject
    lateinit var appStateManager: AppStateManager

    companion object {
        fun newInstance(): TimerMviFragment {
            return TimerMviFragment()
        }
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        retainInstance = true
        EventBus.getDefault().register(this)
    }

    override fun onCreateView(
        inflater: LayoutInflater,
        container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View? {
        return inflater.inflate(
            R.layout.fragment_timer_container,
            container,
            false
        )
    }

    override fun onSaveInstanceState(outState: Bundle) {
        super.onSaveInstanceState(outState)

        outState.putBoolean(KEY_TIMER_STATE, isTimerRun.get())

        if (!isTimerRun.get())
            onSaveCurrentTimerState.onNext(true)

        if (::taskWithProperties.isInitialized) {
            onSaveCurrentFragmentState.onNext(
                LastFragmentState(
                    timerValueHolder,
                    taskWithProperties.id
                )
            )
        }
    }

    override fun saveCurrentFragmentState(): Observable<LastFragmentState> =
        onSaveCurrentFragmentState

    override fun timerControlClick(): Observable<Pair<Boolean, String>> = onTimerControlClick

    override fun errorShown(): Observable<Unit> = onErrorShown

    override fun saveCurrentTimerState(): Observable<Boolean> = onSaveCurrentTimerState

    override fun dismissTimerState(): Observable<Unit> = onDismissTimerState

    override fun onCurrentTrolleyIntent(): Observable<String> = onCurrentTrolleySubject

    override fun errorWrongNfcTypeIntent(): Observable<Unit> = errorWrongNfcTypeSubject

    override fun onViewStateRestored(savedInstanceState: Bundle?) {
        super.onViewStateRestored(savedInstanceState)
        isTimerRun.set(savedInstanceState?.getBoolean(KEY_TIMER_STATE) ?: false)
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)

        buildTimerButtonControls()

        arguments?.getParcelable<TaskWithProperties>(TASK_WITH_PROPERTIES_PARCELABLE)
            ?.let { task ->
                taskWithProperties = task
                isViewStartedFormat = true
                currentTaskId = task.id
            }

        disposable.add(onTimerLaunch
            .doOnNext {
                task_timer_view_started_time.text = DateUtil.getTimeTimerFormat(
                    if (taskWithProperties.startedAt.isNullOrBlank()) {
                        DateUtil.getDateTimeWithDelta(
                            DateTime.now(),
                            appStateManager.getServerTimestamp()
                        )
                    } else {
                        taskWithProperties.startedAt
                    }
                )
                launchTimer()
            }
            .doOnError { Timber.e(it) }
            .subscribe()
        )

        disposable.add(onCurrentRawTaskFinishSubject
            .doOnNext { nfcTaskRowMark ->
                (this.parentFragment as DetailedTaskFragment).initiateTaskFinish(nfcTaskRowMark)
            }
            .doOnError { Timber.e(it) }
            .subscribe()
        )

        onTimerLaunch.onNext(isTimerRun.get())
    }

    private fun launchTimer(): Disposable? {
        startTime = Calendar.getInstance().timeInMillis
        val continueTime: Long = calculateTaskIntervalsSum(
            taskWithProperties,
            appStateManager.getServerTimestamp()
        )

        // Launch timer for standard time calculate. Show standard timer circle UI.
        launchTimerStandardTime(continueTime)

        return Flowable.interval(PERIOD_OF_TIMER_TIME, TimeUnit.SECONDS)
            .filter { isTimerRun.get() }
            .timeInterval()
            .observeOn(AndroidSchedulers.mainThread())
            .subscribe {
                var diff: Long = 0

                try {
                    diff =
                        (Calendar.getInstance().timeInMillis - startTime + continueTime) + timerValueHolder
                    globalTimerValueHolder = diff - continueTime
                } catch (e: ParseException) {
                    e.printStackTrace()
                }

                val diffSeconds = diff / 1000 % 60
                val diffMinutes = diff / (60 * 1000) % 60
                val diffHours = diff / (60 * 60 * 1000) % 24

                task_timer_view_time_text?.let {
                    it.text = (String.format(
                        "%02d:%02d:%02d",
                        diffHours,
                        diffMinutes,
                        diffSeconds
                    ))
                }
            }
    }

    private fun launchTimerStandardTime(continueTime: Long) {
        if (Integer.valueOf(android.os.Build.VERSION.SDK) >= 24) {
            // Task time standard animation launch
            timer_circle_loading.visibility = VISIBLE
            timer_circle_loading?.let { customCircle ->
                val customCircleAnimation = CircleAngleAnimation(
                    customCircle,
                    360,
                    continueTime / 1000,// Started seconds,
                    1800 // Seconds by full circle
                )
                customCircleAnimation.duration = 30 * 60 * 1000
                customCircle.startAnimation(customCircleAnimation)
            }
        } else {
            timer_circle_loading?.let { it.visibility = GONE }
        }
    }

    private fun pauseTimer() {
        this.isTimerRun.set(false)
        task_timer_view_control_button_view.marginStart.plus(16)
        task_timer_view_control_button_view.setImageResource(R.drawable.ic_play_arrow_black_24dp)
        timer_ring_container?.let { layout ->
            layout.animate().alpha(0.5F).withEndAction {
                this.context?.let {
                    layout.setBackgroundDrawable(
                        ContextCompat.getDrawable(
                            it,
                            R.drawable.item_timer_ring_red
                        )
                    )
                    layout.animate().alpha(1F).start()
                }
            }.start()
        }

        setButtonControlDelay()

        onTimerControlClick.onNext(Pair(true, currentTaskId))
    }

    private fun runTimer() {
        timerValueHolder = globalTimerValueHolder
        startTime = Calendar.getInstance().timeInMillis
        this.isTimerRun.set(true)
        task_timer_view_control_button_view.marginStart.minus(16)
        task_timer_view_control_button_view.setImageResource(R.drawable.ic_pause_black_24dp)
        timer_ring_container?.let { layout ->
            layout.animate().alpha(0.5F).withEndAction {
                this.context?.let {
                    layout.setBackgroundDrawable(
                        ContextCompat.getDrawable(
                            it,
                            R.drawable.item_timer_ring_green
                        )
                    )
                    layout.animate().alpha(1F).start()
                }
            }.start()
        }

        onTimerControlClick.onNext(Pair(false, currentTaskId))

        setButtonControlDelay()
    }

    private fun setButtonControlDelay() {
        Observable.just(Unit)
            .subscribeOn(AndroidSchedulers.mainThread())
            .observeOn(AndroidSchedulers.mainThread())
            .filter { task_timer_view_control_button_view != null }
            .map {
                task_timer_view_control_button_view.alpha = .1f
                task_timer_view_control_button_view.isEnabled = false
            }
            .delay(2, TimeUnit.SECONDS)
            .subscribeOn(AndroidSchedulers.mainThread())
            .observeOn(AndroidSchedulers.mainThread())
            .map {
                task_timer_view_control_button_view?.let {
                    it.alpha = 1f
                    it.isEnabled = true
                }
                Unit
            }
            .doOnError { Timber.e(it) }
            .subscribe()
    }

    override fun onStart() {
        super.onStart()
        runTimer()
    }

    override fun render(state: TimerContainerViewState) {
        if (state.errorMessage != null) {
            showSnackbar(state.errorMessage!!, null)
        }

        if (state.isStateChanged) {
            when (state.isTimerRunState) {
                true -> {
                    pauseTimer()
                }
                false -> {
                    runTimer()
                }
            }
            onDismissTimerState.onNext(Unit)
        }
    }

    private fun buildTimerButtonControls() {
        task_timer_view_control_button_view.setOnClickListener {
            when (isTimerRun.get()) {
                true -> {
                    pauseTimer()
                }
                false -> {
                    runTimer()
                }
            }
        }
    }

    override fun onDestroy() {
        disposable.dispose()
        task_timer_view_control_button_view?.let { pauseTimer() }
        EventBus.getDefault().unregister(this)
        super.onDestroy()
    }

    @SuppressLint("DefaultLocale")
    @Subscribe(threadMode = ThreadMode.MAIN_ORDERED)
    fun onNfcTimerScanned(event: NfcTimerScanned) {
        val nfcMessageValueArray = event.message.split('.')
        when {
            nfcMessageValueArray[0].toLowerCase() == TROLLEY_NFC_PREFIX -> {
                (parentFragment as DetailedTaskFragment).showTrolleyFragment(
                    ServiceUtil.cutOffExcessFromNfcMessage(event.message)
                )
            }
            nfcMessageValueArray[0].toLowerCase() == RAW_NFC_PREFIX -> {
                onCurrentRawTaskFinishSubject.onNext(
                    ServiceUtil.cutOffExcessFromNfcMessage(event.message)
                )
            }
            else -> {
                errorWrongNfcTypeSubject.onNext(Unit)
            }
        }
    }
}

data class NfcTimerScanned(val message: String)