package ru.daotech.agroinvest.fragments.earnings

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.functions.Function3
import org.joda.time.DateTime
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.database.DatabaseRepository
import ru.daotech.agroinvest.fragments.ratings.EarningsView
import ru.daotech.agroinvest.models.Worker
import ru.daotech.agroinvest.utils.DateUtil
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

open class EarningsPresenter @Inject constructor(
    private val repository: Repository,
    private val appStateManager: AppStateManager,
    private val databaseRepository: DatabaseRepository
) :
    MviBasePresenter<EarningsView, EarningsViewState>() {

    private var disposable: CompositeDisposable = CompositeDisposable()

    override fun bindIntents() {

        val loadEarningData = intent(EarningsView::loadingIntent)
            .switchMap { loadEarnings() }

        val loadDummy = intent(EarningsView::loadingIntent)
            .switchMap { loadDummy() }

        val errorClearIntent = intent(EarningsView::snackbarDismiss)

        val allIntent = Observable.merge(loadEarningData, loadDummy)
            .onErrorResumeNext { error: Throwable ->
                errorProcessing(error, errorClearIntent)
            }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        disposable.add(
            allIntent
                .subscribe()
        )

        val initialState =
            EarningsViewState(
                worker = null,
                tasksToday = 0,
                tasksYesterday = 0,
                isLoading = false,
                errorMessage = null
            )

        val stateObservable = allIntent.scan(initialState, this::reducer)

        subscribeViewState(stateObservable, EarningsView::render)
    }

    private fun reducer(
        previousState: EarningsViewState,
        partialViewState: EarningsPartialViewState
    ): EarningsViewState {
        return partialViewState.reduce(previousState)
    }


    private fun loadEarnings(): Observable<EarningsPartialViewState> {
        return Observable.combineLatest(
            repository.getCurrentWorker(),
            getTodayTaskCount(),
            getYesterdayTaskCount(),
            Function3<Worker, Int, Int, Triple<Worker, Int, Int>> { worker, tasksToday, tasksYesterday ->
                Triple(worker, tasksToday, tasksYesterday)
            }
        )
            .observeOn(AndroidSchedulers.mainThread())
            .map { (worker, tasksToday, tasksYesterday) ->
                EarningsPartialViewState.SuccessLoadEarnings(
                    worker = worker,
                    tasksToday = tasksToday,
                    tasksYesterday = tasksYesterday
                ) as EarningsPartialViewState
            }
            .startWith(EarningsPartialViewState.Loading as EarningsPartialViewState)
    }

    private fun getYesterdayTaskCount(): Observable<Int> {
        val sd = DateUtil.getDateTimeWithDelta(
            DateTime
                .now()
                .minusDays(1)
                .withTime(
                    0,
                    0,
                    0,
                    0
                ),
            appStateManager.getServerTimestamp()
        )

        val fd = DateUtil.getDateTimeWithDelta(
            DateTime
                .now()
                .minusDays(1)
                .withTime(
                    23,
                    59,
                    59,
                    999
                ),
            appStateManager.getServerTimestamp()
        )

        return appStateManager.getCurrentWorker()
            .flatMap { worker ->
                databaseRepository.getTaskCount(
                    namespace = appStateManager.getSessionId(),
                    workerId = worker.id,
                    startPeriod = sd,
                    finishPeriod = fd
                )
            }
    }

    private fun getTodayTaskCount(): Observable<Int> {
        val sd = DateUtil.getDateTimeWithDelta(
            DateTime
                .now()
                .withTime(
                    0,
                    0,
                    0,
                    0
                ),
            appStateManager.getServerTimestamp()
        )

        val fd = DateUtil.getDateTimeWithDelta(
            DateTime
                .now()
                .withTime(
                    23,
                    59,
                    59,
                    999
                ),
            appStateManager.getServerTimestamp()
        )

        return appStateManager.getCurrentWorker()
            .flatMap { worker ->
                databaseRepository.getTaskCount(
                    namespace = appStateManager.getSessionId(),
                    workerId = worker.id,
                    startPeriod = sd,
                    finishPeriod = fd
                )
            }
    }

    private fun loadDummy(): Observable<EarningsPartialViewState> {
        return Observable.just(EarningsPartialViewState.Dummy as EarningsPartialViewState)
    }

    override fun unbindIntents() {
        disposable.dispose()
    }

    private fun errorProcessing(
        error: Throwable,
        errorClearIntent: Observable<Unit>
    ): Observable<EarningsPartialViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .mergeWith(errorClearIntent)
            .take(1)
            .map {
                EarningsPartialViewState.ClearError as EarningsPartialViewState
            }
            .startWith(EarningsPartialViewState.Error(errorMessage = error.localizedMessage))
    }
}