package ru.daotech.agroinvest.fragments.login

import ru.daotech.agroinvest.database.entities.RecentUser
import java.io.File

sealed class LoginPartialViewState {

    abstract fun reduce(previousState: LoginViewState): LoginViewState

    object EmptyState : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState
        }
    }

    object UpdateAppLoading : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState.copy(
                isUpdateAppLoading = true,
                isLoginDisable = false,
                progressButtonLabel = "Загрузка",
                errorMessage = null
            )
        }
    }

    data class EnableStandardLoginState(val appInfoValue: String) : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState.copy(
                isUpdateAppEnable = false,
                isLoginDisable = false,
                isUpdateAppLoading = false,
                appFile = null,
                errorMessage = null,
                appInfoValue = appInfoValue
            )
        }
    }

    data class EnableUpdateAppState(val appInfoValue: String) : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState.copy(
                isUpdateAppEnable = true,
                isUpdateAppLoading = false,
                isLoginDisable = false,
                errorMessage = null,
                appInfoValue = appInfoValue
            )
        }
    }

    data class DisableLoginAppState(val appInfoValue: String) : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState.copy(
                isUpdateAppEnable = true,
                isLoginDisable = true,
                isUpdateAppLoading = false,
                errorMessage = null,
                appInfoValue = appInfoValue
            )
        }
    }

    object Loading : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState.copy(
                isLoading = true,
                isFailedAttempt = false,
                isLoginDisable = false,
                errorMessage = null
            )
        }
    }

    object DisableLoading : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState.copy(
                isLoading = false,
                isFailedAttempt = false,
                isLoginDisable = false,
                errorMessage = null
            )
        }
    }

    object GoToScan : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState
        }
    }

    object Success : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState.copy(
                isLoading = false,
                isFailedAttempt = false,
                isLoginDisable = false,
                isSuccess = true
            )
        }

    }

    object FailedTry : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            val errorMessage = if (previousState.failedAttempts.inc() >= 3) {
                "Слишком много попыток. Информация отправлена администратору системы."
            } else {
                "Ошибка авторизации. Повторите попытку."
            }
            return previousState.copy(
                isLoading = false,
                errorMessage = errorMessage,
                failedAttempts = previousState.failedAttempts.inc(),
                isLoginDisable = false,
                isFailedAttempt = true
            )
        }

    }

    object ClearError : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState.copy(
                errorMessage = null
            )
        }
    }

    data class ErrorNotRelatedToTry(val errorMessage: String) : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState.copy(
                isLoading = false,
                errorMessage = errorMessage,
                isLoginDisable = false,
                isFailedAttempt = false
            )
        }
    }

    data class QueueCountState(val queueResultMessage: String?) : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState.copy(
                queueResultMessage = queueResultMessage,
                isLoginDisable = false,
                errorMessage = null
            )
        }
    }

    data class StartUpdateInstallingState(val appFile: File) : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState.copy(
                isUpdateAppEnable = true,
                isUpdateAppLoading = true,
                isLoginDisable = false,
                appFile = appFile,
                progressButtonLabel = "Установка",
                errorMessage = null
            )
        }
    }

    data class SetRecentUsers(val users: List<RecentUser>) : LoginPartialViewState() {
        override fun reduce(previousState: LoginViewState): LoginViewState {
            return previousState.copy(
                    recentUsers = users
            )
        }
    }
}