package ru.daotech.agroinvest.fragments.login

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.BuildConfig
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.exceptions.SignInWorkerException
import ru.daotech.agroinvest.models.AppInfo
import ru.daotech.agroinvest.service.events.ServerData
import ru.daotech.agroinvest.service.jsonmodels.AppInfoPayload
import timber.log.Timber
import java.net.UnknownHostException
import java.util.concurrent.TimeUnit
import java.util.concurrent.TimeoutException
import javax.inject.Inject

class LoginPresenter @Inject constructor(
    private val repository: Repository,
    private var onSuccessLogin: (() -> Unit)?,
    private var goToScan: (() -> Unit)?,
    private val appStateManager: AppStateManager
) :
    MviBasePresenter<LoginView, LoginViewState>() {

    private val disposable = CompositeDisposable()

    override fun bindIntents() {

        val loadingIntent = intent(LoginView::loadingIntent)
            .map {
                Timber.d("SOCKET SERVICE: CHANGE SESSION ID")
                appStateManager.setSessionId() // Set new session ID
                appStateManager.setIsFirstLaunch(true)
            }
            .flatMap { repository.loadCurrentAppStatusByUrl() }
            .map { serverData ->
                // Check opportunity for application updating
                val isUpdateDefine =
                    (serverData.versionCode != BuildConfig.VERSION_CODE &&
                            serverData.versionCode > BuildConfig.VERSION_CODE)

                appStateManager.saveActualAppInfo(ServerData(serverData))

                when {
                    (serverData.updateRequired && isUpdateDefine) -> {
                        return@map LoginPartialViewState.DisableLoginAppState(
                            "${BuildConfig.VERSION_NAME} ${BuildConfig.BUILD_TYPE}"
                        )
                    }
                    (!serverData.updateRequired && isUpdateDefine) -> {
                        return@map LoginPartialViewState.EnableUpdateAppState(
                            "${BuildConfig.VERSION_NAME} ${BuildConfig.BUILD_TYPE}"
                        )
                    }
                    else -> {
                        return@map LoginPartialViewState.EnableStandardLoginState(
                            "${BuildConfig.VERSION_NAME} ${BuildConfig.BUILD_TYPE}"
                        )
                    }
                }
            }
            .onErrorResumeNext { error: Throwable ->
                Observable.just(
                    LoginPartialViewState.EnableStandardLoginState(
                        "${BuildConfig.VERSION_NAME} ${BuildConfig.BUILD_TYPE}"
                    )
                )
            }
            .cast(LoginPartialViewState::class.java)
            .observeOn(AndroidSchedulers.mainThread())

        val barcodeRetrieved = intent(LoginView::barcodeRetrieved)
            .switchMap { barcode -> login(barcode) }

        val onUpdateAppInitiateIntent = intent(LoginView::onUpdateAppInitiateIntent)
            .switchMap { fileFolder -> updateApp(fileFolder) }

        val scanButtonClick =
            intent(LoginView::scanButtonClick).map { LoginPartialViewState.GoToScan }

        val loadRecentUsersIntent = intent(LoginView::loadRecentUsersIntent)
            .flatMap { repository.getRecentUsers() }
            .observeOn(AndroidSchedulers.mainThread())
            .map { list -> LoginPartialViewState.SetRecentUsers(list) }

        val recentUsersRetrieved = intent(LoginView::recentUsersRetrieved)
            .map { LoginPartialViewState.SetRecentUsers(emptyList()) }

        val queueCountUpdateIntent = appStateManager.queueCountEventSubject
            .observeOn(AndroidSchedulers.mainThread())
            .map { queueResultMessage ->
                LoginPartialViewState.QueueCountState(queueResultMessage) as LoginPartialViewState
            }

        val errorShown = intent(LoginView::errorShown)
            .map { LoginPartialViewState.ClearError }

        val intentList = listOf(
            barcodeRetrieved,
            scanButtonClick,
            errorShown,
            loadingIntent,
            queueCountUpdateIntent,
            onUpdateAppInitiateIntent,
            loadRecentUsersIntent,
            recentUsersRetrieved
        )

        val allIntent = Observable
            .merge(intentList)
            .onErrorResumeNext { error: Throwable ->
                errorProcessing(error)
            }
            .share()

        disposable.add(
            allIntent.filter { state ->
                state is LoginPartialViewState.Success
            }.subscribe(
                {
                    // Initiate connect to couch database
                    appStateManager.onInitiateConnectToCouchDb.onNext(Unit)
                    onSuccessLogin?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        disposable.add(
            allIntent.filter { state ->
                state is LoginPartialViewState.GoToScan
            }.subscribe(
                {
                    goToScan?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        val initialState =
            LoginViewState(
                isLoading = false,
                failedAttempts = 0,
                isFailedAttempt = false,
                errorMessage = null,
                isSuccess = false,
                queueResultMessage = null,
                isUpdateAppLoading = false,
                isUpdateAppEnable = false,
                isLoginDisable = false,
                progressButtonLabel = null,
                appInfoValue = null,
                appFile = null,
                recentUsers = emptyList()
            )
        val stateObservable = allIntent.scan(initialState, this::reducer)

        subscribeViewState(stateObservable, LoginView::render)
    }

    private fun reducer(
        previousState: LoginViewState,
        partialViewState: LoginPartialViewState
    ): LoginViewState {
        return partialViewState.reduce(previousState)
    }

    private fun errorProcessing(
        error: Throwable
    ): Observable<LoginPartialViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .take(1)
            .map { LoginPartialViewState.ClearError as LoginPartialViewState }
    }

    private fun updateApp(fileFolder: String): Observable<LoginPartialViewState> {
        val appInfo = getAppInfoFromPref()
        return repository.downloadAppUrl(appInfo.appUrl, fileFolder, appInfo.identifier)
//            .subscribeOn(AndroidSchedulers.mainThread())
            .map { LoginPartialViewState.StartUpdateInstallingState(it) as LoginPartialViewState }
            .startWith(LoginPartialViewState.UpdateAppLoading as LoginPartialViewState)
            .doOnError { Timber.e(it.localizedMessage) }
            .onErrorReturn {
                Timber.e(it)
                LoginPartialViewState.ErrorNotRelatedToTry(
                    "Ошибка прав доступа. Перезапустите приложение и повторите попытку."
                )
            }
            .observeOn(AndroidSchedulers.mainThread())
    }

    private fun login(barcode: String): Observable<LoginPartialViewState> {
        // Save barcode to local storage for continued usage
        appStateManager.saveCurrentWorkerCode(barcode);

        return repository.signIn(barcode)
            .observeOn(AndroidSchedulers.mainThread())
            .map { LoginPartialViewState.Success as LoginPartialViewState }
            .startWith(LoginPartialViewState.Loading as LoginPartialViewState)
            .doOnError { Timber.e(it) }
            .onErrorReturn { error ->
                Timber.e(error)
                when (error) {
                    is SignInWorkerException ->
                        LoginPartialViewState.ErrorNotRelatedToTry(error.message!!)
                    is java.lang.IllegalArgumentException ->
                        LoginPartialViewState.FailedTry
                    is UnknownHostException ->
                        LoginPartialViewState.ErrorNotRelatedToTry("Отсутсвует интернет соединение. Проверьте подключение и повторите попытку.")
                    is TimeoutException ->
                        LoginPartialViewState.ErrorNotRelatedToTry("Отсутсвует интернет соединение. Проверьте подключение и повторите попытку.")
                    else ->
                        LoginPartialViewState.ErrorNotRelatedToTry("Произошла ошибка. Повторите попытку.")
                }
            }
            .observeOn(AndroidSchedulers.mainThread())
    }

    private fun getAppInfoFromPref(): AppInfo {
        return AppInfo(
            from = AppInfoPayload(
                versionName = appStateManager.getActualAppInfo()!!.serverDataPayload.versionName,
                appUrl = appStateManager.getActualAppInfo()!!.serverDataPayload.versionUrl,
                versionCode = appStateManager.getActualAppInfo()!!.serverDataPayload.versionName
            )
        )
    }

    override fun unbindIntents() {
        disposable.dispose()
        onSuccessLogin = null
        goToScan = null
    }
}