package ru.daotech.agroinvest.fragments.pages.about

import android.content.Context
import android.content.pm.PackageManager
import android.os.Bundle
import android.view.*
import androidx.core.app.ActivityCompat
import androidx.core.content.ContextCompat
import com.hannesdorfmann.mosby3.mvi.MviFragment
import com.jakewharton.rxbinding2.view.clicks
import dagger.android.support.AndroidSupportInjection
import io.reactivex.Observable
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.subjects.PublishSubject
import kotlinx.android.synthetic.main.fragment_about_page.*
import kotlinx.android.synthetic.main.fragment_about_page.view.*
import ru.daotech.agroinvest.BuildConfig
import ru.daotech.agroinvest.R
import ru.daotech.agroinvest.container.AppBarController
import ru.daotech.agroinvest.container.ContainerActivity
import ru.daotech.agroinvest.di.modules.DATABASE_NAME
import ru.daotech.agroinvest.dialog.builder.AboutDialogBuilder
import ru.daotech.agroinvest.utils.EXTERNAL_WRITE_PERMISSION_CODE
import ru.daotech.agroinvest.utils.ServiceUtil
import ru.daotech.agroinvest.utils.ViewUtil
import timber.log.Timber
import java.io.File
import javax.inject.Inject

class AboutPageFragment : MviFragment<AboutPageView, AboutPagePresenter>(),
    AboutPageView {

    @Inject
    lateinit var aboutPagePresenter: AboutPagePresenter

    @Inject
    lateinit var aboutDialogBuilder: AboutDialogBuilder

    private val onFileUpload: PublishSubject<List<File>> = PublishSubject.create()
    private val onPermissionToExportGranted: PublishSubject<Unit> = PublishSubject.create()

    private var disposable = CompositeDisposable()

    override fun onAttach(context: Context) {
        AndroidSupportInjection.inject(this)
        super.onAttach(context)
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        setHasOptionsMenu(true)
        retainInstance = true
    }

    override fun onPrepareOptionsMenu(menu: Menu) {
        super.onPrepareOptionsMenu(menu)
        (menu.findItem(R.id.menu_search) as MenuItem).isVisible = false

        activity?.let { activity ->
            ViewUtil.manageInternetConnectState(activity, menu)
            ViewUtil.manageSocketConnectState(activity, menu)
        }
    }

    override fun onCreateView(
        inflater: LayoutInflater,
        container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View? {
        (activity as? AppBarController)?.apply {
            setupAppBarWithNavController()
            showAppBar()
        }

        val view = inflater.inflate(R.layout.fragment_about_page, container, false)
        view.textVersion.text = BuildConfig.VERSION_NAME
        view.textBuildType.text = BuildConfig.BUILD_TYPE

        return view
    }

    override fun fileUpload(): Observable<List<File>> = onFileUpload

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        (activity as? AppBarController)?.showAppBar()

        (this.activity as ContainerActivity).actionBar?.setDisplayHomeAsUpEnabled(false)
        (this.activity as ContainerActivity).showToggle()

        aboutDialogBuilder.buildResponseAttentionDialog(requireContext())

        disposable.add(
            onPermissionToExportGranted
                .map {
                    return@map ServiceUtil.exportDB(
                        requireActivity().getDatabasePath(
                            DATABASE_NAME
                        ).parent
                    )
                }
                .map { resultedFile -> onFileUpload.onNext(resultedFile) }
                .doOnError { error ->
                    aboutDialogBuilder.showResultDialog(false)
                    Timber.e(error)
                }
                .subscribe()
        )

        export_db_button
            .clicks()
            .map { checkPermissions() }
            .subscribe()

        super.onViewCreated(view, savedInstanceState)
    }

    private fun checkPermissions() {
        // Check permission of write external storage for next correct database export
        if (ContextCompat.checkSelfPermission(
                requireContext(),
                android.Manifest.permission.WRITE_EXTERNAL_STORAGE
            ) == PackageManager.PERMISSION_DENIED
        ) {
            ActivityCompat.requestPermissions(
                requireActivity(),
                arrayOf(android.Manifest.permission.WRITE_EXTERNAL_STORAGE),
                EXTERNAL_WRITE_PERMISSION_CODE
            )
        } else {
            onPermissionToExportGranted.onNext(Unit)
        }
    }

    override fun onRequestPermissionsResult(
        requestCode: Int,
        permissions: Array<out String>,
        grantResults: IntArray
    ) {
        super.onRequestPermissionsResult(requestCode, permissions, grantResults)

        if (requestCode == EXTERNAL_WRITE_PERMISSION_CODE) {
            if (
                grantResults.isNotEmpty()
                && grantResults[0] == PackageManager.PERMISSION_GRANTED
            ) {
                try {
                    onPermissionToExportGranted.onNext(Unit)
                } catch (ex: SecurityException) {
                    Timber.e(ex)
                }
            }
        }
    }

    override fun render(state: AboutPageViewState) {
        manageErrors(state)
    }

    // On view handling errors
    private fun manageErrors(state: AboutPageViewState) {
        when (state) {
            is Error -> {
                aboutDialogBuilder.showResultDialog(false)
            }
            is AboutPageViewState.UploadFailed -> {
                aboutDialogBuilder.showResultDialog(false)
            }
            is AboutPageViewState.UploadSuccess -> {
                aboutDialogBuilder.showResultDialog(true)
            }
            else -> {
            }
        }
    }

    override fun onDestroy() {
        disposable.clear()
        super.onDestroy()
    }

    override fun loadingIntent(): Observable<Unit> = Observable.just(Unit)

    override fun createPresenter(): AboutPagePresenter {
        return aboutPagePresenter
    }
}