package ru.daotech.agroinvest.fragments.pages.accident

import android.content.Context
import android.os.Bundle
import android.view.*
import android.widget.ArrayAdapter
import android.widget.ListView
import android.widget.ScrollView
import androidx.appcompat.app.AlertDialog
import androidx.core.content.ContextCompat
import androidx.drawerlayout.widget.DrawerLayout
import com.google.android.material.snackbar.Snackbar
import com.hannesdorfmann.mosby3.mvi.MviFragment
import com.jakewharton.rxbinding2.view.clicks
import dagger.android.support.AndroidSupportInjection
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.subjects.PublishSubject
import kotlinx.android.synthetic.main.activity_container.*
import kotlinx.android.synthetic.main.fragment_accident_report.*
import kotlinx.android.synthetic.main.fragment_accident_report.view.*
import ru.daotech.agroinvest.R
import ru.daotech.agroinvest.container.AppBarController
import ru.daotech.agroinvest.container.ContainerActivity
import ru.daotech.agroinvest.dialog.builder.AccidentReportDialogBuilder
import ru.daotech.agroinvest.fragments.pages.accident.AccidentReportFragmentArgs.fromBundle
import ru.daotech.agroinvest.utils.ACCIDENT_FAILED_RESULT_STATE
import ru.daotech.agroinvest.utils.ACCIDENT_SEND_CONTINUATION_RESULT_STATE
import ru.daotech.agroinvest.utils.ACCIDENT_SUCCESS_RESULT_STATE
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

class AccidentReportFragment :
    MviFragment<AccidentReportView, AccidentReportPresenter>(),
    AccidentReportView {

    @Inject
    lateinit var accidentReportPresenter: AccidentReportPresenter

    @Inject
    lateinit var accidentDialogBuilder: AccidentReportDialogBuilder

    protected lateinit var snackbar: Snackbar

    private var disposable = CompositeDisposable()
    private val onLoadInitialIntent: PublishSubject<String> = PublishSubject.create()
    private val onSendResultReport: PublishSubject<String> = PublishSubject.create()
    private val onSendResultReportWithTaskId: PublishSubject<Pair<String, String>> =
        PublishSubject.create()

    private lateinit var taskAccidentReportDialog: AlertDialog

    private lateinit var templateListView: ListView

    override fun onAttach(context: Context) {
        AndroidSupportInjection.inject(this)
        super.onAttach(context)
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        this.requireActivity().drawer_layout?.setDrawerLockMode(DrawerLayout.LOCK_MODE_LOCKED_CLOSED)
        setHasOptionsMenu(true)

        retainInstance = true

        taskAccidentReportDialog = accidentDialogBuilder.buildAccidentRequestDialog(
            requireContext(),
            onSendResultReport,
            onSendResultReportWithTaskId,
            null,
            null
        )
        (this.activity as ContainerActivity).actionBar?.setDisplayShowHomeEnabled(true)
        (this.activity as ContainerActivity).actionBar?.setDisplayHomeAsUpEnabled(true)
    }

    override fun onCreateView(
        inflater: LayoutInflater,
        container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View? {
        (activity as? AppBarController)?.apply {
            setupAppBarWithNavController()
            showAppBar()
        }
        val view = inflater.inflate(
            R.layout.fragment_accident_report,
            container,
            false
        )

        val templates =
            arrayOf(
                getString(R.string.accident_tmplate_task_list),
                getString(R.string.accident_tmplate_worker_day_finish),
                getString(R.string.accident_tmplate_task_start),
                getString(R.string.accident_tmplate_task_finish),
                getString(R.string.accident_tmplate_task_status),
                getString(R.string.accident_tmplate_task_info_showing),
                getString(R.string.accident_tmplate_task_internet_connection),
                getString(R.string.accident_tmplate_nfc_scan)
            )

        val arrayAdapter: ArrayAdapter<String> = ArrayAdapter(
            requireContext(),
            R.layout.item_list_view_accident_template,
            R.id.accident_template_view,
            templates
        )

        templateListView = view.problem_templates_list!!
        templateListView.adapter = arrayAdapter

        templateListView.setOnItemClickListener { _, _, _, _ ->
            accidentDialogBuilder.buildAccidentResponseDialog(requireContext())

            val selectedItemPosition = templateListView.checkedItemPosition
            if (selectedItemPosition >= 0) {
                getTaskIdFromArguments()?.let { taskId ->
                    onSendResultReportWithTaskId.onNext(
                        Pair(
                            taskId,
                            templateListView.adapter.getItem(selectedItemPosition) as String
                        )
                    )
                    return@setOnItemClickListener
                }
                onSendResultReport.onNext(
                    templateListView.adapter.getItem(selectedItemPosition) as String
                )
            } else {
                showSnackbar(getString(R.string.attention_wrong_send_accident_text))
            }
        }

        setListViewHeightBasedOnChildren(templateListView)

        return view
    }

    private fun setListViewHeightBasedOnChildren(listView: ListView) {
        val listAdapter = listView.adapter
            ?: return

        var totalHeight = 0
        var i = 0
        while (i < listAdapter.count) {
            val listItem = listAdapter.getView(i, null, listView)
            listItem.measure(0, 0)
            totalHeight += listItem.measuredHeight
            i++
        }

        val params = listView.layoutParams as ViewGroup.LayoutParams
        params.height = totalHeight + (listView.dividerHeight * (listAdapter.count - 1))
        listView.layoutParams = params
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)

        open_accident_dialog
            .clicks()
            .map { taskAccidentReportDialog.show() }
            .subscribe()

        Observable.interval(2, TimeUnit.SECONDS)
            .filter { this.isVisible }
            .take(1)
            .observeOn(AndroidSchedulers.mainThread())
            .map {
                main_accident_scroll_container?.let {
                    main_accident_scroll_container.scrollTo(0, 0)
                    main_accident_scroll_container.clearFocus()
                    own_accident_report.requestFocus()
                    main_accident_scroll_container.fullScroll(ScrollView.FOCUS_UP)
                }
            }
            .doOnError { Timber.i(it) }
            .subscribe()
    }

    override fun onPrepareOptionsMenu(menu: Menu) {
        super.onPrepareOptionsMenu(menu)
        (menu.findItem(R.id.menu_search) as MenuItem).setVisible(false)
        (menu.findItem(R.id.menu_task_accident_report) as MenuItem).setVisible(false)
    }

    override fun render(state: AccidentReportViewState) {
        manageErrors(state)

        if (state is AccidentReportViewState.TaskLoaded) {
            taskAccidentReportDialog = accidentDialogBuilder.buildAccidentRequestDialog(
                requireContext(),
                onSendResultReport,
                onSendResultReportWithTaskId,
                state.taskWithProperties,
                state.currentWorkerSimple
            )
        } else if (state is AccidentReportViewState.ErrorWithMessage) {
            state.message
        }
    }

    private fun manageErrors(state: AccidentReportViewState) {
        when (state) {
            is Error -> {
                accidentDialogBuilder.showResultDialog(ACCIDENT_FAILED_RESULT_STATE)
            }
            is AccidentReportViewState.SendFailed -> {
                accidentDialogBuilder.showResultDialog(ACCIDENT_FAILED_RESULT_STATE)
            }
            is AccidentReportViewState.SendContinuationState -> {
                accidentDialogBuilder.showResultDialog(ACCIDENT_SEND_CONTINUATION_RESULT_STATE)
            }
            is AccidentReportViewState.ErrorWithMessage -> {
                Snackbar.make(requireView(), state.message, Snackbar.LENGTH_LONG).show()
            }
            is AccidentReportViewState.SendSuccess -> {
                accidentDialogBuilder.showResultDialog(ACCIDENT_SUCCESS_RESULT_STATE)
            }
            else -> {
            }
        }
    }

    override fun onDestroy() {
        disposable.clear()
        super.onDestroy()
    }

    override fun loadingTaskInitialIntent(): Observable<String> = onLoadInitialIntent

    override fun sendResultReport(): Observable<String> = onSendResultReport

    override fun sendResultReportWithTaskId(): Observable<Pair<String, String>> =
        onSendResultReportWithTaskId

    override fun loadingIntent(): Observable<Unit> = Observable.just(Unit)

    override fun createPresenter(): AccidentReportPresenter {
        return accidentReportPresenter
    }

    override fun onResume() {
        if (arguments?.size()!! > 0) {
            arguments?.let { bundle ->
                val args = fromBundle(bundle)
                args.taskId?.let { taskId ->
                    onLoadInitialIntent.onNext(taskId)

                }
            }
        }
        super.onResume()
    }

    private fun getTaskIdFromArguments(): String? {
        if (arguments?.size()!! > 0) {
            arguments?.let { bundle ->
                val args =
                    fromBundle(bundle)
                args.taskId?.let { taskId ->
                    return taskId
                }
            }
        }
        return null
    }

    private fun showSnackbar(message: String) {
        snackbar = Snackbar.make(
            requireActivity().findViewById(android.R.id.content)!!,
            message,
            Snackbar.LENGTH_SHORT
        )
        snackbar.view.setBackgroundColor(
            ContextCompat.getColor(
                requireContext(),
                R.color.accentSnackbarBackground
            )
        )
        snackbar.show()
    }
}