package ru.daotech.agroinvest.fragments.pages.accident

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.Disposable
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.utils.DateUtil
import timber.log.Timber
import java.net.UnknownHostException
import java.util.concurrent.TimeUnit

class AccidentReportPresenter(
    private val appStateManager: AppStateManager,
    private var onAccidentReportFinished: ((Boolean) -> Unit)?,
    private val repository: Repository
) : MviBasePresenter<AccidentReportView, AccidentReportViewState>() {

    private lateinit var disposable: Disposable

    override fun bindIntents() {
        val loadingIntent = intent(AccidentReportView::loadingIntent)
            .flatMap { appStateManager.isWorkerSet() }
            .map { currentWorkerExists ->
                AccidentReportViewState.Success(
                    currentWorkerExists
                ) as AccidentReportViewState
            }
            .onErrorReturn {
                Timber.e(it)
                AccidentReportViewState.Error
            }
            .startWith(AccidentReportViewState.Loading)

        val loadingTaskInitialIntent = intent(AccidentReportView::loadingTaskInitialIntent)
            .flatMap { taskId -> repository.getTaskWithProperties(taskId) }
            .map { taskWithProperties ->
                AccidentReportViewState.TaskLoaded(
                    taskWithProperties,
                    appStateManager.getCurrentWorkerSimple()
                ) as AccidentReportViewState
            }
            .onErrorReturn {
                Timber.e(it)
                AccidentReportViewState.Error
            }
            .startWith(AccidentReportViewState.Loading)

        val sendResultReport = intent(AccidentReportView::sendResultReport)
            .flatMap { report ->
                Observable.just(Unit)
                    .flatMap {
                        val currentTime =
                            DateUtil.getCurrentDateTimestampWithDelta(appStateManager.getServerTimestamp())
                        val lastAccidentTime = appStateManager.getLastAccidentTime()
                        val lastAccidentMessage = appStateManager.getLastAccidentMessage()

                        val delta =
                            if (lastAccidentTime == null || lastAccidentTime.isNullOrBlank()) {
                                null
                            } else {
                                DateUtil.getMillsBetweenTwoDates(currentTime, lastAccidentTime)
                            }

                        when {
                            delta != null && delta <= TimeUnit.MINUTES.toMillis(2L) -> {
                                Observable.just(
                                    AccidentReportViewState.ErrorWithMessage(
                                        "Повторное сообщение может быть отправлено через 2 минуты."
                                    )
                                )
                            }
                            isLastAccidentReportMessageEqualsCurrentMessage(
                                report,
                                lastAccidentMessage
                            ) -> {
                                Observable.just(
                                    AccidentReportViewState.ErrorWithMessage(
                                        "Данное сообщение уже отправлено ранее."
                                    )
                                )
                            }
                            else -> {
                                repository.sendAccidentReport(report)
                                    .map { appStateManager.saveLastAccidentTime(currentTime) }
                                    .map { appStateManager.saveLastAccidentMessage(report) }
                                    .observeOn(AndroidSchedulers.mainThread())
                                    .map { AccidentReportViewState.SendSuccess as AccidentReportViewState }
                                    .onErrorResumeNext { _: Throwable ->
                                        Observable.just(AccidentReportViewState.SendFailed)
                                            .cast(AccidentReportViewState::class.java)
                                    }
                            }
                        }
                    }
                    .onErrorResumeNext { error: Throwable ->
                        return@onErrorResumeNext if (error is UnknownHostException) {
                            Observable.just(AccidentReportViewState.SendContinuationState)
                                .doOnNext {
                                    repository.signalAccidentReportToUpload(report, null)
                                        .subscribe()
                                }
                                .observeOn(AndroidSchedulers.mainThread())
                                .map {
                                    AccidentReportViewState.SendContinuationState
                                            as AccidentReportViewState
                                }
                        } else {
                            Observable.just(AccidentReportViewState.SendFailed)
                                .doOnNext {
                                    repository.signalAccidentReportToUpload(report, null)
                                        .subscribe()
                                }
                                .observeOn(AndroidSchedulers.mainThread())
                                .map { AccidentReportViewState.SendFailed as AccidentReportViewState }
                        }
                    }
            }
            .startWith(AccidentReportViewState.Loading)

        val sendResultReportWithTaskId = intent(AccidentReportView::sendResultReportWithTaskId)
            .switchMap { reportValuePair ->
                val currentTime =
                    DateUtil.getCurrentDateTimestampWithDelta(appStateManager.getServerTimestamp())
                val lastAccidentTime = appStateManager.getLastAccidentTime()
                val lastAccidentMessage = appStateManager.getLastAccidentMessage()

                val delta = lastAccidentTime?.let {
                    DateUtil.getMillsBetweenTwoDates(
                        it,
                        currentTime
                    )
                }

                repository.getTaskWithProperties(reportValuePair.first)
                    .flatMap { taskWithProperties ->

                        when {
                            delta == null || delta <= TimeUnit.MINUTES.toMillis(2L) -> {
                                Observable.just(
                                    AccidentReportViewState.ErrorWithMessage(
                                        "Повторное сообщение может быть отправлено через 2 минуты."
                                    )
                                )
                            }
                            isLastAccidentReportMessageEqualsCurrentMessage(
                                reportValuePair.second,
                                lastAccidentMessage
                            ) -> {
                                Observable.just(
                                    AccidentReportViewState.ErrorWithMessage(
                                        "Данное сообщение уже отправлено ранее."
                                    )
                                )
                            }
                            else -> {
                                repository.sendAccidentComplexReport(
                                    taskWithProperties,
                                    reportValuePair.second
                                )
                                    .map { appStateManager.saveLastAccidentMessage(reportValuePair.second) }
                                    .map { appStateManager.saveLastAccidentTime(currentTime) }
                                    .map { AccidentReportViewState.SendSuccess as AccidentReportViewState }
                                    .onErrorResumeNext { error: Throwable ->
                                        Observable.just(AccidentReportViewState.SendFailed)
                                            .cast(AccidentReportViewState::class.java)
                                    }
                            }
                        }
                    }
                    .onErrorResumeNext { error: Throwable ->
                        Observable.just(AccidentReportViewState.SendContinuationState)
                            .doOnNext {
                                repository.signalAccidentReportToUpload(
                                    reportValuePair.second,
                                    reportValuePair.first
                                )
                                    .subscribe()
                            }
                            .observeOn(AndroidSchedulers.mainThread())
                            .map { AccidentReportViewState.SendContinuationState as AccidentReportViewState }
                    }
            }
            .startWith(AccidentReportViewState.Loading)


        val allIntent = Observable.merge(
            loadingIntent,
            sendResultReport,
            loadingTaskInitialIntent,
            sendResultReportWithTaskId
        )
            .onErrorResumeNext { error: Throwable ->
                errorProcessing(error)
            }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        disposable = loadingIntent
            .filter { state ->
                state is AccidentReportViewState.Success
            }
            .map { state ->
                state as AccidentReportViewState.Success
            }
            .observeOn(AndroidSchedulers.mainThread())
            .subscribe(
                {

                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )

        subscribeViewState(allIntent, AccidentReportView::render)
    }

    private fun errorProcessing(
        error: Throwable
    ): Observable<AccidentReportViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .take(1)
            .map {
                AccidentReportViewState.Error as AccidentReportViewState
            }
    }

    private fun isLastAccidentReportMessageEqualsCurrentMessage(
        currentMessage: String,
        lastMessage: String?
    ): Boolean =
        lastMessage != null && (currentMessage == lastMessage)

    override fun unbindIntents() {
        disposable.dispose()
        onAccidentReportFinished = null
    }
}