package ru.daotech.agroinvest.fragments.pages.update

import android.Manifest
import android.content.Context
import android.content.Intent
import android.content.pm.PackageManager
import android.net.Uri
import android.os.Build
import android.os.Bundle
import android.os.Environment
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.webkit.MimeTypeMap
import androidx.core.content.ContextCompat
import androidx.core.content.FileProvider.getUriForFile
import androidx.drawerlayout.widget.DrawerLayout
import com.google.android.material.snackbar.Snackbar
import com.hannesdorfmann.mosby3.mvi.MviFragment
import com.jakewharton.rxbinding2.view.clicks
import dagger.android.support.AndroidSupportInjection
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.subjects.BehaviorSubject
import io.reactivex.subjects.PublishSubject
import kotlinx.android.synthetic.main.activity_container.*
import kotlinx.android.synthetic.main.fragment_update_app.*
import ru.daotech.agroinvest.*
import ru.daotech.agroinvest.container.AppBarController
import timber.log.Timber
import javax.inject.Inject

class UpdateAppFragment :
    MviFragment<UpdateAppView, UpdateAppPresenter>(),
    UpdateAppView {

    @Inject
    lateinit var updateAppPresenter: UpdateAppPresenter

    @Inject
    lateinit var appStateManager: AppStateManager

    private val clearErrorSubject = PublishSubject.create<Unit>()

    private val appFolderInit: BehaviorSubject<String> = BehaviorSubject.create()

    private lateinit var downloadFolder: String

    protected lateinit var snackbar: Snackbar

    private val compositeDisposable = CompositeDisposable()

    override fun onAttach(context: Context) {
        AndroidSupportInjection.inject(this)
        super.onAttach(context)
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)
        downloadFolder =
            Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_DOWNLOADS)
                .absolutePath

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N) {
            appFolderInit.onNext(requireContext().filesDir.toString())
        } else {
            if (ContextCompat.checkSelfPermission(
                    requireContext(),
                    Manifest.permission.WRITE_EXTERNAL_STORAGE
                )
                != PackageManager.PERMISSION_GRANTED
            ) {

                // Permission is not granted
                // Should we show an explanation?
                if (shouldShowRequestPermissionRationale(
                        Manifest.permission.WRITE_EXTERNAL_STORAGE
                    )
                ) {
                    // Show an explanation to the user *asynchronously* -- don't block
                    // this thread waiting for the user's response! After the user
                    // sees the explanation, try again to request the permission.
                } else {
                    // No explanation needed, we can request the permission.
                    requestPermissions(
                        arrayOf(
                            Manifest.permission.WRITE_EXTERNAL_STORAGE,
                            Manifest.permission.READ_EXTERNAL_STORAGE
                        ),
                        REQUEST_PERMISSIONS_CODE_WRITE_STORAGE
                    )

                    // MY_PERMISSIONS_REQUEST_READ_CONTACTS is an
                    // app-defined int constant. The callback method gets the
                    // result of the request.
                }
            } else {
                // Permission has already been granted
                appFolderInit.onNext(downloadFolder)
            }
        }

        compositeDisposable.add(appStateManager.fileLoadResultSubject
            .observeOn(AndroidSchedulers.mainThread())
            .doOnError { Timber.e(it.localizedMessage) }
            .map { appFile ->
                update_app_progress_bar.hide()
                update_app_download_finished_label.show()
                update_app_confirm_app_button.show()
                update_app_confirm_app_button.clicks()
                    .subscribeOn(AndroidSchedulers.mainThread())
                    .map {
                        val intent = Intent()
                        intent.flags = Intent.FLAG_ACTIVITY_CLEAR_TOP

                        val uri: Uri?
                        var mime: String? =
                            "application/vnd.android.package-archive" // Default Value

                        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N) {
                            intent.action = Intent.ACTION_INSTALL_PACKAGE

                            uri = getUriForFile(
                                requireContext(),
                                BuildConfig.APPLICATION_ID + ".fileprovider",
                                appFile
                            )

                            if (uri != null) {
                                mime = requireContext().contentResolver.getType(uri)
                            }
                        } else {
                            intent.action = Intent.ACTION_VIEW
                            intent.putExtra(Intent.EXTRA_NOT_UNKNOWN_SOURCE, true)
                            intent.flags = Intent.FLAG_ACTIVITY_NEW_TASK

                            appFile.setReadable(true)
                            uri = Uri.fromFile(appFile)

                            mime =
                                MimeTypeMap.getSingleton()
                                    .getMimeTypeFromExtension(appFile.extension)
                        }

                        intent.setDataAndType(uri, mime)
                        intent.addFlags(Intent.FLAG_GRANT_READ_URI_PERMISSION)
                        intent.addFlags(Intent.FLAG_ACTIVITY_CLEAR_TOP)

                        startActivity(intent)
                    }
                    .doOnError { Timber.e(it) }
                    .subscribe()
            }
            .subscribe()
        )
    }

    override fun onRequestPermissionsResult(
        requestCode: Int,
        permissions: Array<String>, grantResults: IntArray
    ) {
        when (requestCode) {
            REQUEST_PERMISSIONS_CODE_WRITE_STORAGE -> {
                // If request is cancelled, the result arrays are empty.
                if ((grantResults.isNotEmpty() && grantResults[0] == PackageManager.PERMISSION_GRANTED)) {
                    // permission was granted, yay! Do the
                    // contacts-related task you need to do.

                    appFolderInit.onNext(downloadFolder)
                } else {
                    Timber.d("${javaClass.name}: Permissions denied! Request them before file uploading.")
                    // permission denied, boo! Disable the
                    // functionality that depends on this permission.
                }
                return
            }

            // Add other 'when' lines to check for other
            // permissions this app might request.
            else -> {
                // Ignore all other requests.
            }
        }
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        this.requireActivity().drawer_layout?.setDrawerLockMode(DrawerLayout.LOCK_MODE_LOCKED_CLOSED)
    }

    override fun onCreateView(
        inflater: LayoutInflater,
        container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View? {
        val view = inflater.inflate(
            R.layout.fragment_update_app,
            container,
            false
        )
        (activity as? AppBarController)?.hideAppBar()
        return view
    }

    override fun createPresenter(): UpdateAppPresenter = updateAppPresenter

    override fun loadingIntent(): Observable<Unit> = Observable.just(Unit)

    override fun downloadIntent(): Observable<String> = appFolderInit

    override fun clearErrorIntent(): Observable<Unit> = clearErrorSubject

    override fun render(state: UpdateAppViewState) {
        manageErrors(state)

        state.appInfo?.let { appInfo ->
            update_app_version_title.text = appInfo.versionName
        }

        if (state.progress < 100 && state.appFile == null) {
            update_app_progress_bar.show()
            update_app_progress_bar.progress = state.progress
            update_app_download_finished_label.hide()
        }
    }

    private fun manageErrors(state: UpdateAppViewState) {
        state.errorMessage?.let { errorMessage ->
            showSnackbar(errorMessage)
        }
    }

    private fun showSnackbar(message: String) {
        snackbar = Snackbar.make(
            requireActivity().findViewById(android.R.id.content)!!,
            message,
            Snackbar.LENGTH_SHORT
        )
        snackbar.view.setBackgroundColor(
            ContextCompat.getColor(
                requireContext(),
                R.color.accentSnackbarBackground
            )
        )
        snackbar.show()
    }

    companion object {
        const val REQUEST_PERMISSIONS_CODE_WRITE_STORAGE = 20128
    }
}