package ru.daotech.agroinvest.fragments.pages.update

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.Disposable
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.models.AppInfo
import ru.daotech.agroinvest.service.jsonmodels.AppInfoPayload
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

class UpdateAppPresenter @Inject constructor(
    private var goToMainScreen: (() -> Unit)?,
    private val repository: Repository,
    private val appStateManager: AppStateManager
) :
    MviBasePresenter<UpdateAppView, UpdateAppViewState>() {

    lateinit var disposable: Disposable

    override fun bindIntents() {

        val appInfo = getAppInfoFromPref()

        val loadingIntent = intent(UpdateAppView::loadingIntent)
            .switchMap { makeLoadingIntent(appInfo) }

        val downloadIntent = intent(UpdateAppView::downloadIntent)
            .delay(2, TimeUnit.SECONDS)
            .take(1)
            .switchMap { appFolder -> makeDownloadIntent(appInfo, appFolder) }
            .cast(UpdateAppPartialViewState::class.java)

        val flowableProgressIntent = appStateManager.fileProgressDownload
            .throttleFirst(10, TimeUnit.MILLISECONDS)
            .observeOn(AndroidSchedulers.mainThread())
            .doOnError(Timber::e)
            .switchMap { progressValue ->
                println(progressValue)
                Observable.just(UpdateAppPartialViewState.ShowDownloadProgressValue(progressValue))
            }
            .cast(UpdateAppPartialViewState::class.java)

        val fileLoadResultIntent = appStateManager.fileLoadResultSubject
            .doOnError(Timber::e)
            .observeOn(AndroidSchedulers.mainThread())
            .switchMap { file ->
                println(file)
                Observable.just(UpdateAppPartialViewState.ShowUpdateButton(file))
            }
            .cast(UpdateAppPartialViewState::class.java)

        val clearErrorIntent = intent(UpdateAppView::clearErrorIntent)
            .map { UpdateAppPartialViewState.ClearError }

        val allIntentList = listOf(
            loadingIntent,
            flowableProgressIntent,
            downloadIntent,
            fileLoadResultIntent,
            clearErrorIntent
        )

        val allIntent = Observable.merge(allIntentList)
            .onErrorResumeNext { error: Throwable -> errorProcessing(error) }
            .share()

        disposable = allIntent
            .filter { state -> state is UpdateAppPartialViewState.ShowNextScreen }
            .subscribe(
                { goToMainScreen?.invoke() },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )

        val initialState = UpdateAppViewState(
            appInfo = null,
            appFile = null,
            progress = 0,
            isLoading = false,
            errorMessage = null
        )

        val stateObservable =
            allIntent.scan(initialState) { previousState, partialViewState ->
                partialViewState.reduce(previousState)
            }

        subscribeViewState(stateObservable, UpdateAppView::render)
    }

    private fun getAppInfoFromPref(): AppInfo {
        return AppInfo(
            from = AppInfoPayload(
                versionName = appStateManager.getActualAppInfo()!!.serverDataPayload.versionName,
                appUrl = appStateManager.getActualAppInfo()!!.serverDataPayload.versionUrl,
                versionCode = appStateManager.getActualAppInfo()!!.serverDataPayload.versionName
            )
        )
    }

    private fun makeLoadingIntent(appInfo: AppInfo): Observable<UpdateAppPartialViewState> {
        return Observable.just(appInfo)
            .map { UpdateAppPartialViewState.ShowAppInfo(it) as UpdateAppPartialViewState }
            .onErrorReturn {
                Timber.e(it)
                UpdateAppPartialViewState.Error("Информация об обновлении не найдена")
            }
    }

    private fun makeDownloadIntent(
        appInfo: AppInfo,
        appFolder: String
    ): Observable<UpdateAppPartialViewState> {
        return repository.downloadAppUrl(appInfo.appUrl, appFolder, appInfo.identifier)
            .subscribeOn(AndroidSchedulers.mainThread())
            .map { UpdateAppPartialViewState.ShowUpdateButton(it) as UpdateAppPartialViewState }
            .doOnError { Timber.e(it.localizedMessage) }
            .onErrorReturn {
                Timber.e(it)
                UpdateAppPartialViewState.Error(
                    "Ошибка прав доступа. Перезапустите приложение и повторите попытку."
                )
            }
    }

    private fun errorProcessing(error: Throwable): Observable<UpdateAppPartialViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .take(1)
            .map { UpdateAppPartialViewState.ClearError as UpdateAppPartialViewState }
            .startWith(UpdateAppPartialViewState.Error(errorMessage = error.localizedMessage))
    }

    override fun unbindIntents() {
        goToMainScreen = null
        disposable.dispose()
    }
}