package ru.daotech.agroinvest.fragments.ratings

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import ru.daotech.agroinvest.business.Repository
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

open class RatingsPresenter @Inject constructor(
    private val repository: Repository,
    private var onItemClick: ((String) -> Unit)?
) :
    MviBasePresenter<RatingsView, RatingsViewState>() {

    private var disposable: CompositeDisposable = CompositeDisposable()

    override fun bindIntents() {

        val loadRatingData = intent(RatingsView::loadingIntent)
            .switchMap { loadRatings() }

        val errorClearIntent = intent(RatingsView::snackbarDismiss)
            .map { RatingsPartialViewState.ClearError as RatingsPartialViewState }

        val allIntent = Observable.merge(loadRatingData, errorClearIntent)
            .onErrorResumeNext { error: Throwable -> errorProcessing(error) }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        disposable.add(
            allIntent
                .filter { state -> state is RatingsPartialViewState.GoToDetailedItem }
                .map { state -> state as RatingsPartialViewState.GoToDetailedItem }
                .subscribe(
                    { state -> onItemClick?.invoke(state.ratingId) },
                    { error -> Timber.e(error, error.localizedMessage) }
                )
        )

        val initialState =
            RatingsViewState(
                ratingsList = null,
                isLoading = false,
                errorMessage = null
            )

        val stateObservable = allIntent.scan(initialState, this::reducer)

        subscribeViewState(stateObservable, RatingsView::render)
    }

    private fun reducer(
        previousState: RatingsViewState,
        partialViewState: RatingsPartialViewState
    ): RatingsViewState {
        return partialViewState.reduce(previousState)
    }


    private fun loadRatings(): Observable<RatingsPartialViewState> {
        return repository.getRatings()
            .observeOn(AndroidSchedulers.mainThread())
            .map { ratings ->
                RatingsPartialViewState.SuccessLoadRatings(ratings) as RatingsPartialViewState
            }
            .doOnError { error -> errorProcessing(error) }
            .startWith(RatingsPartialViewState.Loading as RatingsPartialViewState)
    }

    override fun unbindIntents() {
        disposable.dispose()
        onItemClick = null
    }

    private fun errorProcessing(
        error: Throwable
    ): Observable<RatingsPartialViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .take(1)
            .map { RatingsPartialViewState.ClearError as RatingsPartialViewState }
            .startWith(RatingsPartialViewState.Error(errorMessage = error.localizedMessage))
    }
}