package ru.daotech.agroinvest.fragments.sections.detailed

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import ru.daotech.agroinvest.business.Repository
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

open class DetailedSectionPresenter @Inject constructor(
    private val repository: Repository,
    private var onItemClick: ((String) -> Unit)?
) :
    MviBasePresenter<DetailedSectionView, DetailedSectionViewState>() {

    private var disposable: CompositeDisposable = CompositeDisposable()

    private lateinit var parentId: String

    override fun bindIntents() {

        val loadSectionData = intent(DetailedSectionView::loadingIntent)
            .doOnNext { parentId -> this.parentId = parentId }
            .switchMap { parentId -> this.loadDetailedSection(parentId) }

        val loadParentSection = intent(DetailedSectionView::loadingIntent)
            .doOnNext { parentId -> this.parentId = parentId }
            .switchMap { parentId -> this.loadParentSection(parentId) }

        val errorClearIntent = intent(DetailedSectionView::snackbarDismiss)

        val onSectionItemClick = intent(DetailedSectionView::onSectionClick)
            .map { section -> DetailedSectionPartialViewState.GoToSectionTasks(section.id) }

        val allIntent =
            Observable.merge(
                loadSectionData,
                loadParentSection,
                onSectionItemClick
            )
                .onErrorResumeNext { error: Throwable ->
                    errorProcessing(error, errorClearIntent)
                }
                .observeOn(AndroidSchedulers.mainThread())
                .share()

        disposable.add(
            allIntent
                .filter { state -> state is DetailedSectionPartialViewState.GoToSectionTasks }
                .map { state -> state as DetailedSectionPartialViewState.GoToSectionTasks }
                .subscribe(
                    { state -> onItemClick?.invoke(state.sectionId) },
                    { error -> Timber.e(error, error.localizedMessage) }
                )
        )

        val initialState =
            DetailedSectionViewState(
                sectionsList = null,
                parentSection = null,
                isLoading = false,
                errorMessage = null
            )

        val stateObservable = allIntent.scan(initialState, this::reducer)

        subscribeViewState(stateObservable, DetailedSectionView::render)
    }

    private fun reducer(
        previousState: DetailedSectionViewState,
        partialViewState: DetailedSectionPartialViewState
    ): DetailedSectionViewState {
        return partialViewState.reduce(previousState)
    }


    private fun loadDetailedSection(parentId: String?)
            : Observable<DetailedSectionPartialViewState> {
        return repository.getSectionParent(parentId)
            .observeOn(AndroidSchedulers.mainThread())
            .map { sections ->
                DetailedSectionPartialViewState.SuccessLoadDetailedSection(sections)
                        as DetailedSectionPartialViewState
            }
            .startWith(DetailedSectionPartialViewState.Loading as DetailedSectionPartialViewState)
    }

    private fun loadParentSection(parentId: String): Observable<DetailedSectionPartialViewState> {
        return repository.getSection(parentId)
            .observeOn(AndroidSchedulers.mainThread())
            .map { section ->
                DetailedSectionPartialViewState.SuccessLoadParentSection(section)
                        as DetailedSectionPartialViewState
            }
            .startWith(DetailedSectionPartialViewState.Loading as DetailedSectionPartialViewState)
    }

    override fun unbindIntents() {
        disposable.dispose()
    }

    private fun errorProcessing(
        error: Throwable,
        errorClearIntent: Observable<Unit>
    ): Observable<DetailedSectionPartialViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .mergeWith(errorClearIntent)
            .take(1)
            .map { DetailedSectionPartialViewState.ClearError as DetailedSectionPartialViewState }
            .startWith(DetailedSectionPartialViewState.Error(errorMessage = error.localizedMessage))
    }
}