package ru.daotech.agroinvest.fragments.sections.root

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import ru.daotech.agroinvest.business.Repository
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

open class SectionsPresenter @Inject constructor(
    private val repository: Repository,
    private var onItemClick: ((String) -> Unit)?
) :
    MviBasePresenter<SectionsView, SectionsViewState>() {

    private var disposable: CompositeDisposable = CompositeDisposable()

    override fun bindIntents() {

        val loadSectionData = intent(SectionsView::loadingIntent)
            .switchMap { loadSections() }

        val errorClearIntent = intent(SectionsView::snackbarDismiss)

        val onSectionItemClick = intent(SectionsView::onSectionClick)
            .map { section ->
                SectionsPartialViewState.GoToDetailedItem(section.id)
            }

        val allIntent = Observable.merge(loadSectionData, onSectionItemClick)
            .onErrorResumeNext { error: Throwable ->
                errorProcessing(error, errorClearIntent)
            }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        disposable.add(
            allIntent
                .filter { state -> state is SectionsPartialViewState.GoToDetailedItem }
                .map { state -> state as SectionsPartialViewState.GoToDetailedItem }
                .subscribe(
                    { state ->
                        onItemClick?.invoke(state.sectionId)
                    },
                    { error ->
                        Timber.e(error, error.localizedMessage)
                    }
                )
        )

        val initialState =
            SectionsViewState(
                sectionsList = null,
                isLoading = false,
                errorMessage = null
            )

        val stateObservable = allIntent.scan(initialState, this::reducer)

        subscribeViewState(stateObservable, SectionsView::render)
    }

    private fun reducer(
        previousState: SectionsViewState,
        partialViewState: SectionsPartialViewState
    ): SectionsViewState {
        return partialViewState.reduce(previousState)
    }


    private fun loadSections(): Observable<SectionsPartialViewState> {
        return repository.getRootSections()
            .observeOn(AndroidSchedulers.mainThread())
            .map { sections ->
                SectionsPartialViewState.SuccessLoadSections(sections) as SectionsPartialViewState
            }
            .startWith(SectionsPartialViewState.Loading as SectionsPartialViewState)
    }

    override fun unbindIntents() {
        disposable.dispose()
    }

    private fun errorProcessing(
        error: Throwable,
        errorClearIntent: Observable<Unit>
    ): Observable<SectionsPartialViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .mergeWith(errorClearIntent)
            .take(1)
            .map {
                SectionsPartialViewState.ClearError as SectionsPartialViewState
            }
            .startWith(SectionsPartialViewState.Error(errorMessage = error.localizedMessage))
    }
}