package ru.daotech.agroinvest.fragments.support_notification

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.BehaviorSubject
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.adapters.SupportMessageModelView
import ru.daotech.agroinvest.adapters.WorkerMessageModelView
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.database.entities.SupportMessage
import ru.daotech.agroinvest.utils.CHAT_MESSAGE_SUPPORT_CODE
import ru.daotech.agroinvest.utils.CHAT_MESSAGE_WORKER_CODE
import timber.log.Timber
import java.util.concurrent.TimeUnit

class SupportNotificationPresenter(
    private val repository: Repository,
    private val appStateManager: AppStateManager,
    private var goToAccidentReport: (() -> Unit)?
) : MviBasePresenter<SupportNotificationView, SupportNotificationViewState>() {

    private var compositeDisposable: CompositeDisposable = CompositeDisposable()

    private val onFlowableListSubject: BehaviorSubject<List<SupportMessage>> =
        BehaviorSubject.create()

    override fun bindIntents() {

        val loadingIntent = onFlowableListSubject
            .observeOn(AndroidSchedulers.mainThread())
            .switchMap { supportMessageList ->
                Observable.just(Unit)
                    .map {
                        SupportNotificationViewState.LoadedSupportMessages(
                            supportMessageList.map { supportMessage ->
                                when (supportMessage.messageCode) {
                                    CHAT_MESSAGE_SUPPORT_CODE -> {
                                        SupportMessageModelView(supportMessage)
                                    }
                                    CHAT_MESSAGE_WORKER_CODE -> {
                                        WorkerMessageModelView(
                                            supportMessage,
                                            appStateManager.getCurrentWorkerSimple()!!
                                        )
                                    }
                                    else -> SupportMessageModelView(supportMessage) // Default type
                                }
                            }
                        )
                    }
                    .cast(SupportNotificationViewState::class.java)
                    .startWith(SupportNotificationViewState.Loading as SupportNotificationViewState)
                    .onErrorReturn { SupportNotificationViewState.LoadingDismiss }
            }
            .startWith(SupportNotificationViewState.Loading)
            .cast(SupportNotificationViewState::class.java)

        val goToAccidentIntent = intent(SupportNotificationView::onGoToAccidentIntent)
            .map { SupportNotificationViewState.GoToAccidentState }
            .cast(SupportNotificationViewState::class.java)

        val sendUserSupportMessageIntent =
            intent(SupportNotificationView::onSendUserSupportMessageIntent)
                .flatMap { message -> repository.sendSupportUserMessage(message) }
                .observeOn(AndroidSchedulers.mainThread())
                .map { SupportNotificationViewState.EmptyState }
                .cast(SupportNotificationViewState::class.java)

        val allIntent = Observable.merge(
            loadingIntent,
            goToAccidentIntent,
            sendUserSupportMessageIntent
        )
            .onErrorResumeNext { error: Throwable ->
                errorProcessing(error)
            }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        compositeDisposable.add(
            repository.getAllSupportMessagesByCurrentWorker()
                .throttleLast(2, TimeUnit.SECONDS)
                .subscribe { supportMessages ->
                    if (!supportMessages.isNullOrEmpty() && supportMessages.isNotEmpty()) {
                        onFlowableListSubject.onNext(supportMessages)
                    }
                }
        )

        compositeDisposable.add(
            allIntent
                .filter { state -> state is SupportNotificationViewState.GoToAccidentState }
                .map { state -> state as SupportNotificationViewState.GoToAccidentState }
                .observeOn(AndroidSchedulers.mainThread())
                .subscribe(
                    { goToAccidentReport?.invoke() },
                    { error -> Timber.e(error, error.localizedMessage) }
                )
        )

        compositeDisposable.add(
            allIntent
                .filter { state -> state is SupportNotificationViewState.LoadedSupportMessages }
                .map { state -> state as SupportNotificationViewState.LoadedSupportMessages }
                .observeOn(AndroidSchedulers.mainThread())
                .subscribe(
                    { state ->
                        repository.updateUnreadStatusSupportMessages(
                            state.supportMessagesList.map { baseModel -> baseModel.message }
                        )
                            .subscribe()
                    },
                    { error -> Timber.e(error, error.localizedMessage) }
                )
        )

        subscribeViewState(allIntent, SupportNotificationView::render)
    }

    private fun errorProcessing(
        error: Throwable
    ): Observable<SupportNotificationViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .take(1)
            .map {
                SupportNotificationViewState.Error as SupportNotificationViewState
            }
            .startWith(SupportNotificationViewState.Error)
    }

    override fun unbindIntents() {
        compositeDisposable.clear()
        goToAccidentReport = null
    }
}