package ru.daotech.agroinvest.fragments.tasks.collectionHistory

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.functions.BiFunction
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.BehaviorSubject
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.database.entities.TrolleyInfoImpl
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

open class CollectionHistoryPresenter @Inject constructor(
    private val repository: Repository,
    private var appStateManager: AppStateManager
) :
    MviBasePresenter<CollectionHistoryView, CollectionHistoryViewState>() {

    private var compositeDisposable: CompositeDisposable = CompositeDisposable()
    private lateinit var disposable: Disposable
    private val onFlowableListSubject: BehaviorSubject<List<TrolleyInfoImpl>> = BehaviorSubject.create()

    private fun addLoadCollectionHistroyFlowable() {
        disposable = getCollectionHistoryDisposable()
        compositeDisposable.add(disposable)
    }

    private fun getCollectionHistoryDisposable(): Disposable {
        return repository.getCollectionHistory()
            .throttleLast(5, TimeUnit.SECONDS)
            .observeOn(AndroidSchedulers.mainThread())
            .subscribe { trolleyInfoList ->
                onFlowableListSubject.onNext(trolleyInfoList)
            }
    }

    override fun bindIntents() {

        val loadTasksData = intent(CollectionHistoryView::loadingIntent)
            .switchMap { loadCollectionHistory() }

        val errorClearIntent = intent(CollectionHistoryView::snackbarDismiss)

        val collectionHistoryListSignalIntent = intent(
            CollectionHistoryView::collectionHistoryListSignalIntent
        )
            .throttleFirst(30, TimeUnit.SECONDS)
            .switchMap { makeSignalForcollectionHistoryList() }

        addLoadCollectionHistroyFlowable()

        val errorIntent = intent(CollectionHistoryView::snackbarDismiss)
            .switchMap {
                Observable.just(
                    CollectionHistoryPartialViewState.Error(
                        "Во время загрузки данных произошла ошибка."
                    )
                )
            }

        val allIntent = Observable.merge(
            loadTasksData,
            errorIntent,
            collectionHistoryListSignalIntent
        )
            .onErrorResumeNext { error: Throwable ->
                errorProcessing(error, errorClearIntent)
            }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        val initialState =
            CollectionHistoryViewState(
                trolleyInfoList = null,
                isLoading = false,
                errorMessage = null
            )

        val stateObservable = allIntent.scan(initialState, this::reducer)

        subscribeViewState(stateObservable, CollectionHistoryView::render)
    }

    private fun makeSignalForcollectionHistoryList(): Observable<CollectionHistoryPartialViewState> {
        return Observable.zip<Unit, Unit, Unit>(
            Observable.timer(3, TimeUnit.SECONDS).map { Unit },
            repository.signalToTrolleyInfoHistoryList(workerId = appStateManager.getCurrentWorkerSimple()!!.id),
            BiFunction { _, _ -> Unit }
        )
            .observeOn(AndroidSchedulers.mainThread())
            .onErrorReturn {
                CollectionHistoryPartialViewState.LoadingDismiss
                Timber.e(it)
            }
            .map { CollectionHistoryPartialViewState.LoadingDismiss }
    }

    private fun reducer(
        previousState: CollectionHistoryViewState,
        partialViewState: CollectionHistoryPartialViewState
    ): CollectionHistoryViewState {
        return partialViewState.reduce(previousState)
    }

    private fun loadCollectionHistory(): Observable<CollectionHistoryPartialViewState> {
        return repository.getCollectionHistory()
            .observeOn(AndroidSchedulers.mainThread())
            .map { taskWithProperties ->
                CollectionHistoryPartialViewState.SuccessLoadCollectionHistory(
                    taskWithProperties,
                    appStateManager.getServerTimestamp()
                ) as CollectionHistoryPartialViewState
            }
            .startWith(CollectionHistoryPartialViewState.Loading as CollectionHistoryPartialViewState)
    }

    private fun errorProcessing(
        error: Throwable,
        errorClearIntent: Observable<Unit>
    ): Observable<CollectionHistoryPartialViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .mergeWith(errorClearIntent)
            .take(1)
            .map {
                CollectionHistoryPartialViewState.ClearError as CollectionHistoryPartialViewState
            }
            .map { CollectionHistoryPartialViewState.Error(errorMessage = error.localizedMessage) }
    }

    override fun unbindIntents() {
        compositeDisposable.dispose()
    }
}