package ru.daotech.agroinvest.fragments.tasks.myTasks

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.functions.BiFunction
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.BehaviorSubject
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.database.entities.Task
import ru.daotech.agroinvest.models.RequestTypeEnum
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

open class TasksPresenter @Inject constructor(
    private val appStateManager: AppStateManager,
    private val repository: Repository,
    private var onItemClick: ((String) -> Unit)?,
    private var goToScannedTask: ((String, String) -> Unit)?,
    private var goToTrolleyTied: ((String?) -> Unit)?,
    private var goToBarcodeBox: (() -> Unit)?
) :
    MviBasePresenter<TasksView, TasksViewState>() {

    private var compositeDisposable: CompositeDisposable = CompositeDisposable()

    private val onFlowableListSubject: BehaviorSubject<List<Task>> = BehaviorSubject.create()

    private lateinit var disposable: Disposable

    private fun addFlowable(isFirstLaunch: Boolean = true) {
        disposable =
            when (isFirstLaunch) {
                true -> getTasksWithThrottle(5)
                false -> getTasksWithThrottle(0)
            }

        compositeDisposable.add(disposable)
    }

    private fun getTasksWithThrottle(throttleSeconds: Long): Disposable {
        return repository.getMyTasksFlowable()
            .throttleLast(throttleSeconds, TimeUnit.SECONDS)
            .observeOn(AndroidSchedulers.mainThread())
            .subscribe { tasksList ->
                onFlowableListSubject.onNext(tasksList)
            }
    }

    override fun bindIntents() {
        val loadTasksDataFlowable = onFlowableListSubject
            .switchMap { tasksList -> loadMyTasks(tasksList) }

        addFlowable(appStateManager.getIsFirstLaunch())
        appStateManager.setIsFirstLaunch(false)

        val disposableFilter: Disposable = appStateManager.isFilterRowChangedSubject
            .subscribe {
                disposable.dispose()
                addFlowable(isFirstLaunch = false)
            }

        compositeDisposable.add(disposableFilter)

        val loadTasksData = intent(TasksView::loadingIntent)
            .switchMap { loadMyTasks(null) }

        val taskListSignalIntent = intent(
            TasksView::taskListSignalIntent
        )
            .throttleFirst(30, TimeUnit.SECONDS)
            .switchMap { makeSignalForTaskList() }

        val clearLastTaskStatePreferencesIntent = intent(
            TasksView::clearPreferencesIntent
        )
            .switchMap { clearLastTaskStatePreferences() }

        val errorWrongNfcTypeIntent = intent(
            TasksView::errorWrongNfcTypeIntent
        )
            .switchMap {
                Observable.just(
                    TasksPartialViewState.Error(
                        "Ошибка. Неверный тип NFC метки."
                    )
                )
            }

        val onTaskItemClick = intent(TasksView::onTaskClick)
            .map { task -> TasksPartialViewState.GoToDetailedItem(task.id) }

        val dismissTaskAlreadyInProgressDialogIntent = intent(TasksView::dismissTaskAlreadyInProgressDialogIntent)
            .map { task -> TasksPartialViewState.DismissTaskAlreadyInProgressDialog }

        val openDetailedTaskByIdIntent = intent(
            TasksView::openDetailedTaskByIdIntent
        )
            .map { taskId ->
                TasksPartialViewState.GoToDetailedItem(taskId) as TasksPartialViewState
            }

        val onClearRatingMessageIntent = intent(
            TasksView::clearRatingMessageIntent
        )
            .switchMap { clearRatingMessage() }

        val goToDetailedTaskIntent = intent(
            TasksView::goToDetailedTaskIntent
        )
            .switchMap { launchTaskByNfcCode(it) }

        val goToBarcodeBoxIntent = intent(
            TasksView::goToBarcodeBoxIntent
        )
            .map { TasksPartialViewState.GoingToBarcodeBox }

        val dismissTaskAttentionDialogIntent = intent(
            TasksView::dismissTaskAttentionDialogIntent
        )
            .map { TasksPartialViewState.DismissTaskAttentionDialogIntent as TasksPartialViewState }

        val onCurrentTrolleyTiedIntent = intent(
            TasksView::onCurrentTrolleyTiedIntent
        )
            .map { TasksPartialViewState.GoToTrolleyTied(null) as TasksPartialViewState }

        val onCurrentTrolleyIntent = intent(
            TasksView::onCurrentTrolleyIntent
        )
            .switchMap { trolleyNumber ->
                return@switchMap Observable.just(
                    TasksPartialViewState.GoToTrolleyTied(trolleyNumber) as TasksPartialViewState
                )
            }

        val onCurrentRawIntent = intent(TasksView::onCurrentRawIntent)
            .switchMap { rawNumber ->
                when (!appStateManager.getCurrentRowNumber().isBlank()) {
                    true -> {
                        return@switchMap if (rawNumber == appStateManager.getCurrentRowNumber()) {
                            launchTaskByNfcCode(rawNumber)
                        } else {
                            Observable.just(
                                TasksPartialViewState.ShowErrorTaskAlreadyInProgressAlertAttention
                            )
                        }
                    }
                    false -> {
                        launchTaskByNfcCode(rawNumber)
                    }
                }
            }

        val errorClearIntent = intent(TasksView::snackbarDismiss)

        val allIntentList = listOf(
            loadTasksDataFlowable,
            loadTasksData,
            onTaskItemClick,
            onClearRatingMessageIntent,
            openDetailedTaskByIdIntent,
            clearLastTaskStatePreferencesIntent,
            taskListSignalIntent,
            onCurrentRawIntent,
            errorWrongNfcTypeIntent,
            onCurrentTrolleyIntent,
            onCurrentTrolleyTiedIntent,
            dismissTaskAlreadyInProgressDialogIntent,
            goToDetailedTaskIntent,
            dismissTaskAttentionDialogIntent,
            goToBarcodeBoxIntent
        )

        val allIntent = Observable.merge(
            allIntentList
        )
            .onErrorResumeNext { error: Throwable ->
                errorProcessing(error, errorClearIntent)
            }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        compositeDisposable.add(
            allIntent
                .filter { state -> state is TasksPartialViewState.GoToDetailedItem }
                .map { state -> state as TasksPartialViewState.GoToDetailedItem }
                .subscribe(
                    { state -> onItemClick?.invoke(state.taskId) },
                    { error -> Timber.e(error, error.localizedMessage) }
                )
        )

        compositeDisposable.add(
            allIntent
                .filter { state -> state is TasksPartialViewState.GoToTrolleyTied }
                .map { state -> state as TasksPartialViewState.GoToTrolleyTied }
                .subscribe(
                    { state -> goToTrolleyTied?.invoke(state.nfcTrolleyMark) },
                    { error -> Timber.e(error, error.localizedMessage) }
                )
        )

        compositeDisposable.add(
            allIntent.filter { state ->
                state is TasksPartialViewState.GoingToBarcodeBox
            }.subscribe(
                {
                    goToBarcodeBox?.invoke()
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )
        )

        val initialState =
            TasksViewState(
                taskList = null,
                isLoading = false,
                errorMessage = null,
                ratingMessage = null,
                dismissSnackbar = false,
                serverTimestamp = null,
                isRowFilterDefined = false,
                isTaskAlreadyInProgress = false,
                nfcTrolleyMark = null
            )

        val stateObservable = allIntent.scan(initialState, this::reducer)

        subscribeViewState(stateObservable, TasksView::render)
    }

    private fun clearLastTaskStatePreferences(): Observable<TasksPartialViewState> {
        return repository.clearLastTaskStatePreferences()
            .map { TasksPartialViewState.ClearLastTaskStatePreferences as TasksPartialViewState }
    }

    private fun reducer(
        previousState: TasksViewState,
        partialViewState: TasksPartialViewState
    ): TasksViewState {
        return partialViewState.reduce(previousState)
    }

    private fun loadMyTasks(tasks: List<Task>?): Observable<TasksPartialViewState> {
        return Observable.just(Unit)
            .map {
                when (tasks) {
                    null -> TasksPartialViewState.Loading
                    else -> if (!appStateManager.getFilterRowNumber().isNullOrBlank()) {
                        TasksPartialViewState.SuccessLoadTasksByFilter(
                            tasks,
                            appStateManager.getServerTimestamp()
                        )
                    } else {
                        TasksPartialViewState.SuccessLoadTasks(
                            tasks,
                            appStateManager.getServerTimestamp()
                        )
                    }
                }
            }
            .startWith(TasksPartialViewState.Loading as TasksPartialViewState)
            .onErrorReturn { TasksPartialViewState.LoadingDismiss }
    }

    private fun makeSignalForTaskList(): Observable<TasksPartialViewState> {
        return Observable.zip<Unit, Unit, Unit>(
            Observable.timer(3, TimeUnit.SECONDS).map { Unit },
            repository.signalToTaskList(workerId = appStateManager.getCurrentWorkerSimple()!!.id),
            BiFunction { _, _ -> Unit }
        )
            .observeOn(AndroidSchedulers.mainThread())
            .onErrorReturn {
                TasksPartialViewState.LoadingDismiss
                Timber.e(it)
            }
            .map { TasksPartialViewState.LoadingDismiss }
    }

    private fun launchTaskByNfcCode(nfcCode: String): Observable<TasksPartialViewState> {
        return repository.getMyTasksByRowNumber(nfcCode)
            .observeOn(AndroidSchedulers.mainThread())
            .debounce(1, TimeUnit.SECONDS)
            .map { tasks ->
                tasks
                    .filter { task -> task.rowNumber == nfcCode.toInt() }
                    .forEach { task ->
                        return@map when {
                            tasks.count { it.rowNumber == nfcCode.toInt() } > 1 -> {
                                TasksPartialViewState
                                    .Error(
                                        "Задача ${task.rowNumber} в процессе. " +
                                                "Завершите ее для начала новой задачи"
                                    )
                            }
                            else -> {
                                goToScannedTask?.invoke(task.id, nfcCode)
                                TasksPartialViewState.ClearError
                            }
                        }
                    }
                return@map TasksPartialViewState.Error(
                    "Задач на метку не назначено. Перейдите к другой задаче"
                )
            }
            .cast(TasksPartialViewState::class.java)
            .doOnError { Timber.e(it) }
    }

    private fun clearRatingMessage(): Observable<TasksPartialViewState> {
        return Observable.timer(3, TimeUnit.SECONDS)
            .map {
                TasksPartialViewState.ClearError as TasksPartialViewState
            }
            .startWith(TasksPartialViewState.ShowRatingMessage(null))
    }

    private fun errorProcessing(
        error: Throwable,
        errorClearIntent: Observable<Unit>
    ): Observable<TasksPartialViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .mergeWith(errorClearIntent)
            .take(1)
            .map {
                TasksPartialViewState.ClearError as TasksPartialViewState
            }
            .startWith(TasksPartialViewState.Error(errorMessage = error.localizedMessage))
    }

    override fun unbindIntents() {
        super.unbindIntents()
        compositeDisposable.dispose()

        onItemClick = null
        goToScannedTask = null
        goToTrolleyTied = null
        goToBarcodeBox = null
    }
}