package ru.daotech.agroinvest.fragments.tasks.sectionTasks

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import ru.daotech.agroinvest.business.Repository
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

open class SectionTasksPresenter @Inject constructor(
    private val repository: Repository,
    private var onItemClick: ((String) -> Unit)?
) :
    MviBasePresenter<SectionTasksView, SectionTasksViewState>() {

    private var disposable: CompositeDisposable = CompositeDisposable()

    private lateinit var sectionId: String

    override fun bindIntents() {

        val loadTasksData = intent(SectionTasksView::loadingIntent)
            .doOnNext { sectionId ->
                this.sectionId = sectionId
            }
            .switchMap { sectionId ->
                this.loadSectionTasks(
                    sectionId
                )
            }

        val loadParentSection = intent(SectionTasksView::loadingIntent)
            .doOnNext { sectionId ->
                this.sectionId = sectionId
            }
            .switchMap { sectionId ->
                this.loadParentSection(
                    sectionId
                )
            }

        val errorClearIntent = intent(SectionTasksView::snackbarDismiss)

        val onTaskItemClick = intent(SectionTasksView::onTaskClick)
            .map { task ->
                SectionTasksPartialViewState.GoToDetailedItem(task.id)
            }

        val allIntent = Observable.merge(loadTasksData, loadParentSection, onTaskItemClick)
            .onErrorResumeNext { error: Throwable ->
                errorProcessing(error, errorClearIntent)
            }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        disposable.add(
            allIntent
                .filter { state -> state is SectionTasksPartialViewState.GoToDetailedItem }
                .map { state -> state as SectionTasksPartialViewState.GoToDetailedItem }
                .subscribe(
                    { state -> onItemClick?.invoke(state.taskId) },
                    { error ->
                        Timber.e(error, error.localizedMessage)
                    }
                )
        )

        val initialState =
            SectionTasksViewState(
                taskWithPropertiesList = null,
                parentSection = null,
                isLoading = false,
                errorMessage = null
            )

        val stateObservable = allIntent.scan(initialState, this::reducer)

        subscribeViewState(stateObservable, SectionTasksView::render)
    }

    private fun reducer(
        previousState: SectionTasksViewState,
        partialViewState: SectionTasksPartialViewState
    ): SectionTasksViewState {
        return partialViewState.reduce(previousState)
    }


    private fun loadSectionTasks(sectionId: String): Observable<SectionTasksPartialViewState> {
        return repository.getSectionTasks(sectionId)
            .observeOn(AndroidSchedulers.mainThread())
            .map { taskWithPropertiesList ->
                SectionTasksPartialViewState.SuccessLoadTasks(taskWithPropertiesList) as SectionTasksPartialViewState
            }
            .startWith(SectionTasksPartialViewState.Loading as SectionTasksPartialViewState)
    }

    private fun loadParentSection(parentId: String): Observable<SectionTasksPartialViewState> {
        return repository.getSection(parentId)
            .observeOn(AndroidSchedulers.mainThread())
            .map { section ->
                SectionTasksPartialViewState.SuccessLoadParentSection(section) as SectionTasksPartialViewState
            }
            .startWith(SectionTasksPartialViewState.Loading as SectionTasksPartialViewState)
    }

    override fun unbindIntents() {
        disposable.dispose()
    }

    private fun errorProcessing(
        error: Throwable,
        errorClearIntent: Observable<Unit>
    ): Observable<SectionTasksPartialViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .mergeWith(errorClearIntent)
            .take(1)
            .map {
                SectionTasksPartialViewState.ClearError as SectionTasksPartialViewState
            }
            .startWith(SectionTasksPartialViewState.Error(errorMessage = error.localizedMessage))
    }
}