package ru.daotech.agroinvest.fragments.tasks.tasksHistory

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.business.Repository
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

open class TasksHistoryPresenter @Inject constructor(
    private val repository: Repository,
    private var onItemClick: ((String) -> Unit)?,
    private var appStateManager: AppStateManager
) :
    MviBasePresenter<TasksHistoryView, TasksHistoryViewState>() {

    private var disposable: CompositeDisposable = CompositeDisposable()

    override fun bindIntents() {

        val loadTasksData = intent(TasksHistoryView::loadingIntent)
            .switchMap { loadTasksHistory() }

        val errorClearIntent = intent(TasksHistoryView::snackbarDismiss)

        val onTaskItemClick = intent(TasksHistoryView::onTaskClick)
            .map { task ->
                TasksHistoryPartialViewState.Error("Задача уже завершена: " + task.name)
//                TasksHistoryPartialViewState.GoToDetailedItem(task.id)
            }
            .cast(TasksHistoryPartialViewState::class.java)

        val allIntent = Observable.merge(
            loadTasksData,
            onTaskItemClick
            //onSendNotSentTasksIntent
        )
            .onErrorResumeNext { error: Throwable ->
                errorProcessing(error, errorClearIntent)
            }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        disposable.add(
            allIntent
                .filter { state -> state is TasksHistoryPartialViewState.GoToDetailedItem }
                .map { state -> state as TasksHistoryPartialViewState.GoToDetailedItem }
                .subscribe(
                    { state -> onItemClick?.invoke(state.taskId) },
                    { error ->
                        Timber.e(error, error.localizedMessage)
                    }
                )
        )

        val initialState =
            TasksHistoryViewState(
                taskWithPropertiesList = null,
                isLoading = false,
                errorMessage = null,
                uploadDataMessage = null,
                serverTimestamp = null
            )

        val stateObservable = allIntent.scan(initialState, this::reducer)

        subscribeViewState(stateObservable, TasksHistoryView::render)
    }

    private fun reducer(
        previousState: TasksHistoryViewState,
        partialViewState: TasksHistoryPartialViewState
    ): TasksHistoryViewState {
        return partialViewState.reduce(previousState)
    }

    private fun loadTasksHistory(): Observable<TasksHistoryPartialViewState> {
        return repository.getTasksHistory()
            .observeOn(AndroidSchedulers.mainThread())
            .map { taskWithProperties ->
                TasksHistoryPartialViewState.SuccessLoadTasks(
                    taskWithProperties,
                    appStateManager.getServerTimestamp()
                ) as TasksHistoryPartialViewState
            }
            .startWith(TasksHistoryPartialViewState.Loading as TasksHistoryPartialViewState)
    }

    override fun unbindIntents() {
        disposable.dispose()
    }

    private fun errorProcessing(
        error: Throwable,
        errorClearIntent: Observable<Unit>
    ): Observable<TasksHistoryPartialViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .mergeWith(errorClearIntent)
            .take(1)
            .map {
                TasksHistoryPartialViewState.ClearError as TasksHistoryPartialViewState
            }
            .map { TasksHistoryPartialViewState.Error(errorMessage = error.localizedMessage) }
    }
}