package ru.daotech.agroinvest.fragments.trolley_tied

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.utils.DateUtil
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

class TrolleyTiedPresenter @Inject constructor(
    private val repository: Repository,
    private val appStateManager: AppStateManager
) : MviBasePresenter<TrolleyTiedView, TrolleyTiedViewState>() {

    override fun bindIntents() {

        val loadDetailedTaskFromDb =
            intent(TrolleyTiedView::loadingIntent)
                .switchMap { this.loadTrolleyInfo() }

        val onTakeTrolleyButtonClick =
            intent(TrolleyTiedView::takeTrolleyButtonClick)
                .flatMap { trolleyNumber ->
                    // Save current time's mark for calculate trolley tied time interval
                    appStateManager.saveCurrentTrolleyTiedTimeMark(
                        DateUtil.getCurrentDateTimestampWithDelta(appStateManager.getServerTimestamp())
                    )

                    repository.signalTrolleyTiedToUpload(trolleyNumber)
                }
                .observeOn(AndroidSchedulers.mainThread())
                .switchMap { Observable.just(TrolleyTiedPartialViewState.SuccessTakeLoaded) }
                .cast(TrolleyTiedPartialViewState::class.java)
                .onErrorResumeNext { error: Throwable ->
                    Observable.just(TrolleyTiedPartialViewState.ClearError)
                        .map {
                            TrolleyTiedPartialViewState.Error(
                                "Произошла ошибка во время взятия в работу новой тележки."
                            )
                        }
                        .cast(TrolleyTiedPartialViewState::class.java)
                }

        val onUntiedTrolleyButtonClick =
            intent(TrolleyTiedView::untiedTrolleyButtonClick)
                .flatMap { trolleyNumber ->
                    if (trolleyNumber == appStateManager.getCurrentTrolleyNumber()) {
                        repository.signalTrolleyUntiedToUpload(trolleyNumber)
                            .observeOn(AndroidSchedulers.mainThread())
                            .switchMap { Observable.just(TrolleyTiedPartialViewState.SuccessUntiedLoaded) }
                            .cast(TrolleyTiedPartialViewState::class.java)
                    } else {
                        return@flatMap Observable.just(TrolleyTiedPartialViewState.ClearError)
                            .map {
                                TrolleyTiedPartialViewState.Error(
                                    "Номера тележек не совпадают."
                                )
                            }
                            .cast(TrolleyTiedPartialViewState::class.java)
                    }
                }
                .onErrorResumeNext { error: Throwable ->
                    Observable.just(TrolleyTiedPartialViewState.ClearError)
                        .map {
                            TrolleyTiedPartialViewState.Error(
                                "Произошла ошибка во время отвязывания тележки."
                            )
                        }
                        .cast(TrolleyTiedPartialViewState::class.java)
                }

        val onNfcMessageScannedIntent =
            intent(TrolleyTiedView::nfcMessageScannedIntent)
                .flatMap { nfcMessage ->
                    Observable.just(TrolleyTiedPartialViewState.ClearError)
                        .map {
                            TrolleyTiedPartialViewState.NfcMessageScanned(nfcMessage)
                        }
                        .cast(TrolleyTiedPartialViewState::class.java)
                }

        val errorMessageIntent =
            intent(TrolleyTiedView::errorMessageIntent)
                .switchMap { errorMessage ->
                    Observable.just(TrolleyTiedPartialViewState.ClearError)
                        .map {
                            TrolleyTiedPartialViewState.Error(errorMessage)
                        }
                        .cast(TrolleyTiedPartialViewState::class.java)
                }

        val allIntentList = listOf(
            loadDetailedTaskFromDb,
            onTakeTrolleyButtonClick,
            onUntiedTrolleyButtonClick,
            errorMessageIntent,
            onNfcMessageScannedIntent
        )

        val allIntent = Observable.merge(allIntentList)
            .onErrorResumeNext { error: Throwable ->
                errorProcessing(error)
            }
            .observeOn(AndroidSchedulers.mainThread())
            .share()

        val initialState =
            TrolleyTiedViewState(
                isLoading = false,
                errorMessage = null,
                isTakeSuccess = false,
                isUntiedSuccess = false,
                trolleyNumber = appStateManager.getCurrentTrolleyNumber(),
                nfcMode = appStateManager.getCurrentWorkerSimple()?.nfcMode,
                nfcMessage = null
            )

        val stateObservable = allIntent
            .scan(initialState) { previousState, partialViewState ->
                partialViewState.reduce(previousState)
            }

        subscribeViewState(stateObservable, TrolleyTiedView::render)

    }

    private fun loadTrolleyInfo(): Observable<TrolleyTiedPartialViewState> {
        return Observable.just(TrolleyTiedPartialViewState.EmptyState)
            .cast(TrolleyTiedPartialViewState::class.java)
    }

    private fun errorProcessing(error: Throwable): Observable<TrolleyTiedPartialViewState> {
        Timber.e(error)
        return Observable.timer(3, TimeUnit.SECONDS)
            .map { Unit }
            .map { TrolleyTiedPartialViewState.ClearError as TrolleyTiedPartialViewState }
            .startWith(
                TrolleyTiedPartialViewState
                    .Error(errorMessage = error.localizedMessage)
            )
    }

    override fun unbindIntents() {
        super.unbindIntents()
    }
}
