package ru.daotech.agroinvest.fragments.welcome

import com.hannesdorfmann.mosby3.mvi.MviBasePresenter
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.Disposable
import io.reactivex.functions.Function3
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.business.Repository
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

class WelcomePresenter @Inject constructor(
    private val appStateManager: AppStateManager,
    private var goToMainScreen: (() -> Unit)?,
    private var goToDetailedTask: ((String) -> Unit)?,
    private val repository: Repository
) :
    MviBasePresenter<WelcomeView, WelcomeViewState>() {

    lateinit var disposable: Disposable

    override fun bindIntents() {

        val loadingIntent = intent(WelcomeView::loadingIntent)
            .switchMap { makeLoadingIntent() }

        val clearErrorIntent = intent(WelcomeView::clearErrorIntent)
            .map {
                WelcomePartialViewState.ClearError
            }

        val signalTaskInfoIntent = intent(WelcomeView::sendSignal)
            .switchMap { makeSignalForTaskList() }
            .take(1)

        val allIntent =
            Observable.merge(loadingIntent, signalTaskInfoIntent, clearErrorIntent).share()

        disposable = allIntent
            .filter { state ->
                state is WelcomePartialViewState.ShowNextScreen
            }
            .flatMap { repository.getLastFragmentInfo() }
            .subscribe(
                { lastFragmentState ->
                    if (!lastFragmentState.currentTaskId.isBlank() && lastFragmentState.currentTaskInterval != 0L) {
                        goToDetailedTask?.invoke(lastFragmentState.currentTaskId)
                    } else {
                        goToMainScreen?.invoke()
                    }
                },
                { error ->
                    Timber.e(error, error.localizedMessage)
                }
            )

        val initialState = WelcomeViewState()

        val stateObservable = allIntent.scan(initialState) { previousState, partialViewState ->
            partialViewState.reduce(previousState)
        }

        subscribeViewState(stateObservable, WelcomeView::render)
    }

    private fun makeLoadingIntent(): Observable<WelcomePartialViewState> {
        return appStateManager.getCurrentWorker()
            .map { worker ->
                WelcomePartialViewState.ShowWorkerInfo(worker) as WelcomePartialViewState
            }.onErrorReturn {
                Timber.e(it)
                WelcomePartialViewState.Error("Информация о пользователе не найдена")
            }
    }

    private fun makeSignalForTaskList(): Observable<WelcomePartialViewState> {
        return appStateManager.getCurrentWorker()
            .flatMap { worker ->
                Observable.zip<Unit, Unit, Unit, Unit>(
                    Observable.timer(4, TimeUnit.SECONDS).map { Unit },
                    repository.signalToTaskList(workerId = worker.id),
                    repository.signalToTrolleyInfoHistoryList(workerId = worker.id),
                    Function3 { _, _, _ -> Unit }
                )
                    .timeout(5, TimeUnit.SECONDS)
                    .onErrorResumeNext { error: Throwable ->
                        Timber.d(error)
                        Observable.just(Unit)
                    }
            }
            .observeOn(AndroidSchedulers.mainThread())
            .onErrorReturn {
                WelcomePartialViewState.Error("Информация о задачах не найдена")
            }
            .map {
                WelcomePartialViewState.ShowNextScreen
            }
    }

    override fun unbindIntents() {
        goToMainScreen = null
    }
}