package ru.daotech.agroinvest.network

import com.beust.klaxon.Klaxon
import com.beust.klaxon.KlaxonException
import com.google.gson.Gson
import io.reactivex.Observable
import io.reactivex.Single
import io.reactivex.schedulers.Schedulers
import okhttp3.MediaType
import okhttp3.MultipartBody
import okhttp3.RequestBody
import okhttp3.ResponseBody
import retrofit2.Response
import ru.daotech.agroinvest.BuildConfig
import ru.daotech.agroinvest.asObject
import ru.daotech.agroinvest.exceptions.SignOutCamundaException
import ru.daotech.agroinvest.network.models.CamundaResponse
import ru.daotech.agroinvest.network.models.FileUploadResponse
import ru.daotech.agroinvest.network.models.VariableRequest
import ru.daotech.agroinvest.service.events.Event
import ru.daotech.agroinvest.service.events.SignIn
import ru.daotech.agroinvest.service.events.SignInMessage
import ru.daotech.agroinvest.service.jsonmodels.*
import timber.log.Timber
import java.io.File
import java.io.StringReader
import java.util.concurrent.TimeUnit
import javax.inject.Inject

const val FILE_MANAGE_URL = ""

class RetrofitRepository @Inject constructor(
    private val agroinvestService: AgroinvestService
) : NetworkRepository {

    override fun signOut(signOutRequest: VariableRequest, delay: Long): Single<Response<Any>> {
        return Single.timer(delay, TimeUnit.SECONDS)
            .flatMap {
                agroinvestService.camundaSignOutRequest(signOutRequest)
                    .subscribeOn(Schedulers.io())
            }
            .timeout(15, TimeUnit.SECONDS)
            .subscribeOn(Schedulers.io())
    }

    override fun signalToCamunda(payload: String): Single<Response<CamundaResponse>> {
        return agroinvestService.signalToCamunda(payload)
            .subscribeOn(Schedulers.io())
    }

    override fun sendTaskAccidentReport(payload: VariableRequest): Observable<ResponseBody> {
        return agroinvestService.sendTaskAccidentReport(payload)
            .subscribeOn(Schedulers.io())
    }

    override fun signInByCacheUrlWithAttempts(workerCode: String): Observable<Event> {
        return Observable.timer(BuildConfig.signManageTimeoutSeconds.toLong(), TimeUnit.SECONDS)
            .flatMap { agroinvestService.signInByCacheUrl(BuildConfig.cacheSignInUrl, workerCode) }
            .subscribeOn(Schedulers.io())
            .map { response ->
                val jsonObj =
                    Klaxon().parseJsonObject(StringReader(response))

                var responseModel: QueueMessagePayload? = null

                try {
                    responseModel = Klaxon().parseFromJsonObject<SignInMessagePayload>(jsonObj)
                } catch (ex: KlaxonException) {

                }

                try {
                    responseModel = Klaxon().parseFromJsonObject<WorkerPayload>(jsonObj)
                } catch (ex: KlaxonException) {

                }

                responseModel?.let {
                    when (responseModel) {
                        is WorkerPayload -> {
                            return@map SignIn(responseModel)
                        }
                        is SignInMessagePayload -> {
                            return@map SignInMessage(responseModel)
                        }
                        else -> {
                            throw SignOutCamundaException("Некорректный ответ с сервера. Повторите попытку.")
                        }
                    }
                }
                throw SignOutCamundaException("Некорректный ответ с сервера. Повторите попытку.")
            }
            .doOnError { Timber.e(it.localizedMessage) }
            .subscribeOn(Schedulers.io())
    }

    override fun signInByCacheUrl(workerCode: String): Observable<SignIn> {
        return agroinvestService.signInByCacheUrl(BuildConfig.cacheSignInUrl, workerCode)
            .subscribeOn(Schedulers.io())
            .map { response ->
                val workerPayloadJsonObject =
                    Klaxon().parseJsonObject(StringReader(response))
                SignIn(Klaxon().parseFromJsonObject<WorkerPayload>(workerPayloadJsonObject)!!)
            }
            .doOnError(Timber::e)
            .subscribeOn(Schedulers.io())
    }

    override fun loadTasksDataByCacheUrl(workerId: String): Observable<List<TaskPayload>> {
        return agroinvestService.loadTasksDataByCacheUrl(BuildConfig.cacheTasksUrl, workerId)
            .map { response ->
                Gson().fromJson(response, ArrayList<String>()::class.java)
                    .asObject<List<TaskPayload>>()
            }
            .doOnError { Timber.e(it) }
            .subscribeOn(Schedulers.io())
    }

    override fun loadCurrentAppStatusByUrl(): Observable<ServerDataPayload> {
        return agroinvestService.loadCurrentAppStatusByUrl(BuildConfig.appStatusUrl)
            .map { response ->
                Gson().fromJson(
                    response,
                    ServerDataPayload::class.java
                )
            }
            .doOnError(Timber::e)
            .subscribeOn(Schedulers.io())
    }

    override fun fileUploadRequest(
        meta: VariableRequest,
        files: List<File>,
        filename: String
    ): Observable<FileUploadResponse> {
        val filePartList = ArrayList<MultipartBody.Part>()

        val filePart = MultipartBody.Part.createFormData(
            "database",
            filename,
            RequestBody.create(MediaType.parse("multipart/form-data"), files[0])
        )
        filePartList.add(filePart)

        if (files.size > 1) {
            val filePart_sh = MultipartBody.Part.createFormData(
                "database",
                "$filename-sh",
                RequestBody.create(MediaType.parse("multipart/form-data"), files[1])
            )

            val filePart_wal = MultipartBody.Part.createFormData(
                "database",
                "$filename-wal",
                RequestBody.create(MediaType.parse("multipart/form-data"), files[2])
            )

            filePartList.add(filePart_sh)
            filePartList.add(filePart_wal)
        }

        return agroinvestService.fileUploadRequest(
            "${BuildConfig.fileRequestSchema}${BuildConfig.fileRequestHostName}",
            filePartList
        )
            .map { it }
            .subscribeOn(Schedulers.io())
    }

    override fun screenshotUploadRequest(
        meta: VariableRequest,
        file: File,
        filename: String
    ): Observable<FileUploadResponse> {
        val filePartList = ArrayList<MultipartBody.Part>()

        val filePart = MultipartBody.Part.createFormData(
            "database",
            filename,
            RequestBody.create(MediaType.parse("multipart/form-data"), file)
        )
        filePartList.add(filePart)

        return agroinvestService.fileUploadRequest(
            "${BuildConfig.fileRequestSchema}${BuildConfig.fileRequestHostName}",
            filePartList
        )
            .map { it }
            .subscribeOn(Schedulers.io())
    }
}