package ru.daotech.agroinvest.service

import android.annotation.SuppressLint
import android.app.Notification
import android.app.Service
import android.content.Intent
import android.os.*
import dagger.android.AndroidInjection
import io.reactivex.Observable
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.subjects.PublishSubject
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.BuildConfig
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.coordinators.RootCoordinator
import timber.log.Timber
import java.util.concurrent.TimeUnit
import javax.inject.Inject

private const val CAMUNDA_SERVICE_START = "CamundaServiceStartArguments"

class CamundaService : Service() {

    @Inject
    lateinit var repository: Repository

    @Inject
    lateinit var appStateManager: AppStateManager

    @Inject
    lateinit var rootCoordinator: RootCoordinator

    private var startMode: Int = START_STICKY  // indicates how to behave if the service is killed
    private var binder: IBinder? = null        // interface for clients that bind
    private var allowRebind: Boolean = false   // indicates whether onRebind should be used

    private var serviceLooper: Looper? = null
    private var serviceHandler: ServiceHandler? = null

    private var compositeDisposable = CompositeDisposable()
    private val onFlushCamunda: PublishSubject<Unit> = PublishSubject.create()
    private val onRequestCamundaTaskList: PublishSubject<Unit> = PublishSubject.create()

    // Handler that receives messages from the thread
    private inner class ServiceHandler(looper: Looper) : Handler(looper) {
        @SuppressLint("BinaryOperationInTimber")
        override fun handleMessage(msg: Message) {
            try {
                Timber.d("CAMUNDA SERVICE: MESSAGE %s", msg)

                compositeDisposable.add(onFlushCamunda
                    .flatMap { repository.flushCamundaQueue() }
                    .doOnError { Timber.e(it) }
                    .onErrorReturnItem(Unit)
                    .subscribe(
                        { Timber.d("CAMUNDA SERVICE: QUEUE FLUSH: Done") },
                        { Timber.e("CAMUNDA SERVICE: QUEUE FLUSH: $it") }
                    )
                )

                compositeDisposable.add(onRequestCamundaTaskList
                    .flatMap { repository.signalToTaskListByCacheUrl() }
                    .doOnError { Timber.e(it) }
                    .onErrorReturnItem(Unit)
                    .subscribe(
                        { Timber.d("CAMUNDA TASK LIST REQUEST: Done") },
                        { Timber.e("CAMUNDA TASK LIST REQUEST: $it") }
                    )
                )

                // Handling task info's response from simple HTTP request
                compositeDisposable.add(appStateManager.responseTaskInfoFromHttpSubject
                    .flatMapSingle { taskPayloadList ->
                        val multipleTaskInfoHandler = ArrayList<Observable<Unit>>()
                        taskPayloadList.forEach { payload ->
                            multipleTaskInfoHandler.add(
                                repository.saveTask(
                                    payload
                                )
                            )
                        }
                        return@flatMapSingle Observable.merge(multipleTaskInfoHandler)
                            .toList()
                    }
                    .doOnError { Timber.e(it) }
                    .subscribe(
                        {
                            Timber.d("QUEUE SERVICE: TASK LIST UPSERT: Done")
                        },
                        { Timber.e("QUEUE SERVICE: TASK LIST UPSERT: $it") }
                    )
                )

                Observable.interval(10, TimeUnit.SECONDS)
                    .doOnNext {
                        Timber.d(
                            "CAMUNDA SERVICE: NETWORK STATUS: "
                                    + appStateManager.connectionReceiver.isConnected
                        )
                        onFlushCamunda.onNext(Unit)
                    }
                    .subscribe()

                Observable.interval(BuildConfig.camundaTaskListInterval.toLong(), TimeUnit.SECONDS)
                    .doOnNext {
                        Timber.d("CAMUNDA TASK LIST REQUEST:")
                        appStateManager.getCurrentWorkerSimple()?.let {
                            onRequestCamundaTaskList.onNext(Unit)
                        }
                    }
                    .subscribe()

            } catch (e: InterruptedException) {
                // Restore interrupt status
                Thread.currentThread().interrupt()
            }
        }
    }

    override fun onCreate() {
        // The service is being created
        super.onCreate()

        AndroidInjection.inject(this)

        Timber.d("CAMUNDA SERVICE: CREATE")

        // Start up the thread running the service. Note that we create a
        // separate thread because the service normally runs in the process's
        // main thread, which we don't want to block. We also make it
        // background priority so CPU-intensive work will not disrupt our UI
        HandlerThread(CAMUNDA_SERVICE_START, Process.THREAD_PRIORITY_BACKGROUND).apply {
            start()

            // Get the HandlerThread's Looper and use it for our Handler
            serviceLooper = looper
            serviceHandler = ServiceHandler(looper)
        }

        startForegroundService()
    }

    override fun onStartCommand(intent: Intent?, flags: Int, startId: Int): Int {
        // The service is starting, due to a call to startService()
        Timber.d("CAMUNDA SERVICE: START")

        // For each start request, send a message to start a job and deliver the
        // start ID so we know which request we're stopping when we finish the job
        serviceHandler?.obtainMessage()?.also { msg ->
            msg.arg1 = startId
            serviceHandler?.sendMessage(msg)
        }

        return startMode
    }

    override fun onBind(intent: Intent): IBinder? {
        // A client is binding to the service with bindService()
        Timber.d("CAMUNDA SERVICE: BIND")

        return binder
    }

    override fun onUnbind(intent: Intent): Boolean {
        // All clients have unbound with unbindService()
        Timber.d("CAMUNDA SERVICE: UNBIND")

        return allowRebind
    }

    override fun onRebind(intent: Intent) {
        // A client is binding to the service with bindService(),
        // after onUnbind() has already been called
        Timber.d("CAMUNDA SERVICE: REBIND")

        return super.onRebind(intent)
    }

    override fun onDestroy() {
        // The service is no longer used and is being destroyed
        Timber.d("CAMUNDA SERVICE: DESTROY")

        stopForegroundService()
        compositeDisposable.clear()

        super.onDestroy()
    }

    private fun startForegroundService() {
        Timber.d("CAMUNDA SERVICE: FOREGROUND SERVICE START")

        val notification: Notification? = NotificationCreator.getNotification(this)

        startForeground(NotificationCreator.getNotificationId(), notification)
    }

    private fun stopForegroundService() {
        Timber.d("CAMUNDA SERVICE: FOREGROUND SERVICE STOP")

        stopForeground(true)
        stopSelf()
    }
}
