package ru.daotech.agroinvest.service

import android.app.Notification
import android.app.Service
import android.content.Intent
import android.os.*
import com.beust.klaxon.Klaxon
import com.rabbitmq.client.*
import dagger.android.AndroidInjection
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.plugins.RxJavaPlugins
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.PublishSubject
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.BuildConfig
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.models.Worker
import ru.daotech.agroinvest.service.events.*
import ru.daotech.agroinvest.service.jsonmodels.*
import ru.daotech.agroinvest.utils.CHAT_MESSAGE_SUPPORT_CODE
import timber.log.Timber
import java.io.IOException
import java.io.StringReader
import java.util.concurrent.ExecutorService
import java.util.concurrent.Executors
import java.util.concurrent.TimeUnit
import javax.inject.Inject

private const val QUEUE_SERVICE_START = "QueueServiceStartArguments"
private const val SESSION_EXCHANGE_NAME = "common" // TODO: replace with "session"
private const val TASK_INFO_EXCHANGE_NAME = "taskInfoExchange"
private const val PERSONAL_QUEUE_NAME = "worker"

class QueueService : Service() {

    @Inject
    lateinit var repository: Repository

    @Inject
    lateinit var appStateManager: AppStateManager

    private val onSaveInfoToDb: PublishSubject<TaskPayload> = PublishSubject.create()
    private val onSaveSupportMessageInfoToDb: PublishSubject<SupportMessagePayload> =
        PublishSubject.create()
    private val onSaveTrolleyInfoToDb: PublishSubject<TrolleyInfoPayload> = PublishSubject.create()
    private val onSaveWorkerInfoToPrefs: PublishSubject<Worker> = PublishSubject.create()
    private val onUnbindTrolley: PublishSubject<TrolleyUnbindPayload> = PublishSubject.create()
    private val onSignOut: PublishSubject<SignOutPermanent> = PublishSubject.create()

    private val onConnectQueue: PublishSubject<Unit> = PublishSubject.create()

    private var startMode: Int = START_STICKY  // indicates how to behave if the service is killed
    private var binder: IBinder? = null        // interface for clients that bind
    private var allowRebind: Boolean = false   // indicates whether onRebind should be used
    private var isConnectionEstablished: Boolean =
        false   // indicates whether onRebind should be used

    private var serviceLooper: Looper? = null
    private var serviceHandler: ServiceHandler? = null

    private val factory = ConnectionFactory()

    private lateinit var connection: Connection

    private lateinit var channelMain: Channel
    private lateinit var channelTaskSync: Channel
    private lateinit var channelPersonal: Channel

    private lateinit var consumerMain: Consumer
    private lateinit var consumerTaskSync: Consumer
    private lateinit var consumerPersonal: Consumer

    private lateinit var sessionQueue: AMQP.Queue.DeclareOk
    private lateinit var taskInfoQueue: AMQP.Queue.DeclareOk
    private lateinit var personalQueue: AMQP.Queue.DeclareOk

    private var compositeDisposable = CompositeDisposable()
    private lateinit var disposableSession: Disposable
    private lateinit var disposableWorker: Disposable
    private lateinit var disposableConnectivity: Disposable
    private lateinit var disposableTimer: Disposable

    private val args: MutableMap<String, Any> = HashMap()

    // Handler that receives messages from the thread
    private inner class ServiceHandler(looper: Looper) : Handler(looper) {
        override fun handleMessage(msg: Message) {
            // Normally we would do some work here, like download a file.
            // For our sample, we just sleep for 5 seconds.
            try {
                Timber.d("QUEUE SERVICE: MESSAGE %s", msg)

                onConnectQueue.onNext(Unit)
            } catch (e: InterruptedException) {
                // Restore interrupt status.
                Thread.currentThread().interrupt()
            }
        }
    }

    inner class QueueConsumer(channel: Channel) : DefaultConsumer(channel) {
        override fun handleDelivery(
            consumerTag: String?,
            envelope: Envelope?,
            properties: AMQP.BasicProperties?,
            body: ByteArray?
        ) {
            Timber.d("QUEUE SERVICE: Handle message from ${envelope?.exchange} with binding key ${envelope?.routingKey}")

            getEventFromMessage(body!!)?.let { event ->
                appStateManager.eventSubject.onNext(event)

                // Define TaskInfo type for database saving
                if (event is TaskInfo) {
                    onSaveInfoToDb.onNext(event.taskPayload)
                }

                // Define TrolleyInfo type for database saving
                if (event is TrolleyInfo) {
                    onSaveTrolleyInfoToDb.onNext(event.trolleyInfoPayload)
                }

                // Define WorkerInfo type for prefs saving
                if (event is WorkerInfo) {
                    onSaveWorkerInfoToPrefs.onNext(event.worker)
                }

                if (event is TrolleyUnbind) {
                    onUnbindTrolley.onNext(event.trolleyUnbindPayload)
                }

                if (event is SupportMessageInfo) {
                    onSaveSupportMessageInfoToDb.onNext(event.supportMessagePayload)
                }

                if (event is UploadBackup) {
                    appStateManager.uploadBackupMessageEventSubject.onNext(Unit)
                }

                if (event is SignInMessage) {
                    Timber.d("Was interrupted? -${event.signInMessagePayload.interrupt}. SignInMessage is '${event.signInMessagePayload.message}'")
                }

                if (event is SignOutPermanent) {

                    compositeDisposable.add(
                        onSignOut
                            .subscribeOn(AndroidSchedulers.mainThread())
                            .observeOn(AndroidSchedulers.mainThread())
                            .subscribe(
                                { repository.signOutFromQueue() },
                                Timber::e
                            )
                    )

                    onSignOut.onNext(event)
                }

                if (event is SignOut) {
                    // Send signal to socket disconnect
                    appStateManager.onEmitWorkerSignStatus.onNext(false)
                }

                Timber.d("QUEUE SERVICE: Event sent")
            }

            markAsAcknowledged(envelope!!.deliveryTag)

            Timber.d("QUEUE SERVICE: Message handled")
        }

        private fun getEventFromMessage(body: ByteArray): Event? {
            return try {
                val bodyString = String(body)

                Timber.d("QUEUE SERVICE: JSON Event body: $bodyString")

                val jsonObj =
                    Klaxon().parseJsonObject(StringReader(bodyString)) //Converting string into Json Object make parsing faster. Magic!
                val queueMessage = Klaxon().parseFromJsonObject<QueueMessage>(jsonObj)

                when (queueMessage?.type) {
                    "serverData" -> ServerData(queueMessage.payload as ServerDataPayload)
                    "signIn" -> SignIn(queueMessage.payload as WorkerPayload)
                    "workerInfo" -> WorkerInfo(queueMessage.payload as WorkerPayload)
                    "signOutRemote" -> SignOutPermanent(queueMessage.payload as SignOutPayload)
                    "signOut" -> SignOut()
                    "signMessage" -> SignInMessage(queueMessage.payload as SignInMessagePayload)
                    "broadcastMessage" -> BroadcastMessage(queueMessage.payload as BroadcastMessagePayload)
                    "taskInfo" -> TaskInfo(queueMessage.payload as TaskPayload)
                    "taskStart" -> TaskStart(queueMessage.payload as TaskStartPayload)
                    "taskFinish" -> TaskFinish(queueMessage.payload as TaskFinishPayload)
                    "taskCancel" -> TaskCancel(queueMessage.payload as TaskCancelPayload)
                    "trolleyInfo" -> TrolleyInfo(queueMessage.payload as TrolleyInfoPayload)
                    "trolleyUnbind" -> TrolleyUnbind(queueMessage.payload as TrolleyUnbindPayload)
                    "supportMessage" -> SupportMessageInfo(queueMessage.payload as SupportMessagePayload)
                    "barcodeBox" -> BarcodeBoxMessage(queueMessage.payload as BarcodeBoxMessagePayload)
                    "uploadBackup" -> UploadBackup(queueMessage.payload as UploadBackupPayload)
                    "makeScreenshot" -> ScreenshotMessage(queueMessage.payload as ScreenshotMessagePayload)
                    else -> null
                }
            } catch (exception: Exception) {
                Timber.e(exception)
                null
            }
        }

        private fun markAsAcknowledged(deliveryTag: Long) {
            Timber.d("QUEUE SERVICE: Mark ACK %s", deliveryTag)

            try {
                if (channel.isOpen) {
                    channel.basicAck(deliveryTag, false)
                }
            } catch (any: Exception) {
                Timber.e(any, "QUEUE SERVICE: Error occurred during message handling")
            }
        }
    }

    override fun onCreate() {
        // The service is being created
        super.onCreate()

        AndroidInjection.inject(this)

        Timber.d("QUEUE SERVICE: CREATE")

        // Start up the thread running the service.  Note that we create a
        // separate thread because the service normally runs in the process's
        // main thread, which we don't want to block.  We also make it
        // background priority so CPU-intensive work will not disrupt our UI.
        HandlerThread(QUEUE_SERVICE_START, Process.THREAD_PRIORITY_BACKGROUND).apply {
            start()

            // Get the HandlerThread's Looper and use it for our Handler
            serviceLooper = looper
            serviceHandler = ServiceHandler(looper)

            // Init RabbitMQ Factory
            factory.isAutomaticRecoveryEnabled = true
            factory.isTopologyRecoveryEnabled = true
            factory.setUri(BuildConfig.queueHost)
            factory.virtualHost = BuildConfig.queueVirtualHost
            factory.requestedHeartbeat = 10 // 10 sec
            factory.connectionTimeout = 0 // 600000 for 600 seconds. 0 infinite
            factory.networkRecoveryInterval = 5000 // 5 sec

            args["x-max-priority"] = 5

            disposableSession = appStateManager.isSessionIdChangedSubject
                .subscribeOn(Schedulers.io())
                .observeOn(Schedulers.single())
                .subscribe(
                    { oldSessionId ->
                        rebindSessionQueue(oldSessionId = oldSessionId)
                    },
                    { error -> Timber.e(error, error.localizedMessage) }
                )

            disposableWorker = appStateManager.isWorkerChangedSubject
                .subscribeOn(Schedulers.io())
                .observeOn(Schedulers.single())
                .subscribe(
                    { worker ->
                        rebindWorkerQueues(oldWorker = worker)
                    },
                    { error -> Timber.e(error, error.localizedMessage) }
                )

            disposableConnectivity = appStateManager.connectionReceiver.connectivitySubject
                .subscribeOn(Schedulers.io())
                .observeOn(Schedulers.single())
                .subscribe(
                    { connectivityData ->
                        Timber.d("QUEUE SERVICE: CONNECTIVITY DATA: %s", connectivityData)
                        Timber.d(
                            "QUEUE SERVICE: CONNECTIVITY DATA: %s",
                            ::channelMain.isInitialized
                        )

                        if (connectivityData.first) {
                            if (!isConnectionEstablished) {
                                Timber.d("QUEUE SERVICE: REOPENING QUEUE")

                                onConnectQueue.onNext(Unit)
                            }
                        }
                    },
                    { error -> Timber.e(error, error.localizedMessage) }
                )

            compositeDisposable.add(onSaveInfoToDb
                .flatMap { taskPayload -> repository.saveTask(taskPayload) }
                .doOnError { Timber.e(it) }
                .subscribe(
                    { Timber.d("QUEUE SERVICE: TASK UPSERT: Done") },
                    { Timber.e("QUEUE SERVICE: TASK UPSERT: $it") }
                )
            )

            compositeDisposable.add(onSaveSupportMessageInfoToDb
                .flatMap { supportMessageInfo ->
                    repository.saveSupportMessageInfo(
                        supportMessageInfo,
                        CHAT_MESSAGE_SUPPORT_CODE
                    )
                }
                .doOnError { Timber.e(it) }
                .subscribe(
                    {
                        Timber.d("QUEUE SERVICE: SUPPORT MESSAGE UPSERT: Done")
                    },
                    { Timber.e("QUEUE SERVICE: SUPPORT MESSAGE UPSERT: $it") }
                )
            )

            compositeDisposable.add(onSaveTrolleyInfoToDb
                .flatMap { trolleyInfoPayload -> repository.saveTrolleyInfo(trolleyInfoPayload) }
                .doOnError { Timber.e(it) }
                .subscribe(
                    { Timber.d("QUEUE SERVICE: TROLLEY_INFO UPSERT: Done") },
                    { Timber.e("QUEUE SERVICE: TROLLEY_INFO UPSERT: $it") }
                )
            )

            compositeDisposable.add(onSaveWorkerInfoToPrefs
                .flatMap { worker -> repository.saveWorkerInfo(worker) }
                .doOnError { Timber.e(it) }
                .subscribe(
                    { Timber.d("QUEUE SERVICE: WORKER_INFO UPSERT: Done") },
                    { Timber.e("QUEUE SERVICE: WORKER_INFO UPSERT: $it") }
                )
            )

            compositeDisposable.add(onUnbindTrolley
                .flatMap { trolleyUnbindPayload ->
                    repository.unbindTrolley(trolleyUnbindPayload)
                }
                .doOnError { Timber.e(it) }
                .subscribe(
                    { Timber.d("QUEUE SERVICE: TROLLEY_UNBIND UPSERT: Done") },
                    { Timber.e("QUEUE SERVICE: TROLLEY_UNBIND UPSERT: $it") }
                )
            )

            compositeDisposable.add(onConnectQueue
                .subscribeOn(Schedulers.io())
                .observeOn(Schedulers.single())
                .throttleFirst(10, TimeUnit.SECONDS) // equal to network recovery interval
                .doOnError { Timber.e(it) }
                .subscribe(
                    {
                        Timber.d("QUEUE SERVICE: QUEUE CONNECT SUBJECT")

                        connectQueue()
                    },
                    { Timber.e("QUEUE SERVICE: QUEUE CONNECT SUBJECT ERROR: $it") }
                )
            )

            initTimer()

            startForegroundService()
        }
    }

    override fun onStartCommand(intent: Intent?, flags: Int, startId: Int): Int {
        // The service is starting, due to a call to startService()
        Timber.d("QUEUE SERVICE: START")

        // For each start request, send a message to start a job and deliver the
        // start ID so we know which request we're stopping when we finish the job
        serviceHandler?.obtainMessage()?.also { msg ->
            msg.arg1 = startId
            serviceHandler?.sendMessage(msg)
        }

        return startMode
        //return super.onStartCommand(intent, flags, startId)
    }

    override fun onBind(intent: Intent): IBinder? {
        // A client is binding to the service with bindService()
        Timber.d("QUEUE SERVICE: BIND")

        return binder
    }

    override fun onUnbind(intent: Intent): Boolean {
        // All clients have unbound with unbindService()
        Timber.d("QUEUE SERVICE: UNBIND")

        return allowRebind
    }

    override fun onRebind(intent: Intent) {
        // A client is binding to the service with bindService(),
        // after onUnbind() has already been called
        Timber.d("QUEUE SERVICE: REBIND")

        return super.onRebind(intent)
    }

    override fun onDestroy() {
        // The service is no longer used and is being destroyed
        Timber.d("QUEUE SERVICE: DESTROY")

        disconnectQueue(true)
        stopForegroundService()

        super.onDestroy()
    }

    private fun startForegroundService() {
        Timber.d("QUEUE SERVICE: FOREGROUND SERVICE START")

        val notification: Notification? = NotificationCreator.getNotification(this)

        startForeground(NotificationCreator.getNotificationId(), notification)
    }

    private fun stopForegroundService() {
        Timber.d("QUEUE SERVICE: FOREGROUND SERVICE STOP")

        stopForeground(true)
        stopSelf()
    }

    private fun connectQueue() {
        try {
            // Channel/Connection may not be opened yet https://www.rabbitmq.com/api-guide.html
            Timber.d("QUEUE SERVICE: QUEUE CONNECT: %s", ::connection.isInitialized)

            if (!::connection.isInitialized || !connection.isOpen) {
                val es: ExecutorService = Executors.newFixedThreadPool(20)

                disconnectQueue(dispose = false)

                connection = factory.newConnection(es)
                isConnectionEstablished = true

                connection.addShutdownListener { cause ->
                    // initTimer()
                    // disconnectQueue(false)
                    Timber.d("QUEUE SERVICE: QUEUE CONNECTION SHUTDOWN ${cause.isHardError} | ${cause.isInitiatedByApplication} | $cause")
                }
            }

            if (::connection.isInitialized && connection.isOpen) {

                if (!::channelMain.isInitialized || !channelMain.isOpen) {
                    channelMain = connection.createChannel()
                    channelMain.basicQos(1)

                    consumerMain = QueueConsumer(channelMain)

                    channelMain.addShutdownListener { cause ->
                        Timber.d(
                            "QUEUE SERVICE: QUEUE CHANNEL MAIN SHUTDOWN %s",
                            cause.toString()
                        )
                    }
                }

                if (!::channelTaskSync.isInitialized || !channelTaskSync.isOpen) {
                    channelTaskSync = connection.createChannel()
                    channelTaskSync.basicQos(1)

                    consumerTaskSync = QueueConsumer(channelTaskSync)

                    channelTaskSync.addShutdownListener { cause ->
                        // disconnectQueue(false)
                        Timber.d(
                            "QUEUE SERVICE: QUEUE CHANNEL SYNC SHUTDOWN %s",
                            cause.toString()
                        )
                    }
                }

                if (!::channelPersonal.isInitialized || !channelPersonal.isOpen) {
                    channelPersonal = connection.createChannel()
                    channelPersonal.basicQos(1)

                    consumerPersonal = QueueConsumer(channelPersonal)

                    channelPersonal.addShutdownListener { cause ->
                        Timber.d(
                            "QUEUE SERVICE: QUEUE CHANNEL PERSONAL SHUTDOWN %s",
                            cause.toString()
                        )
                    }
                }

                if (::channelMain.isInitialized && ::consumerMain.isInitialized && channelMain.isOpen) {
                    sessionQueue = channelMain.queueDeclare(
                        SESSION_EXCHANGE_NAME + "." + appStateManager.getSessionId(),
                        false,
                        false,
                        true,
                        args
                    )

                    Timber.d(
                        "QUEUE SERVICE: QUEUE  CONSUMER COUNT " + channelMain.consumerCount(
                            sessionQueue.queue
                        )
                    )

                    if (channelMain.consumerCount(sessionQueue.queue) <= 0) {
                        channelMain.basicConsume(
                            sessionQueue.queue,
                            false,
                            //consumerMainTag,
                            consumerMain
                        )

                        Timber.d(
                            "QUEUE SERVICE: QUEUE  CONSUMER COUNT " + channelMain.consumerCount(
                                sessionQueue.queue
                            )
                        )
                    }
                }

                if (::channelTaskSync.isInitialized && ::consumerTaskSync.isInitialized && channelTaskSync.isOpen) {
                    taskInfoQueue = channelTaskSync.queueDeclare(
                        TASK_INFO_EXCHANGE_NAME + "." + appStateManager.getSessionId(),
                        false,
                        false,
                        true,
                        args
                    )

                    if (::consumerTaskSync.isInitialized && channelTaskSync.consumerCount(
                            taskInfoQueue.queue
                        ) <= 0
                    ) {
                        channelTaskSync.basicConsume(
                            taskInfoQueue.queue,
                            false,
                            //consumerTaskSyncTag,
                            consumerTaskSync
                        )
                    }
                }

                rebindSessionQueue(oldSessionId = null)
                rebindWorkerQueues(oldWorker = null)
            }
        } catch (sse: ShutdownSignalException) {
            // possibly check if channel was closed
            // by the time we started action and reasons for
            // closing it
            Timber.d("QUEUE SERVICE: QUEUE DECLARE SSE %s", sse.toString())
        } catch (ioe: IOException) {
            // check why connection was closed
            Timber.d("QUEUE SERVICE: QUEUE DECLARE IOE %s", ioe.toString())
        } catch (ex: InterruptedException) {
            Timber.d("QUEUE SERVICE: QUEUE DECLARE Interrupted")
        } catch (any: Exception) {
            Timber.e(any, "QUEUE SERVICE: QUEUE DECLARE Exception occurred")
        }

        Timber.d("Ending handle intent")
    }

    private fun rebindSessionQueue(oldSessionId: String?) {
        Timber.d("QUEUE SERVICE: QUEUE SESSION REBIND")

        if (::channelMain.isInitialized && channelMain.isOpen) {
            channelMain.exchangeDeclare(
                SESSION_EXCHANGE_NAME,
                BuiltinExchangeType.TOPIC,
                false,
                false,
                args
            )
        }

        if (oldSessionId != null) {
            Timber.d("QUEUE SERVICE: QUEUE SESSION UNBIND %s", oldSessionId)

            unbindSessionQueue(sessionId = oldSessionId)
        }

        val newSessionId = appStateManager.getSessionId()
        Timber.d("QUEUE SERVICE: QUEUE SESSION BIND %s", newSessionId)

        if (::channelMain.isInitialized && channelMain.isOpen) {
            channelMain.queueBind(sessionQueue.queue, SESSION_EXCHANGE_NAME, newSessionId)
        }
    }

    private fun unbindSessionQueue(sessionId: String) {
        Timber.d("QUEUE SERVICE: QUEUE SESSION UNBIND %s", sessionId)

        if (::channelMain.isInitialized && channelMain.isOpen) {
            try {
                RxJavaPlugins.setErrorHandler(Throwable::printStackTrace)

                // Do queue unbinding to background thread for expect NetworkMainThreadException
                Observable.just(
                    channelMain.queueUnbind(sessionQueue.queue, SESSION_EXCHANGE_NAME, sessionId)
                )
                    .observeOn(Schedulers.single())
                    .doOnError { Timber.e(it) }
                    .subscribe()
            } catch (ex: NetworkOnMainThreadException) {
                Timber.e(ex)
            }
        }
    }

    private fun rebindWorkerQueues(oldWorker: Worker?) {
        Timber.d("QUEUE SERVICE: QUEUE WORKER REBIND")

        if (oldWorker != null) {
            Timber.d("QUEUE SERVICE: QUEUE WORKER UNBIND %s", oldWorker)

            unbindWorkerQueues(worker = oldWorker)
        }

        val newWorker = appStateManager.getCurrentWorkerSimple()

        Timber.d("QUEUE SERVICE: QUEUE WORKER BIND %s", newWorker)

        if (newWorker != null) {
            try {
                if (::channelMain.isInitialized && channelMain.isOpen) {
                    channelMain.run {
                        newWorker.exchangers.forEach {
                            if (it.name != TASK_INFO_EXCHANGE_NAME) {
                                channelMain.exchangeDeclare(it.name, BuiltinExchangeType.TOPIC)
                                channelMain.queueBind(sessionQueue.queue, it.name, it.bindingKey)
                            }
                        }
                    }
                }

                if (::channelTaskSync.isInitialized && channelTaskSync.isOpen) {
                    channelTaskSync.run {
                        newWorker.exchangers.forEach {
                            if (it.name == TASK_INFO_EXCHANGE_NAME) {
                                channelTaskSync.exchangeDeclare(it.name, BuiltinExchangeType.TOPIC)
                                channelTaskSync.queueBind(
                                    taskInfoQueue.queue,
                                    it.name,
                                    it.bindingKey
                                )
                            }
                        }
                    }
                }

                if (::channelPersonal.isInitialized && channelPersonal.isOpen) {
                    personalQueue = channelPersonal.queueDeclare(
                        PERSONAL_QUEUE_NAME + "." + newWorker.id,
                        false,
                        false,
                        false,
                        args
                    )

                    if (::consumerPersonal.isInitialized && channelPersonal.consumerCount(
                            personalQueue.queue
                        ) <= 0
                    ) {
                        channelPersonal.basicConsume(
                            personalQueue.queue,
                            false,
                            //consumerPersonalTag,
                            consumerPersonal
                        )
                    }
                }
            } catch (ex: Exception) {
                Timber.e(ex, "QUEUE SERVICE: Worker exchange were not binded")
            }
        }
    }

    private fun unbindWorkerQueues(worker: Worker?) {
        Timber.d("QUEUE SERVICE: QUEUE WORKER UNBIND %s", worker)

        if (worker != null) {
            try {
                if (::channelMain.isInitialized && channelMain.isOpen) {
                    channelMain.run {
                        worker.exchangers.forEach {
                            if (it.name != TASK_INFO_EXCHANGE_NAME) {
                                channelMain.queueUnbind(sessionQueue.queue, it.name, it.bindingKey)
                            }
                        }
                    }
                }

                if (::channelTaskSync.isInitialized && channelTaskSync.isOpen) {
                    channelTaskSync.run {
                        worker.exchangers.forEach {
                            if (it.name == TASK_INFO_EXCHANGE_NAME) {
                                channelTaskSync.queueUnbind(
                                    taskInfoQueue.queue,
                                    it.name,
                                    it.bindingKey
                                )
                            }
                        }
                    }
                }

                if (::channelPersonal.isInitialized && ::personalQueue.isInitialized && channelPersonal.isOpen) {
                    channelPersonal.run {
                        channelTaskSync.queueDelete(personalQueue.queue)
                    }
                }
            } catch (ex: Exception) {
                Timber.e(ex, "QUEUE SERVICE: Worker exchange were not binded")
            }
        }
    }

    private fun disconnectQueue(dispose: Boolean = true) {
        try {
            Timber.d("QUEUE SERVICE: QUEUE DISCONNECT: %s", dispose)

            if (::connection.isInitialized) {
                // Do queue unbinding to background thread for expect NetworkMainThreadException
//                Observable.just(connection.close())
//                    .subscribeOn(Schedulers.io())
//                    .observeOn(AndroidSchedulers.mainThread())
//                    .doOnError { Timber.e(it) }
//                    .subscribe()

            }
        } catch (ex: InterruptedException) {
            Timber.d("Interrupted")
        } catch (any: Exception) {
            Timber.e(any, "Exception occurred")
        } finally {
            if (dispose) {
                compositeDisposable.dispose()
                disposableSession.dispose()
                disposableWorker.dispose()
                disposableConnectivity.dispose()
                Timber.d("QUEUE SERVICE: QUEUE CONNECTION TIMER DISPOSE")
                if (::disposableTimer.isInitialized) {
                    disposableTimer.dispose()
                }
            }
        }

        Timber.d("Ending handle intent")
    }

    private fun initTimer() {
        if (::disposableTimer.isInitialized) {
            disposableTimer.dispose()
        }

        Timber.d("QUEUE SERVICE: QUEUE CONNECTION TIMER INIT")

        disposableTimer = Observable.timer(15, TimeUnit.SECONDS)
            .subscribeOn(Schedulers.io())
            .observeOn(Schedulers.single())
            .repeat()
            .subscribe({
                Timber.d("QUEUE SERVICE: QUEUE CONNECTION TIMER LOOP")

                onConnectQueue.onNext(Unit)
            }, { error -> Timber.e(error, error.localizedMessage) })
    }
}
