package ru.daotech.agroinvest.service

import android.app.Notification
import android.app.NotificationManager
import android.app.PendingIntent
import android.app.Service
import android.content.Context
import android.content.Intent
import android.os.IBinder
import androidx.core.app.NotificationCompat
import com.beust.klaxon.Klaxon
import com.beust.klaxon.KlaxonException
import com.github.nkzawa.emitter.Emitter
import com.github.nkzawa.socketio.client.IO
import com.github.nkzawa.socketio.client.Socket
import dagger.android.AndroidInjection
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.subjects.PublishSubject
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.BuildConfig
import ru.daotech.agroinvest.R
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.container.ContainerActivity
import ru.daotech.agroinvest.network.models.ChatDataResponse
import ru.daotech.agroinvest.network.models.ChatMessageRequest
import ru.daotech.agroinvest.network.models.ChatMessageResponse
import ru.daotech.agroinvest.service.events.*
import ru.daotech.agroinvest.service.jsonmodels.*
import ru.daotech.agroinvest.utils.CHAT_MESSAGE_SUPPORT_CODE
import timber.log.Timber
import java.io.StringReader
import java.net.URISyntaxException
import java.util.concurrent.TimeUnit
import javax.inject.Inject

class SocketService : Service() {

    @Inject
    lateinit var repository: Repository

    @Inject
    lateinit var appStateManager: AppStateManager

    private var binder: IBinder? = null
    private var allowRebind: Boolean = false

    private val socketEmitDisposable = CompositeDisposable()
    private val socketCallbackDisposable = CompositeDisposable()
    private val socketStateDisposable = CompositeDisposable()
    private lateinit var signInSocketDisposable: Disposable

    private val onNewServerMessageReceive: PublishSubject<String> = PublishSubject.create()
    private val onSignOutSubject: PublishSubject<SignOutPermanent> = PublishSubject.create()
    private val onSaveInfoToDbSubject: PublishSubject<TaskPayload> = PublishSubject.create()
    private val onCharDataUpdateSubject: PublishSubject<ChatDataResponse> = PublishSubject.create()

    private lateinit var socketInstance: Socket

    override fun onCreate() {
        super.onCreate()
        AndroidInjection.inject(this)
    }

    override fun onStartCommand(intent: Intent?, flags: Int, startId: Int): Int {
        initSocketConnection()

        return super.onStartCommand(intent, flags, startId)
    }

    override fun onBind(intent: Intent?): IBinder? {
        Timber.i("SOCKET SERVICE: BIND")
        return binder
    }

    private fun initSocketConnection() {
        try {
            if (!::socketInstance.isInitialized) {
                socketInstance = IO.socket(BuildConfig.socketUrl)

//          Local socket url
//                socketInstance = IO.socket("http://192.168.1.34:5000")

            }

            if (::socketInstance.isInitialized && !socketInstance.connected()) {
                socketInstance.connect()
            }

        } catch (ex: URISyntaxException) {
            Timber.e(ex.localizedMessage)
        }

        appStateManager.onEmitSessionSend.onNext(Unit)

        emmitSocketEvents()
        handleSocketEvents()
    }

    private fun handleSocketStateSignals() {
        if (socketStateDisposable.isDisposed || socketStateDisposable.size() != 0) {
            socketStateDisposable.clear()
        }

        signInSocketDisposable =
            appStateManager.onSocketStateRestart
                .map { authenticateWorker() }
                .doOnError { Timber.e(it) }
                .onErrorReturn { Unit }
                .subscribe()
    }

    private fun disconnectSocket() {
        if (!::socketInstance.isInitialized) {
            return
        }

        socketInstance.disconnect()
        socketInstance.close()
    }

    private fun emmitSocketEvents() {
        if (!::socketInstance.isInitialized) {
            return
        }

        // Avoiding double subscribes
        if (!socketEmitDisposable.isDisposed || socketEmitDisposable.size() != 0) {
            socketEmitDisposable.clear()
        }

        socketEmitDisposable.add(
            onNewServerMessageReceive
                .flatMap { message ->
                    repository.saveSupportMessageInfo(
                        SupportMessagePayload(
                            "SupportMessage",
                            message
                        ),
                        CHAT_MESSAGE_SUPPORT_CODE
                    )
                        .observeOn(AndroidSchedulers.mainThread())
                        .map {
                            sendGeneralNotification(
                                applicationContext,
                                CHAT_MESSAGE_SUPPORT_CODE,
                                getString(R.string.new_message_notification_title),
                                message
                            )
                        }
                }
                .doOnError { Timber.e(it) }
                .subscribe()
        )

        socketEmitDisposable.add(
            appStateManager.onSendMessageSocket
                .map { message ->
                    Timber.d("QUEUE SERVICE: Event new-message")

                    if (::socketInstance.isInitialized && socketInstance.connected()) {
                        val currentWorkerId = appStateManager.getCurrentWorkerSimple()!!.id
                        socketInstance.emit(
                            "new-message",
                            Klaxon().toJsonString(
                                ChatMessageRequest(
                                    room = "worker.${currentWorkerId}",
                                    text = message,
                                    sender = currentWorkerId
                                )
                            )
                        )
                    }
                }
                .doOnError { Timber.e(it) }
                .subscribe()
        )

        socketEmitDisposable.add(
            appStateManager.onEmitWorkerSignStatus
                .throttleFirst(2000, TimeUnit.MILLISECONDS)
                .map { status ->
                    when (status) {
                        true -> authenticateWorker()
                        else -> deAuthenticateWorker()
                    }
                }
                .onErrorReturn { Unit }
                .doOnError { Timber.e(it) }
                .subscribe()
        )

        socketEmitDisposable.add(
            appStateManager.onEmitSessionSend
                .throttleFirst(2000, TimeUnit.MILLISECONDS)
                .map { connectSession() }
                .onErrorReturn { Unit }
                .doOnError { Timber.e(it) }
                .subscribe()
        )

        socketEmitDisposable.add(
            onCharDataUpdateSubject
                .debounce(2000, TimeUnit.MILLISECONDS)
                .flatMapSingle { responseData -> repository.updatePersonalChat(responseData) }
                .doOnError { Timber.e(it) }
                .onErrorReturn { Unit }
                .subscribe()
        )
    }

    private fun connectSession() {
        appStateManager.getSessionId().let { sessionId ->
            if (::socketInstance.isInitialized && socketInstance.connected()) {
                Timber.d("SOCKET SERVICE: Event session-connect")
                socketInstance.emit(
                    "session-connect",
                    sessionId
                )
            }
        }
    }

    private fun authenticateWorker() {
        appStateManager.getCurrentWorkerSimple()?.let { worker ->
            if (::socketInstance.isInitialized && socketInstance.connected()) {
                Timber.d("SOCKET SERVICE: Event worker-connect")

                socketInstance.emit(
                    "worker-connect",
                    Klaxon().toJsonString(worker)
                )

                appStateManager.onEmitSessionSend.onNext(Unit)

                if (::signInSocketDisposable.isInitialized) {
                    signInSocketDisposable.dispose()
                }
            }
        }
    }

    private fun deAuthenticateWorker() {
        if (::socketInstance.isInitialized && socketInstance.connected()) {
            Timber.d("SOCKET SERVICE: Event worker-disconnect")

            socketInstance.emit(
                "worker-disconnect",
                emptyList<String>()
            )

            handleSocketStateSignals()
        }
    }

    private fun handleSocketEvents() {
        if (!::socketInstance.isInitialized) {
            return
        }

        socketInstance.apply {
            on(
                "last-message",
                onNewMessageFromServer
            ) // To update if someone send a message to chat room.

            on(
                "update-chat",
                onInitialMessagesFromServer
            ) // To know if the user left the chat room.

            once(
                "receiverInstance",
                receiverInstance
            )

            once(
                "connectError",
                onConnectError
            )

            once(
                "disconnect",
                onDisconnect
            )

            on(
                "connect",
                onConnect
            )

            once(
                "timeoutError",
                onTimeoutError
            )

            once(
                "app-event",
                onSocketEventReceiver
            )

            once(
                "reconnect_attempt",
                onReconnectAttempt
            )

            once(
                "reconnect",
                onReconnect
            )

            once(
                "reconnecting",
                onReconnecting
            )

            once(
                "reconnect_error",
                onReconnectError
            )
        }
    }

    private var onNewMessageFromServer = Emitter.Listener {
        Timber.d("SOCKET SERVICE: Event new-message ${it[0] as String?}")

        val lastMessageJsonResponse = it[0] as String?

        if (lastMessageJsonResponse.isNullOrBlank()) {
            return@Listener
        }

        val chatMessageData = Klaxon().parseJsonObject(StringReader(lastMessageJsonResponse))
        if (!chatMessageData.isNullOrEmpty()) {
            Klaxon().parseFromJsonObject<ChatMessageResponse>(chatMessageData)?.let { lastMessage ->
                if (lastMessage.sender == "sender") {
                    onNewServerMessageReceive.onNext(lastMessage.text)
                }
            }
        }
    }

    private var onInitialMessagesFromServer = Emitter.Listener {
        Timber.d("SOCKET SERVICE: Event update-chat ${it[0] as String?}")

        val chatMessageDataJson = it[0] as String?

        if (chatMessageDataJson.isNullOrBlank()) {
            return@Listener
        }

        val chatMessageData = Klaxon().parseJsonObject(StringReader(chatMessageDataJson))

        try {
            if (!chatMessageData.isNullOrEmpty()) {
                Klaxon().parseFromJsonObject<ChatDataResponse>(chatMessageData)?.let { chatResponse ->
                    onCharDataUpdateSubject.onNext(chatResponse)
                }
            }
        } catch (ex: KlaxonException) {
            Timber.e(ex.localizedMessage)
        }
    }

    private var receiverInstance = Emitter.Listener {
        Timber.d("SOCKET SERVICE: Event handled")
    }

    private var onConnectError = Emitter.Listener { _ ->
        Timber.d("SOCKET SERVICE: Connection error handled")
    }

    private var onDisconnect = Emitter.Listener { _ ->
        Timber.d("SOCKET SERVICE: Disconnect socket handled")

        // Update socket connectivity state - DISCONNECTED
        appStateManager.onSocketConnectivityState.onNext(false)
    }

    private var onConnect = Emitter.Listener {
        Timber.d("SOCKET SERVICE: Connect socket handled")

        if (::socketInstance.isInitialized && socketInstance.connected()) {
            appStateManager.onEmitWorkerSignStatus.onNext(true)
        }

        // Refill disposable
        handleSocketStateSignals()

        // Update socket connectivity state - CONNECTED
        appStateManager.onSocketConnectivityState.onNext(true)
    }

    private var onTimeoutError = Emitter.Listener {
        Timber.d("SOCKET SERVICE: Timeout error handled")
    }

    private var onReconnectAttempt = Emitter.Listener {
        Timber.d("SOCKET SERVICE: Reconnect attempt handled")
    }

    private var onReconnect = Emitter.Listener {
        Timber.d("SOCKET SERVICE: Reconnect handled")

        // Update socket connectivity state - CONNECTED
        appStateManager.onSocketConnectivityState.onNext(true)
    }

    private var onReconnecting = Emitter.Listener {
        Timber.d("SOCKET SERVICE: Reconnecting handled")
    }

    private var onReconnectError = Emitter.Listener {
        Timber.d("SOCKET SERVICE: Reconnect error handled")

        // Update socket connectivity state - DISCONNECTED
        appStateManager.onSocketConnectivityState.onNext(false)
    }

    private var onSocketEventReceiver = Emitter.Listener { it ->
        Timber.d("SOCKET SERVICE: Handle message event ${it[0] as String?}")

        (it[0] as String?)?.let {

            handleSocketEvents(it)?.let { event ->

                // Avoiding double subscribes
                if (!socketCallbackDisposable.isDisposed || socketCallbackDisposable.size() != 0) {
                    socketCallbackDisposable.clear()
                }

                appStateManager.eventSubject.onNext(event)

                // Define TaskInfo type for database saving
                if (event is TaskInfo) {

                    socketCallbackDisposable.add(
                        onSaveInfoToDbSubject
                            .flatMap { taskPayload ->
                                repository.saveTask(taskPayload)
                            }
                            .doOnError { error -> Timber.e(error.localizedMessage) }
                            .onErrorReturn { Unit }
                            .subscribe(
                                {
                                    Timber.d("SOCKET SERVICE: TASK UPSERT: Done")
                                },
                                {
                                    Timber.e("SOCKET SERVICE: TASK UPSERT: $it")
                                }
                            )
                    )

                    onSaveInfoToDbSubject.onNext(event.taskPayload)
                }

                // Define SignOutRemote event type
                if (event is SignOutPermanent) {

                    if (socketCallbackDisposable.isDisposed || socketCallbackDisposable.size() <= 0) {
                        socketCallbackDisposable.add(
                            onSignOutSubject
                                .subscribeOn(AndroidSchedulers.mainThread())
                                .observeOn(AndroidSchedulers.mainThread())
                                .subscribe(
                                    { repository.signOutFromQueue() },
                                    Timber::e
                                )
                        )
                    }

                    onSignOutSubject.onNext(event)
                }

                if (event is SignOut) {
                    // Disconnect current worker from room
                    appStateManager.onEmitWorkerSignStatus.onNext(false)
                }
            }
        }
    }

    private fun handleSocketEvents(bodyString: String): Event? {
        return try {
            //Converting string into Json Object make parsing faster. Magic!
            val jsonObj =
                Klaxon().parseJsonObject(StringReader(bodyString))
            val queueMessage = Klaxon().parseFromJsonObject<QueueMessage>(jsonObj)

            when (queueMessage?.type) {
                "signIn" -> SignIn(queueMessage.payload as WorkerPayload)
                "signOut" -> SignOut()
                "signOutRemote" -> SignOutPermanent(queueMessage.payload as SignOutPayload)
                "taskInfo" -> TaskInfo(queueMessage.payload as TaskPayload)
                else -> null
            }
        } catch (exception: Exception) {
            Timber.e(exception)
            null
        }
    }

    private fun sendGeneralNotification(
        context: Context,
        uniqueId: Int,
        title: String,
        contentText: String
    ) {
        val notificationManagerCompat =
            context.getSystemService(Context.NOTIFICATION_SERVICE) as NotificationManager
        val builder = NotificationCompat.Builder(context)
        builder.setAutoCancel(true)
        builder.setContentTitle(title)
        builder.setContentText(contentText)
        builder.setGroup("Chat message group")
        builder.setDefaults(Notification.DEFAULT_ALL)
        builder.setStyle(
            NotificationCompat.BigTextStyle()
                .setSummaryText(title)
                .setBigContentTitle(title)
                .bigText(contentText)
        )
        val requestsViewIntent = Intent(context, ContainerActivity::class.java)
        requestsViewIntent.flags = (Intent.FLAG_ACTIVITY_CLEAR_TOP
                or Intent.FLAG_ACTIVITY_SINGLE_TOP or Intent.FLAG_ACTIVITY_NEW_TASK)
        requestsViewIntent.addFlags(
            Intent.FLAG_ACTIVITY_CLEAR_TOP
                    or Intent.FLAG_ACTIVITY_SINGLE_TOP or Intent.FLAG_ACTIVITY_NEW_TASK
        )
        val requestsViewPending = PendingIntent.getActivity(
            context,
            Integer.valueOf(uniqueId),
            requestsViewIntent,
            0
        )
        builder.setContentIntent(requestsViewPending)
        builder.setSmallIcon(R.drawable.foreground_service_icon)
        builder.setShowWhen(true)
        val notification: Notification = builder.build()
        notificationManagerCompat.notify(Integer.valueOf(uniqueId), notification)
    }

    private fun getNotification(): Notification? {
        val notification: Notification
        val builder: NotificationCompat.Builder = NotificationCompat.Builder(this)
        builder.setColor(resources.getColor(R.color.colorAccent)).setAutoCancel(true)
        notification = builder.build()
        notification.flags = Notification.FLAG_FOREGROUND_SERVICE or Notification.FLAG_AUTO_CANCEL
        return notification
    }

    private fun stopForegroundService() {
        Timber.d("SOCKET SERVICE: FOREGROUND SERVICE STOP")

        socketInstance.io().off()
        socketInstance.close()
        socketInstance.disconnect()

        socketEmitDisposable.clear()
        socketCallbackDisposable.clear()
        socketStateDisposable.clear()

        socketEmitDisposable.dispose()
        socketCallbackDisposable.dispose()
        socketStateDisposable.dispose()

        stopForeground(true)
        stopSelf()
    }

    override fun onUnbind(intent: Intent): Boolean {
        Timber.d("SOCKET SERVICE: UNBIND")

        return allowRebind
    }

    override fun onRebind(intent: Intent) {
        Timber.d("SOCKET SERVICE: REBIND")

        return super.onRebind(intent)
    }

    override fun onDestroy() {
        super.onDestroy()

        disconnectSocket()
        stopForegroundService()
    }
}