package ru.daotech.agroinvest.service.couch

import android.app.Service
import android.content.Intent
import android.os.IBinder
import com.couchbase.lite.*
import dagger.android.AndroidInjection
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.PublishSubject
import ru.daotech.agroinvest.AppStateManager
import ru.daotech.agroinvest.BuildConfig
import ru.daotech.agroinvest.asObject
import ru.daotech.agroinvest.business.Repository
import ru.daotech.agroinvest.exceptions.CouchDatabaseNotInitializedYedException
import ru.daotech.agroinvest.models.Worker
import ru.daotech.agroinvest.service.events.Event
import ru.daotech.agroinvest.service.events.SignOut
import ru.daotech.agroinvest.service.events.TaskInfo
import ru.daotech.agroinvest.service.jsonmodels.TaskPayload
import timber.log.Timber
import java.net.URI
import javax.inject.Inject

class CouchDbService : Service() {

    @Inject
    lateinit var repository: Repository

    @Inject
    lateinit var appStateManager: AppStateManager

    private var binder: IBinder? = null

    private val onSaveInfoToDb: PublishSubject<TaskPayload> = PublishSubject.create()

    private lateinit var disposable: Disposable
    private val compositeDisposable = CompositeDisposable()

    private lateinit var databaseInstance: Database
    private lateinit var replicator: Replicator
    private lateinit var currentListenerToken: ListenerToken

    override fun onCreate() {
        super.onCreate()
        AndroidInjection.inject(this)
        CouchbaseLite.init(this)
    }

    override fun onBind(intent: Intent?): IBinder? {
        Timber.i("COUCHDB SERVICE: BIND")
        return binder
    }

    override fun onStartCommand(intent: Intent?, flags: Int, startId: Int): Int {
        buildDisposableHandler()

        return super.onStartCommand(intent, flags, startId)
    }

    private fun buildDisposableHandler() {
        if (::disposable.isInitialized && !disposable.isDisposed) {
            disposable.dispose()
        }

        disposable = appStateManager.onInitiateConnectToCouchDb
            .map {
                appStateManager.getCurrentWorkerSimple()?.let {
                    buildCouchDbConnection(it)
                }
            }
            .subscribeOn(Schedulers.io())
            .doOnError { Timber.e(it.localizedMessage) }
            .subscribe()

        compositeDisposable.add(
            appStateManager.onCloseCouchDatabaseConnection
                .map { disposeReplicator() }
                .doOnError { Timber.e(it) }
                .subscribe()
        )

        compositeDisposable.add(
            onSaveInfoToDb
                .flatMap { taskPayload ->
                    repository.saveTask(taskPayload)
                }
                .subscribeOn(Schedulers.single())
                .observeOn(Schedulers.single())
                .doOnError { Timber.e(it) }
                .subscribe(
                    {
                        Timber.d("COUCHDB RESPONSE: TASK UPSERT: Done")
                    },
                    { Timber.e("COUCHDB RESPONSE: TASK UPSERT: $it") }
                )
        )
    }

    private fun disposeReplicator() {
        if (::replicator.isInitialized) {
            replicator.stop()
            if (::currentListenerToken.isInitialized) {
                try {
                    replicator.removeChangeListener(currentListenerToken)
                } catch (ex: NullPointerException) {
                    Timber.e(ex.localizedMessage)
                }
            }
        }
    }

    private fun buildCouchDbConnection(worker: Worker?) {
        val databaseConfig = DatabaseConfiguration()
        databaseInstance = Database(BuildConfig.couchName, databaseConfig)

        // Create replicators to push and pull changes to and from the cloud
        val targetEndpoint: Endpoint = URLEndpoint(URI(BuildConfig.couchUrl))
        val replicatorConfiguration = ReplicatorConfiguration(databaseInstance, targetEndpoint)

        worker?.let {
            // Add authentication by login and password
            replicatorConfiguration.authenticator = BasicAuthenticator(
                "worker.${it.id}",
                it.id
            )
        }
        replicatorConfiguration.replicatorType = ReplicatorHelper.getReplicatorTypeFor(
            true,
            false
        )
        replicatorConfiguration.isContinuous = true

        disposeReplicator() // for correct reinitializing

        replicator = Replicator(replicatorConfiguration)

        // Listen to replicator change events
        replicator.addChangeListener { change: ReplicatorChange ->
            Timber.d("COUCHDB RESPONSE: CHANGE EVENT CAUGHT $change")
            if (change.status.error != null) {
                Timber.d("Error code: %s", change.status.error.code)
            }
        }

        // Listen to replicator change documents
        currentListenerToken =
            replicator.addDocumentReplicationListener { replica: DocumentReplication ->
                replica.documents.forEach { document ->
                    uploadResponseToInnerDatabase(document)?.let { event ->
                        if (event is TaskInfo) {
                            onSaveInfoToDb.onNext(event.taskPayload)
                        }

                        if (event is SignOut) {
                            if (appStateManager.getCurrentWorkerSimple() != null) {
                                appStateManager.eventSubject.onNext(event)
                            }
                        }
                    }
                }
            }

        replicator.start()
    }

    private fun uploadResponseToInnerDatabase(replicaDocument: ReplicatedDocument): Event? {
        if (!::databaseInstance.isInitialized) {
            throw CouchDatabaseNotInitializedYedException
        }

        val document = databaseInstance.getDocument(replicaDocument.id)

        return try {
            when (document.toMap()["type"]) {
                "task" -> {
                    val couchBasePayload =
                        document.toMap()["payload"]?.asObject<TaskPayload>()
                    TaskInfo(couchBasePayload!!)
                }
                "signOut" -> {
                    SignOut()
                }
                else -> null
            }
        } catch (exception: Exception) {
            Timber.e(exception)
            null
        }
    }

    override fun onDestroy() {
        super.onDestroy()

        if (::disposable.isInitialized) {
            disposable.dispose()
        }

        compositeDisposable.clear()
    }
}