package ru.daotech.agroinvest.utils

import android.annotation.SuppressLint
import android.content.Context
import android.content.ContextWrapper
import android.database.Cursor
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.net.Uri
import android.nfc.NdefMessage
import android.nfc.NdefRecord
import android.os.Build
import android.os.Environment
import android.os.VibrationEffect
import android.os.Vibrator
import android.provider.OpenableColumns
import android.text.format.DateFormat
import android.util.Base64
import android.view.View
import org.joda.time.DateTime
import ru.daotech.agroinvest.database.relation.TaskWithProperties
import ru.daotech.agroinvest.models.Worker
import timber.log.Timber
import java.io.*
import java.io.File.separator
import java.util.*
import java.util.regex.Pattern
import kotlin.collections.ArrayList
import kotlin.experimental.and


class ServiceUtil {

    companion object {

        private const val DB_FILE_FORMAT = ".db"
        const val SQLITE_FILE_FORMAT = ".sqlite"
        const val SCREEN_FILE_FORMAT = ".jpg"
        private const val FILEPATH_SEPARATOR = "/"
        private const val BOTTOM_LINE_SEPARATOR = "_"
        private const val VIBRATE_DURATION_MILLISECONDS = 100L
        private const val FULL_NAME_SPLITTER = " "
        private const val SHORTEN_NAME_SPLITTER = ". "

        fun cutOffExcessFromNfcMessage(nfcMessage: String): String {
            val regex = Pattern.compile("[0-9]+")
            val regexMatcher = regex.matcher(nfcMessage)
            if (regexMatcher.find()) {
                return regexMatcher.group()
            }
            return nfcMessage
        }

        fun exportDB(databasePath: String): List<File>? {
            try {
                val filesToExport = ArrayList<File>()
                val dbFileFolder = File(databasePath)
                for (dbFile in dbFileFolder.listFiles()) {
                    val fileInputStream = FileInputStream(dbFile)

                    val outFileName: String =
                        Environment.getExternalStorageDirectory().absolutePath +
                                separator +
                                dbFile.name +
                                DB_FILE_FORMAT
                    val file = File(outFileName)
                    file.createNewFile()

                    // Open the empty db as the output stream
                    val output: OutputStream = FileOutputStream(file)

                    // Transfer bytes from the inputfile to the outputfile
                    val buffer = ByteArray(1024)
                    var length: Int
                    while (fileInputStream.read(buffer)
                            .also { length = it } > 0
                    ) {
                        output.write(buffer, 0, length)
                    }

                    // Close the streams
                    output.flush()
                    output.close()
                    fileInputStream.close()
                    filesToExport.add(file)
                }
                return filesToExport
            } catch (e: IOException) {
                Timber.e(e)
            }
            return null
        }

        fun calculateTaskIntervalsSum(
            taskWithProperties: TaskWithProperties,
            serverTimestamp: Int
        ): Long {

            var intervalTime: Long = 0

            for (interval in taskWithProperties.intervals) {
                if (!interval.finishedAt.isBlank() && !interval.startedAt.isBlank()) {
                    intervalTime += DateUtil.getMillsBetweenTwoDates(
                        finishedAt = interval.finishedAt,
                        startedAt = interval.startedAt
                    )
                }
            }

            // If task in progress, the resulted time of intervals increase on current time mark
            if (!taskWithProperties.startedAt.isNullOrBlank()
                && taskWithProperties.finishedAt.isNullOrBlank()
            ) {
                with(taskWithProperties.intervals.last()) {
                    if (finishedAt.isBlank() && !startedAt.isBlank()) {
                        intervalTime += DateUtil.getMillsBetweenTwoDates(
                            finishedAt = DateUtil.getDateTimeWithDelta(
                                DateTime.now(),
                                serverTimestamp
                            ),
                            startedAt = startedAt
                        )
                    }
                }
            }
            return intervalTime
        }

        /**
         * Parse the NDEF message contents and return it
         * https://stackoverflow.com/questions/7917567/strange-character-on-android-ndef-record-payload
         */
        fun processNdefMessages(ndefMessages: List<NdefMessage?>): String? {
            return ndefMessages.firstOrNull()?.records?.firstOrNull()?.run {
                var message: String? = null
                if (this.tnf == NdefRecord.TNF_WELL_KNOWN
                    && NdefRecord.RTD_TEXT!!.contentEquals(this.type)
                ) {
                    val status = this.payload.first()
                    val enc = status and 0x80.toByte()
                    val charset = if (enc == 0.toByte()) {
                        Charsets.UTF_8
                    } else {
                        Charsets.UTF_16
                    }
                    val ianaLength = (status and 0x3F.toByte()) + 1 // IANA length + first byte
                    message = String(
                        this.payload,
                        ianaLength,
                        this.payload.size - ianaLength,
                        charset
                    )
                }
                message
            }
        }

        fun makeVibrationEffect(context: Context) {
            val vib: Vibrator =
                context.getSystemService(Context.VIBRATOR_SERVICE) as Vibrator

            if (vib.hasVibrator()) {
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
                    // void vibrate (VibrationEffect vibe)
                    vib.vibrate(
                        VibrationEffect.createOneShot(
                            VIBRATE_DURATION_MILLISECONDS,
                            // The default vibration strength of the device.
                            VibrationEffect.DEFAULT_AMPLITUDE
                        )
                    )
                } else {
                    // This method was deprecated in API level 26
                    vib.vibrate(VIBRATE_DURATION_MILLISECONDS)
                }
            }
        }

        @SuppressLint("Recycle")
        fun getFileName(uri: Uri): String? {
            return uri.path?.split(FILEPATH_SEPARATOR)?.last()
        }

        fun buildUploadedFilename(
            worker: Worker?,
            formatCode: String,
            serverTimestamp: Int
        ): String {
            val filename = StringBuilder()
            return filename
                .append(worker?.id)
                .append(BOTTOM_LINE_SEPARATOR)
                .append(worker?.personalNumber)
                .append(BOTTOM_LINE_SEPARATOR)
                .append(DateUtil.getCurrentDateTimeWithDeltaForRequest(serverTimestamp))
                .append(formatCode)
                .toString()
        }

        fun viewToFile(view: View): File? {
            val now = Date()
            DateFormat.format("yyyy-MM-dd_hh:mm:ss", now)
            try { // image naming and path  to include sd card  appending name you choose for file

                val outFileName: String =
                    Environment.getExternalStorageDirectory().absolutePath +
                            separator +
                            now +
                            ".jpg"

                // create bitmap screen capture
                view.setDrawingCacheEnabled(true)
                val bitmap: Bitmap = Bitmap.createBitmap(view.getDrawingCache())
                view.setDrawingCacheEnabled(false)

                val imageFile = File(outFileName)

                try {
                    imageFile.createNewFile()
                } catch (ex: SecurityException) {
                    Timber.e("Permission for file creating was denied")
                }

                val outputStream = FileOutputStream(imageFile)

                val quality = 100
                bitmap.compress(Bitmap.CompressFormat.JPEG, quality, outputStream)
                outputStream.flush()
                outputStream.close()
                return imageFile
            } catch (e: Throwable) {
                e.printStackTrace()
            }
            return null
        }

        fun getShortenFullName(fullName: String): String {
            val fullNameArr = fullName.split(FULL_NAME_SPLITTER)
            var shortenName = fullNameArr[0] + FULL_NAME_SPLITTER
            fullNameArr.forEachIndexed { index, _ ->
                if (index != 0) {
                    shortenName += fullNameArr[index].toCharArray()[0] + SHORTEN_NAME_SPLITTER
                }
            }
            return shortenName
        }

        fun bitmapToFile(bitmap: Bitmap, context: Context): Map<String, Any> {
            // Get the context wrapper
            val wrapper = ContextWrapper(context)

            // Initialize a new file instance to save bitmap object
            var file = wrapper.getDir("Images", Context.MODE_PRIVATE)
            val fileName = "${UUID.randomUUID()}.jpg"
            file = File(file, fileName)

            try {
                // Compress the bitmap and save in jpg format
                val stream: OutputStream = FileOutputStream(file)
                bitmap.compress(Bitmap.CompressFormat.JPEG, 100, stream)
                stream.flush()
                stream.close()
            } catch (e: IOException) {
                e.printStackTrace()
            }

            val resultMap: MutableMap<String, Any> = HashMap()
            resultMap["fileName"] = fileName
            resultMap["fileUri"] = Uri.parse(file.absolutePath) as Uri

            return resultMap
        }
    }
}